//
// The sample code given here modifies the coordinates of a vertices
// subject to the basic physical laws governing dynamics.
//

//////////  The data types will need to be modified
//////////  to fit whatever framework you are using.

typedef float VECTOR[3];
typedef float MATRIX[4][4];

typedef struct tagVERTEX {
             // add other members as required
 VECTOR p;   // p[2] is the height coordinate
} VERTEX;

///////// The following are standard functions and definitions that are
///////// usually part of a compilers library.
///////// Substitute then with the ones you wish to use.

extern float fabs(float D);       // return floating point absolute value of D
extern float sqrt(float D);       // return floating point square root of  D

#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif

////////// The following functions will need to be provided to fulfil the
////////// described action.

// scale the vector A with the scalar s,   R = sA
extern void ScaleVector(float s,VECTOR A, VECTOR R);
// copy a vector from Source to Destination
extern void CopyVector(VECTOR Source, VECTOR Destination);
// add two vectors R = A + B
extern void AddVectors(VECTOR A, VECTOR B, VECTOR R);
// subtract two vectors R = A - B
extern void SubtractVectors(VECTOR A, VECTOR B, VECTOR R);
// vector dot product R = A . B
extern float VectorDotProduct(VECTOR A, VECTOR B);
// vector cross product R = A x B
extern void VectorCrossProduct(VECTOR A, VECTOR B, VECTOR R);
// transform a from one vector to another using the matrix M
extern void  TransformVector(MATRIX M, VECTOR Va, VECTOR Vr);
// rotate round axis was given in the descrition of the modeling algorithm
// to create surfaces of revolution.
extern void RotateRoundAxis(MATRIX transform, VECTOR base,
                            VECTOR direction, float angle);
// normalize the length of a vector
extern void Normalize(VECTOR v);

//////// a macro provide a useful way to implement a small function

// Euclidian distance squared between two points defined with
// position vectors "a2 and "b"
#define E_Distance2(a,b) (                                          \
                         (a[0]-(double)b[0]) * (a[0]-(double)b[0])  \
                        +(a[1]-(double)b[1]) * (a[1]-(double)b[1])  \
                        +(a[2]-(double)b[2]) * (a[2]-(double)b[2])  \
                         )

////////////////////////////////////////////////////////////////////////////
////////////////////     The examples    ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////

//
// prototypes for functions used internally
//
static float CalculatePosition(long, float, float, float, float,
                               float, float);
static void GetTandA(int, float, float, float *, float *);


static float GetPseudoTime(float, VECTOR, float, float);
static void GetSpin(int, VECTOR, VECTOR, float, float [4][4]);
static void GetVandA(float, float, long, float *, float *);
static void GetPosition(VECTOR, VECTOR, VECTOR,
                        float, float, float,
                        float, float, float,
                        VECTOR, long);
static void SpinIntoPosition(VECTOR, VECTOR,
                             float, float [4][4]);
static void MoveIntoPosition(VECTOR, VECTOR, float);


////////////////////////////////////////////////////////////////////////////
// The first action simulates the appearance of a model falling under
// gravity. There are two options, in the first the model appears to act
// a a rigid body, the second gives the illusion that the model is "floppy"
// because each vertex is treated individually and therefore when a vertex
// hits the "floor" it bounces independently from the other vertices in
// the model
////////////////////////////////////////////////////////////////////////////

void ExecuteAnimationEffectFall(
  long Nvertices,          // The number of vertices to be modified
  VERTEX *VertexList,      // The list of vertices to be modified
  float t,                 // This is the animation parameter that specifies
                           // how far the action has "progressed". When
                           // t=0 the action is at the start, when t=1
                           // the action is over, when t=0.5 the action is
                           // half completed.
  float z_ground,          // The Z coordinate of "floor" level.
  int nb,                  // The number of bounces the model is to perform
  float mu,                // The coefficient of restitution < 0.5
  int type                 // Control whether all the vertices act as one
                           // or each bounces when it hits the floor
                           // if type=0 the model behaves as a single
                           // rigid unit if type=1 the illusion of
                           // flaxibility is created.

){
 long i;
 float z,zmax=-1.e20,zmin=1.e20,d,t0,a;
 //
 // Find the overall height of the model
 //
 for(i=0;i<Nvertices;i++)
 {
   if(VertexList[i].p[2] > zmax)zmax=VertexList[i].p[2];
   if(VertexList[i].p[2] < zmin)zmin=VertexList[i].p[2];
 }
 //
 // Determine the distance the model is to fall when
 //
 d=zmin-z_ground;
 //
 // Calculate the acceleration used to simulate that due to gravity "a"
 // and the proportion of the allotted time for the model to fall to the
 // floor for the first time.
 //
 GetTandA(nb,mu,d,&t0,&a);
 if(type == 0)
 {
   //
   // Calculate the new Z coordinate appropriate for time "t"
   //
   z=(float)CalculatePosition(nb,zmax,zmax,t,mu,a,d);
   //
   // Apply it to all vertices
   //
   z=zmax-z;
   for(i=0;i<Nvertices;i++)VertexList[i].p[2] -= z;
 }
 else
 {
   //
   // Calculate the new Z coordinate appropriate for time "t" do it for
   // each vertex one at a time (this allows each vertex to bounce when
   // it hits the floor.
   //
   for(i=0;i<Nvertices;i++)
     VertexList[i].p[2]=(float)CalculatePosition(
         nb,zmax,VertexList[i].p[2],t,mu,a,d);
 }
 return;
}

static float CalculatePosition(
 long n,
 float zm,
 float z0,
 float t,
 float mu,
 float a,
 float d
){
 //
 // Calculate the "Z" coordinate of an object falling from a height "z0"
 // and bouncing when it hits the ground.  "t" is the fraction (0-1) of
 // "when" the desired position should be returned. (t=0 at start of fall.
 // When t=1 the object has hit the ground after the prescribed number of
 // bounces "n".
 //
 int i;
 float dfall,ta,dt,h,deltaT;
 dfall=z0-(zm-d);
 ta=sqrt(2.0*dfall/a);
 //
 // First hit
 //
 if(t < ta)return (z0-a*t*t*0.5);
 dt=mu*ta; h=dfall;
 //
 // Find in which bounce the model is at time "t"
 //
 if(n > 0)for(i=0;i<3*n;i++)
 {
   h=mu*mu*h;
   if(t < ta+dt)
   {
     //
     // on the way up
     //
     deltaT = dt - (t - ta);
     return (z0-(dfall-h+a*deltaT*deltaT*0.5));
   }
   ta += dt;
   if(t < ta+dt)
   {
     //
     // on the way down
     //
     deltaT = (t - ta);
     return (z0-(dfall-h+a*deltaT*deltaT*0.5));
   }
   ta += dt;
   dt *= mu;
 }
 return (z0-dfall);
}

static void GetTandA(
 int nb,
 float mu,
 float d,
 float *t,
 float *a
){
 //
 // Determine the acceleration "a" and time for first fall "t" given
 // a specific number of bounces "nb", distance to fall "d" and coefficent
 // of restitiution "mu"
 //
 int i;
 float t0;
 if(nb > 0)
 {
   //
   // add in the effect of a number of bounces
   //
   t0=mu;
   for(i=0;i<nb-1;i++)t0 = mu*(t0 + 1.0);
   t0 *= 2.0;
   t0 += 1.0;
 }
 else t0=1.0;
 t0  = 1.0/t0;
 *a=2.0*d/(t0*t0);
 *t=t0;
 return;
}



////////////////////////////////////////////////////////////////////////////
// The second action simulates the effect of an explosion on a model.
// The simulation works by treating groups of three vertices as the vertices
// that define disconnected triangular facets. Each faect is projected
// radially from the seat of an explosion with a high initial velocity.
// The facets (given by three consecutive vertices in the vertex list) may
// optionally fall "to the floor" under the influence of gravity and/or
// be projected at the initial instant of the explosion or at a time later
// which is proporitional to their distance from the site of the explosion
// (to simulate a shock front).
////////////////////////////////////////////////////////////////////////////


void ExecuteAnimationEffectExplode(
  long Nvertices,          // The number of vertices to be modified
  VERTEX *VertexList,      // The list of vertices to be modified
  float t,                 // This is the animation parameter that specifies
                           // how far the action has "progressed". When
                           // t=0 the action is at the start, when t=1
                           // the action is over, when t=0.5 the action is
                           // half completed.
  VECTOR po,               // A position vector to the site of the explosion
  int type,                // type=0 gives a radial explosion without the
                           // influence of gravity, type=1 brings into play
                           // gravitational forces causing the facets to
                           // fall back to ground
  int ishock,              // Percentage value. When 0 all parts of model
                           // begin explosion simultaneously when 100% the
                           // furthest parts fly off after half the time
                           // alloted to the action. (t=0.5)
  int tumble,              // if tumble=1 them facets tumble while moving away
  int tumbletype,          // if tumbletype=0 the facets sping round a vector
                           // which points "up" if tumbletype=1 the facets
                           // tumble round two axes one of which precesses
                           // round the other for added effect
  float df,                // distance of "floor" below the site of explosion
  float dex                // maximum size of the explosion
){
 long i;
 float tt,V,A,spin[4][4],floor,zmax,dmin,dmax,di,shock;
 VECTOR p,pp,d,dd,d0,d1,d2;
 //
 // fractional (0-1) percentage of the shock front
 //
 shock=(float)ishock/100.0;
 //
 // position of the "floor"
 //
 floor=po[2]-df;
 if(type || ishock > 0)
 { // gavitational influences or shock front are in use
   //
   // determine the maximum dimensions of the object because under these
   // circumstances they effect how the action scales in time
   //
   dmax = zmax = -1.e20; dmin = 1.e30;
   for(i=0;i<Nvertices;i++)
   {
     di=E_Distance2(po,VertexList[i].p);
     if(di > dmax)dmax=di;
     if(di < dmin)dmin=di;
     if((float)VertexList[i].p[2] > zmax)zmax=(float)VertexList[i].p[2];
   }
   //
   // if the model is too small then using a shock front is impossible
   //
   if(dmax-dmin < 1.0)ishock=0;
   else
   {
     dmin=sqrt(dmin);
     dmax=shock/(sqrt(dmax)-dmin);
   }
   //
   // modify the distance to fall so that even the "topmost" facets in the
   // the model will fall to ground by the end of the effect.
   //
   if(type)df += (zmax - po[2]);
 }
 //
 // Get appropriate values for initial velocity "V" and gravitional
 // acceleration "A" that will allow the action to attain the maximum
 // height and distance fallen during the alloted time (unity).
 //
 GetVandA(df,dex,type,&V,&A);
 //
 // move all the facets into position, this assumes that each facet is
 // surrounded by three vertices and they are ordered sequentially in the
 // list of vertices for the model.
 //
 for(i=0;i<Nvertices;i+=3)
 {
   AddVectors(VertexList[i].p,VertexList[i+1].p,p);
   AddVectors(VertexList[i+2].p,p,p);
   //
   // position vector to center of facet
   //
   ScaleVector(0.333,p,p);
   //
   // vectors from center of facet to it's vertices
   //
   SubtractVectors(VertexList[i  ].p,p,d0);
   SubtractVectors(VertexList[i+1].p,p,d1);
   SubtractVectors(VertexList[i+2].p,p,d2);
   //
   // vector from site of exposion to center of facet
   //
   SubtractVectors(p,po,d);
   //
   // If a shock wave is in use calculate a pseudo "time" for the facet
   // so that "tt" will give the illusion of the effect being delayed
   // because the facet is at a distance |d| from the explosion.
   //
   if(ishock > 0)tt=GetPseudoTime(t,d,dmin,dmax);
   else          tt=t;
   Normalize(d);
   //
   // Calculate the new position "pp" for the center of the
   // facet at "pseudo" time "tt".  The vector "dd" is returned
   // and this is the direction along the trajectory followed by the
   // the facet.
   //
   GetPosition(p,d,po,df,dex,tt,V,A,floor,pp,type);
   if(!tumble)
   {
     //
     // if the facet is not tumbling then determine new position vectors
     // for it's vertices given the new position of the center "pp"
     //
     MoveIntoPosition(pp,d0,floor);
     MoveIntoPosition(pp,d1,floor);
     MoveIntoPosition(pp,d2,floor);
   }
   else
   {
     //
     // Calculate a "spin" transformation in the form of a "4 x 4" matrix
     //
     GetSpin(tumbletype,p,d,tt,spin);
     //
     // Move the facet into the position and orientation it occupies at
     // time "tt" using "spin" transformation and the other parameters
     // which are the same as for the "MoveIntoPosition" function.
     //
     SpinIntoPosition(pp,d0,floor,spin);
     SpinIntoPosition(pp,d1,floor,spin);
     SpinIntoPosition(pp,d2,floor,spin);
   }
   //
   // Apply the new positions to the vertex list.
   //
   CopyVector(d0,VertexList[i  ].p);
   CopyVector(d1,VertexList[i+1].p);
   CopyVector(d2,VertexList[i+2].p);
 }
 return;
}

static float GetPseudoTime(
 float t,
 VECTOR dv,
 float dmin,
 float dmax
){
 //
 // get the distance from the site of the explosion to the facet under
 // consideration, scale the time so that "t" is adjusted that facets
 // closest to the explosion "equal to dmin" fly off when "t=0".
 // "dmin" and "dmax" are the distances from the site of explosion to
 // the nearest and farthest facets in the model.
 //
 float d;
 d=sqrt(dv[0]*dv[0]+dv[1]*dv[1]+dv[2]*dv[2]);
 d=(d-dmin)*dmax;
 if(t < d)return 0.0;
 //
 // For facets further from the explosion - put the clock back so
 // they don't fly off until later!
 //
 return t=t-d;
}

static void GetSpin(
 int type,
 VECTOR p,
 VECTOR d,
 float t,
 float spin[4][4]
){
 //
 // Calculate a transformation matix that will simulate a spinning motion
 // when applied to the facets in a model. It is assumed that "spin" is
 // to apply to a coordinate frame of reference centered on each facet
 // it is applied to.
 //
 static VECTOR up={0.0,0.0,1.0},zero={0.0,0.0,0.0};
 VECTOR a;
 float angle;
 if(type == 0)CopyVector(up,a);
 else
 {
   //
   // Carrying out this additional step gives the illusion that the
   // facets are twisting and tumbling as they spin.
   //
   VectorCrossProduct(d,up,a);
   Normalize(a);
 }
 angle = t*2*3.1415926;
 //
 // The "spin" transformation is obtained by a rotation around an axis
 // which is proportional to the time elapsed.  A full rotation is
 // made by the time the effect has reached it's climax.
 //
 RotateRoundAxis(spin,zero,a,angle);
 return;
}

static void GetVandA(
 float df,
 float dex,
 long type,
 float *V,   // initial velocity
 float *A    // acceleration due to gravity
){
 //
 // Calculate appropriate values for the projectiles initial velociy
 // and acceleration due to gravity given the type of explosion,
 // how far the 'missiles' have to fall ("df") and how big the explosion is
 // to look "(dex)" (a value on the same scale as the distance fallen).
 // Note: If "dex" << "df" then the explosion will just look like a fall!
 //
 if(type == 0)
 {
   *V = 2.0*dex;
   *A = 0.0;
 }
 else{
   *A = sqrt(2.0*dex) + sqrt(2.0*(dex+df));
   *A = (*A) * (*A);
   *V = sqrt(2.0*(*A)*dex);
 }
 return;
}

static void GetPosition(
 VECTOR p,    // center of facet
 VECTOR d,    // direction face travelling in when ejected
 VECTOR po,   // site of explosion
 float df,    // distance to fall
 float dex,   // 'size' of the explosion
 float t,     // time we need to know where facet is
 float V,     // initial velocity (magnitude)
 float A,     // pseudo gravitational acceleration
 float floor, // distance from explosion to floor.
 VECTOR pp,   // returned (new position)
 long type
){
 //
 // Calculate the position to move the center of the facet to.
 // The function does this by determined in which bounce (it only examines
 // the first three) the facet lies at time t. It then proceed to
 // calculate the new position "pp".
 //
 if(type == 0)
 {
   //
   // Radial explosion only (no bounces) facets fly off and gradually
   // slow down.
   //
   ScaleVector(V*t*(1.0-0.5*t),d,pp);
   AddVectors(p,pp,pp);
 }
 else
 {
   float ca,sa,r,ti,c;
   VECTOR dxy;
   dxy[0]=d[0]; dxy[1]=d[1]; dxy[2]=0.0;
   Normalize(dxy);
   ca=VectorDotProduct(d,dxy);
   sa=sqrt(1.0 - ca*ca);
   if(p[2] < po[2])sa = -sa;
   c=(V*V*sa*sa - 2.0*A*(floor-p[2]));
   if(c >= 0)
   {
     if(c > 0)c=sqrt(c);
     ti = (V*sa + c)/A;
     if(t > ti)
     {
       //
       // not before first impact.
       //
       float vx,vy,dt;
       VECTOR pnew;
       r = V*ca*ti;
       pnew[0]=p[0]+r*dxy[0];
       pnew[1]=p[1]+r*dxy[1];
       pnew[2]=p[2]+V*sa*ti-0.5*A*ti*ti;
       vx=V*ca; vy=A*ti-V*sa;
       V=sqrt(vx*vx+vy*vy);
       ca=vx/V; sa=vy/V; V *= 0.5;
       t=t-ti;
       dt=2.0*V*sa/A;
       if(ti+dt > 1.0)
       {
         //
         // not before second impact  (assume in third bounce)
         //
         CopyVector(pnew,pp);
         pp[2]=max(pp[2],floor);
         return;
       }
       //
       // in second impact
       //
       if(t > dt)t=dt;
       r = V*ca*t;
       pp[0]=pnew[0]+r*dxy[0];
       pp[1]=pnew[1]+r*dxy[1];
       pp[2]=pnew[2]+V*sa*t-0.5*A*t*t;
       pp[2]=max(pp[2],floor);
       return;
     }
   }
   //
   // first impact of projectile
   //
   r = V*ca*t;
   pp[0]=p[0]+r*dxy[0];
   pp[1]=p[1]+r*dxy[1];
   pp[2]=p[2]+V*sa*t-0.5*A*t*t;
   pp[2]=max(pp[2],floor);
 }
 return;
}

static void MoveIntoPosition(
 VECTOR pp,
 VECTOR dp,
 float flr
){
 //
 //  move the vertex into position from "pp" to "dp" by
 //  adding offset "dp" (undergone by the center of the facet)
 //  and making sure that it doesn't fall through the floor.
 //
 AddVectors(pp,dp,dp);
 dp[2]=max(dp[2],flr);
 return;
}

static void SpinIntoPosition(
 VECTOR pp,
 VECTOR dp,
 float flr,
 float spin[4][4]
){
 //
 //  spin the vertex into position from "pp" to "dp" by
 //  applying the "spin" transformation then adding offset "dp"
 //  and making sure that it doesn't fall through the floor.
 //
 VECTOR dpp;
 TransformVector(spin,dp,dpp);
 AddVectors(pp,dpp,dp);
 dp[2]=max(dp[2],flr);
 return;
}
