//
// Some animation helper functions are presented in this code section
// The algorithms, formula and equations are discussed in the text.
//

//////////  The data types will need to be modified
//////////  to fit whatever framework you are using.

typedef float VECTOR[3];
typedef float MATRIX[4][4];
typedef float QUATERNION[4];


///////// The following are standard functions and definitions that are
///////// usually part of a compilers library.
///////// Substitute then with the ones you wish to use.

// return floating point absolute value of D
extern float fabs(float D);
// return floating point square root of  D
extern float sqrt(float D);
// return the sin of the angle T (radians)
extern float sin(float T);
// return the cos of the angle T (radians)
extern float cos(float T);
// return the arc cos of T
extern float acos(float T);
// return the inverse tangent of dy/dx in range (-pi, pi)
extern float atan2(float dy, float dx);

//////////////////////////////////////////////////////////////////////
//////////////// Functions ///////////////////////////////////////////


void MultiplyVectorByMatrix(MATRIX m,
           float x, float y, float z,
           float *xx, float *yy, float *zz)
{
 //
 // To use a three dimensional vector in "non-homogeneous" form the
 // multiplication by a 4x4 transformation matrix is implemented in this
 // function by refering to the elements of the vectors directly
 //
   *xx=m[0][0]*x+m[0][1]*y+m[0][2]*z+m[0][3];
   *yy=m[1][0]*x+m[1][1]*y+m[1][2]*z+m[1][3];
   *zz=m[2][0]*x+m[2][1]*y+m[2][2]*z+m[2][3];
 return;
}

void GetBasisTransform(VECTOR x, VECTOR y, VECTOR z,
                              VECTOR u, VECTOR v, VECTOR w,
                              MATRIX T)
{
 //
 // Given two orthonormal bases, the vectors  x,y,z  and  u,v,w  this
 // function determines "T" a transformation matrix which transforms any
 // vector expressed in terms of the  x,y,z basis to one expressed in
 // terms of the  u,v,w.
 // for example   vectors  u = Tx
 //                        v = Ty
 //                        w = Ty
 //
 int i,j;
 float a[3][3];
 MATRIX t;
 float det;
 for(i=0;i<4;i++)for(j=0;j<4;j++)
  if(i == j)T[i][j]=t[i][j]=1.0;
  else      T[i][j]=t[i][j]=0.0;
 a[0][0]=x[0]; a[0][1]=x[1]; a[0][2]=x[2];
 a[1][0]=y[0]; a[1][1]=y[1]; a[1][2]=y[2];
 a[2][0]=z[0]; a[2][1]=z[1]; a[2][2]=z[2];
 det=a[0][0]*(a[1][1]*a[2][2] - a[2][1]*a[1][2])
    -a[0][1]*(a[1][0]*a[2][2] - a[2][0]*a[1][2])
    +a[0][2]*(a[1][0]*a[2][1] - a[2][0]*a[1][1]);
 if(fabs(det) < 1.e-10)return;
 det = 1.0/det;
 t[0][0] = +(a[1][1]*a[2][2] - a[2][1]*a[1][2])*det;
 t[1][0] = -(a[1][0]*a[2][2] - a[2][0]*a[1][2])*det;
 t[2][0] = +(a[1][0]*a[2][1] - a[2][0]*a[1][1])*det;
 t[0][1] = -(a[0][1]*a[2][2] - a[2][1]*a[0][2])*det;
 t[1][1] = +(a[0][0]*a[2][2] - a[2][0]*a[0][2])*det;
 t[2][1] = -(a[0][0]*a[2][1] - a[2][0]*a[0][1])*det;
 t[0][2] = +(a[0][1]*a[1][2] - a[1][1]*a[0][2])*det;
 t[1][2] = -(a[0][0]*a[1][2] - a[1][0]*a[0][2])*det;
 t[2][2] = +(a[0][0]*a[1][1] - a[1][0]*a[0][1])*det;
 MultiplyVectorByMatrix(t,u[0],v[0],w[0],&T[0][0],&T[0][1],&T[0][2]);
 MultiplyVectorByMatrix(t,u[1],v[1],w[1],&T[1][0],&T[1][1],&T[1][2]);
 MultiplyVectorByMatrix(t,u[2],v[2],w[2],&T[2][0],&T[2][1],&T[2][2]);
}

static void InvertRotationMatrix(MATRIX T,MATRIX t)
{
 //
 // This function returns the inverse of a PURELY rotational transformation
 // matrix.  "t" is the inverse of  "T"    t = t**(-1)
 // e.g. if vectors "a" and "b" are related by  b = Ta then  a = tb
 //
 int i,j;
 float a[3][3];
 float det;
 for(i=0;i<4;i++)for(j=0;j<4;j++)
  if(i == j)t[i][j]=1.0;
  else      t[i][j]=0.0;
 for(i=0;i<3;i++)for(j=0;j<3;j++)a[i][j]=T[i][j];
 det=a[0][0]*(a[1][1]*a[2][2] - a[2][1]*a[1][2])
    -a[0][1]*(a[1][0]*a[2][2] - a[2][0]*a[1][2])
    +a[0][2]*(a[1][0]*a[2][1] - a[2][0]*a[1][1]);
 if(fabs(det) < 1.e-10)return;
 det = 1.0/det;
 t[0][0] = +(a[1][1]*a[2][2] - a[2][1]*a[1][2])*det;
 t[1][0] = -(a[1][0]*a[2][2] - a[2][0]*a[1][2])*det;
 t[2][0] = +(a[1][0]*a[2][1] - a[2][0]*a[1][1])*det;
 t[0][1] = -(a[0][1]*a[2][2] - a[2][1]*a[0][2])*det;
 t[1][1] = +(a[0][0]*a[2][2] - a[2][0]*a[0][2])*det;
 t[2][1] = -(a[0][0]*a[2][1] - a[2][0]*a[0][1])*det;
 t[0][2] = +(a[0][1]*a[1][2] - a[1][1]*a[0][2])*det;
 t[1][2] = -(a[0][0]*a[1][2] - a[1][0]*a[0][2])*det;
 t[2][2] = +(a[0][0]*a[1][1] - a[1][0]*a[0][1])*det;
}

#define EPSILON 1.0e-6

void MatrixToQuaternion(MATRIX m, QUATERNION q)
{
 //
 // Convert the transformation matrix "m" to the equivalent quaternion "q".
 // Note that ONLY the rotational component of the matrix can be written
 // in quaternion form. Any translational component in "m" is ignored.
 //
 float w4,x2;
 q[0]=0.25*(1.0+m[0][0]+m[1][1]+m[2][2]);
 if(q[0] > EPSILON){
   q[0]=sqrt(q[0]);
   w4=1.0/(4.0*q[0]);
   q[1]=(m[1][2]-m[2][1])*w4;
   q[2]=(m[2][0]-m[0][2])*w4;
   q[3]=(m[0][1]-m[1][0])*w4;
 }
 else{
   q[0]=0.0;
   q[1]= -0.5*(m[1][1]+m[2][2]);
   if(q[1] > EPSILON){
     q[1]=sqrt(q[1]);
     x2=1.0/(2.0*q[1]);
     q[2]=m[0][1]*x2;
     q[3]=m[0][2]*x2;
   }
   else{
     q[1]=0.0;
     q[2]=0.5*(1.0-m[2][2]);
     if(q[2] > EPSILON){
       q[2]=sqrt(q[2]);
       q[3]=m[1][2]/(2.0*q[2]);
     }
     else{
       q[2]=0.0;
       q[3]=1.0;
     }
   }
 }
 return;
}

void QuaternionToMatrix(QUATERNION q, MATRIX m)
{
 //
 // Convert the quaternion "q" to the equivalent transformation matrix "m".
 //
 float w,x,y,z,x2,y2,z2;
 w=q[0]; x=q[1]; y=q[2]; z=q[3]; x2=x*x; y2=y*y; z2=z*z;
 m[0][0]=1.0-2*y2-2*z2; m[0][1]=2*x*y+2*w*z;   m[0][2]=2*x*z-2*w*y;
 m[1][0]=2*x*y-2*w*z;   m[1][1]=1.0-2*x2-2*z2; m[1][2]=2*y*z+2*w*x;
 m[2][0]=2*x*z+2*w*y;   m[2][1]=2*y*z-2*w*x;   m[2][2]=1.0-2*x2-2*y2;
 m[3][3]=1.0; m[0][3]=m[1][3]=m[2][3]=m[3][0]=m[3][1]=m[3][2]=0.0;
 return;
}

void Slerp(float mu, QUATERNION q1, QUATERNION q2, QUATERNION q)
{
 //
 // Perform "sperical" interpolation between two quaternions given
 // the parameter mu. The output quaternion "q" is such that:
 // when  mu=0   q=q1
 // when  mu=1   q=q2
 //
 float sum,beta1,beta2,theta;
 sum=q1[0]*q2[0]+q1[1]*q2[1]+q1[2]*q2[2]+q1[3]*q2[3];
 if(sum > 1.0)sum=1.0;
 theta=acos(sum);
 if(theta < EPSILON){ beta1=1.0-mu; beta2=mu;}
 else{
  beta1=sin((1.0-mu)*theta)/sin(theta);
  beta2=sin(mu*theta)/sin(theta);
 }
 q[0]=beta1*q1[0]+beta2*q2[0];
 q[1]=beta1*q1[1]+beta2*q2[1];
 q[2]=beta1*q1[2]+beta2*q2[2];
 q[3]=beta1*q1[3]+beta2*q2[3];
}
