// Code to calculate whether a line between two points intersects
// a triangular polygon.

//////////  The data type (VECTOR) will need to be modified
//////////  to fit whatever framework you are using.

typedef float VECTOR[3];

////////// The following functions will need to be provided to fulfil the
////////// described action.

extern float  Dot(VECTOR A, VECTOR B);
//  Scale the vector A with the scalar s,   R = sA
extern void   ScaleVector(float s,VECTOR A, VECTOR R);
// add two vectors R = A + B
extern void AddVectors(VECTOR A, VECTOR B, VECTOR R);
// subtract two vectors R = A - B
extern void SubtractVectors(VECTOR A, VECTOR B, VECTOR R);

///////// The following are standard functions that are usually part
///////// of a compilers library.
///////// Substitute then with the ones you wish to use.

extern float fabs(float D);       // return floating point absolute value of D
extern float sqrt(float D);       // return floating point square root of  D

// The definination of factors needed to determine if the intersection
// is near the edge of the polygon and if the line passes through the
// polygon.


#define TOL1    0.0005
#define TOL2    0.9995
#define TOL5   -0.001
#define TOL6    1.001

#define TOL8    1.00000
#define TOL9  100.0
#define TOLA  100.0
#define TOLC  100.0

long PointInsidePolygon(VECTOR vi,
                        VECTOR p0, VECTOR p1, VECTOR p2,
                        long *code, long *id)
{
// vi is position vector of point to test if it lies inside triangular
// polygon formed by p0 - p1 - p2  (vi lies in the plane of the polygon)
/* code  = 0 point vi is in plane and inside face                */
/*       = 1   "      "      "        on edge id id's the edge   */
/*       = 2   "      "      "        at vertex id id's the edge */
// "id" returns which of the edges 0 or 1 or 2  the intersection hits
// if the intersection is on an edge
 VECTOR w,u,v;
 float a,b,ab,den,uu,vv,uv,wu,wv;
 SubtractVectors(vi,p0,w);
 SubtractVectors(p2,p0,u);
 SubtractVectors(p1,p0,v);
 uu=Dot(u,u);
 vv=Dot(v,v);
 uv=Dot(u,v);
 wu=Dot(w,u);
 wv=Dot(w,v);
 den=uu*vv-uv*uv;
 if(fabs(den) < 1.0e-6)return 0; // denominator zero
 a=(wu*vv - wv*uv)/den;
 b=(wv*uu - wu*uv)/den;
/* a=1 at p2      b=1 at p1  */
 ab=a+b;
 if(a < TOL5 || a > TOL6 || b < TOL5 || b > TOL6 || ab > TOL6)return 0;
 if(a >= TOL1 && a <= TOL2 && b >= TOL1 && b <= TOL2 && ab <= TOL2){
   *code=0;  /* inside face */
   return 1;
 }
 else if(a < TOL1){ /* along edge 0-1 */
   if(b < TOL1){ /* p0 */
     *code=2; *id=0;
     return 1;
   }
   else if(b > TOL2){ /* p1 */
     *code=2; *id=1;
     return 1;
   }
   else{  /* on edge 0-1 */
     *code=1; *id=0;
     return 1;
   }
 }
 else if(b < TOL1){ /* along edge 0-2 */
   if(a < TOL1){ /* p0 */
     *code=2; *id=0;
     return 1;
   }
   else if(a > TOL2){ /* p2 */
     *code=2; *id=2;
     return 1;
   }
   else{ /* on edge 0-2 */
     *code=1; *id=2;
     return 1;
   }
 }
 else if(ab > TOL2){/* on edge 1-2 */
   *code=1; *id=1;
   return 1;
 }
 return 0;
}

long LineThroughPolygon(VECTOR pa, VECTOR pb, VECTOR pc,
                        VECTOR n,
                        VECTOR p0, VECTOR p1,
                        VECTOR p2,
                        long full)
{
// what type of check
/* full = 0 implies that the edge intersects the tool face internally.
            At least one vertex of the edge does not lie in the plane
            of the tool face face.
   full = 1 implies that the edge crosses the tool face but it may cross
            the face at one of it's edges, however neither of the vertices
            of the edge lie in/near the plane of the tool face.
   full = 2 implies that if both points lie in plane of tool face then
            it is checked to see if it is a bounding edge.
   pa, pb and pc are the vertices of the polygon
   p0 and p1 are points at end of edge
   returns p2 point of intersection
   return code = 0 no intersection
               = 1 intersection at one end of edge
               = 2 if inside
*/
 long code,id;
 VECTOR da,db,dba;
 float d,d0,d1,d01,mu,mu1;
 SubtractVectors(pa,p0,da);
 SubtractVectors(pa,p1,db);
 d0=Dot(da,n);
 d1=Dot(db,n);
 d01 =d0*d1;
 if(fabs(d0) > TOLA || fabs(d1) > TOLA){         /* TOLA => not in plane */
   /* at least one vertex lies out of the tool plane */
   if(d01 <= TOL8){          /* TOL8 => line crosses the polygon */
     SubtractVectors(p1,p0,dba);
     mu1=Dot(dba,n);
     if(fabs(mu1) > TOL9){
       mu=d0/mu1;
       if(full > 0){   /* not near either end of edge */
         if(fabs(d0) < TOLA || fabs(d1) < TOLA)return 0;
       }
       ScaleVector(mu,dba,dba);
       AddVectors(p0,dba,p2);   /* p2 is the intersection point */
       if(PointInsidePolygon(p2,pa,pb,pc,&code,&id)){
         if(code == 0 || full > 0)return 1; /* allow Xn even on face edges */
       }
     }
   }
 }
 else if(full == 2){ /* both points lie in plane of tool face */
   /* check to see if this edge is a boundary edge   */
   AddVectors(p0,p1,dba);
   ScaleVector(0.5,dba,dba);        /* midpoint of edge  */
   SubtractVectors(pa,dba,da);
   if(fabs(Dot(da,n)) < TOLC){  /*    "  is in plane */
     if(PointInsidePolygon(dba,pa,pb,pc,&code,&id))return 2; // inside polygon
   }
 }
 return 0;
}
