//
// Algorithm to fill an outline with triangular facets
//

//////////  The data types (VECTOR and VERTEX) will need to be modified
//////////  to fit whatever framework you are using.

// The following parameters may need to be adjusted for different scales
// of outline. They are currently set for the case in which the minimum
// lenght of an edge in the outline is 10.0. The parameter MIMIMUM_LENGTH
// should be set so that it is a bit greater than MIMIMUM_SCALE**2.
// MINIMUM_SCALE should be set to approx 20% of the the shortest lenght.
// resolution. It would of course be possible to modify the code to
// calculate the smallest length and set these accordingly. Doing this
// would have the added advantage that if any edge in the outline was
// found to have zero length an error could be flagged and the capping
// abandoned.

#define MINIMUM_LENGTH 5.0
#define MINIMUM_SCALE  2.0

typedef float VECTOR[3];

typedef struct tagVERTEX {
 // add other members as required
 VECTOR p;
} VERTEX;

////////// The following functions will need to be provided to fulfil the
////////// described action.

// create an edge in database returns the identity of new edge
extern long   newE(VERTEX *, VERTEX *);
// create a facet in database  returns the identity of new edge
extern long   newF(VERTEX *, VERTEX *, VERTEX *);
// make vector V unit length
extern void   Normalize(VECTOR V);
// R = vector cross product of U x V
extern void   Cross(VECTOR U, VECTOR V, VECTOR R);
// return the dot product of A.B
extern float  Dot(VECTOR A, VECTOR B);
//  Scale the vector A with the scalar s,   R = sA
extern void   ScaleVector(float s,VECTOR A, VECTOR R);
// copy a vector from Source to Destination
extern void   CopyVector(VECTOR Source, VECTOR Destination);
// add two vectors R = A + B
extern void AddVectors(VECTOR A, VECTOR B, VECTOR R);
// subtract two vectors R = A - B
extern void SubtractVectors(VECTOR A, VECTOR B, VECTOR R);

///////// The following are standard functions that are usually part
///////// of a compilers library.
///////// Substitute then with the ones you wish to use.

extern void *malloc(long X);      // allocate X bytes, return pointer to it.
extern void free(void *P);        // release memory block pointed to by P
extern float fabs(float D);       // return floating point absolute value of D
extern float sqrt(float D);       // return floating point square root of  D

/////////////////////// The following structures are used by the function
/////////////////////// and will be used by caller to set up the data.

// Each element in the list describing a part of the outline of the
// curve to be capped is described in a "LIST" structure.

typedef struct tagLIST {
  struct tagVERTEX* v; // pointer to the vertex associated with this edge
                       // of a piece of the outline.
  long   last,         // last entry in the list that is currently used
         next,         // next entry in the list that is currently used
         inc;          // this edge is included in current piece.
  VECTOR n;    // A vector perpendicular to this edge. It should also be
               // perpendicular to a vector that is normal to the plane in
               // which this piece of the outline lies.
  VECTOR p;    // Vector giving position of vertex v (can be copy of v->p)
} LIST;

//
// The curve may consist of several pieces. Each piece is described by
// a "LISTOFLISTS" data structure.  The full curve will be given by
// an array of "LISTOFLISTS" structures.
//

typedef struct tagLISTOFLISTS {
  struct tagLIST* List; // Array of structures with information about a vertex/
                        // edge pair. It is assumed that the vertex is at the
                        // start of the edge which then joins it
                        // to the vertex in the next entry in the List.
  long   n,             // Number of edges in this piece of the outline
         I,             // inside / outside flag
         done;          // Finished processing this piece
  VECTOR N;             // A vector normal to the plane containing this piece.
                        // The normal can point to either side of the plane
                        // it will be modified for consistency with the other
                        // N's and whether the curve described in "List" is
                        // clockwise or anti-clockwise.
} LISTOFLISTS;

////
//// functions uses internally by this algorithm
////

static void  ReverseList(LIST *, long );
static long  IsEdgeInside(LIST *, long, long , long , long , VECTOR, VECTOR);
static void  UpdateOutlinePiece(long, long, LISTOFLISTS *);
static void  UpdateOutlineNormals(long, LISTOFLISTS *);
static long  IsEdgeNotInWay(long ,LIST *, VECTOR, VECTOR , long, long);
static long  IsCurveInside(LISTOFLISTS *fl, long i, long );

// FillPolygon is the entry function of the algorithm.

void FillPolygon(long N, LISTOFLISTS *ListOfLists)
//
// The function to fill a polygonal outline in N closed pieces. The pieces
// are given in an array of LISTOFLISTS structures that are filled before
// calling to function FillPolygon.
//
{
 VERTEX* LastFace[3];
 long i,j,k,n,nTotal,h_count;
 LISTOFLISTS *newListOfLists;
 LIST *List;
 VECTOR v1,v2,v3,PieceNormal,EdgeNormal;
 long te,le,ne,fe;

 for(i=0;i<N;i++)
 {
   // is curve inside another curve
   // function returns -1 if not or index of curve it is inside
   ListOfLists[i].I=IsCurveInside(ListOfLists,i,N);
   ListOfLists[i].done=0;
 }

 for(i=0;i<N;i++)
 { // is it a hole or inside a hole ?
   h_count=0;
   // if inside work outwards to see how may curves it lies inside
   if((j=ListOfLists[i].I) >= 0)while(j >= 0){
     h_count++;
     j=ListOfLists[j].I;
   }
   if(h_count %2 == 0)ListOfLists[i].I = -1; // curve inside a hole
 }
 // make normals consistent for outline curves
 for(i=0;i<N;i++)
 {
   if(ListOfLists[i].I < 0)  //  not a hole
   {
     n=ListOfLists[i].n;
     List=ListOfLists[i].List;
     for(j=0;j<n;j++)
     { // we need an edge that is inside
       te=j; ne=List[te].next; fe=List[ne].next;
       if(IsEdgeInside(List,n,te,ne,fe,v1,ListOfLists[i].N))goto N1;
     }
     // failed to get a normal
     goto FINISHED;
     N1:
     CopyVector(List[te].p,v1);
     CopyVector(List[ne].p,v2);
     CopyVector(List[fe].p,v3);
     SubtractVectors(v2,v1,v2);
     SubtractVectors(v3,v1,v3);
     Cross(v2,v3,v1);
     Normalize(v1);
     CopyVector(v1,ListOfLists[i].N);
     // update the normals for each edge to reflect this new curve normal
     UpdateOutlineNormals(i,ListOfLists);
   }
 }
 // make normals consistent for hole curves
 for(i=0;i<N;i++)
 {
   if(ListOfLists[i].I >= 0)  // a hole outline
   {
     n=ListOfLists[i].n;
     List=ListOfLists[i].List;
     for(j=0;j<n;j++)
     { // need an inside edge
       te=j; ne=List[te].next; fe=List[ne].next;
       if(IsEdgeInside(List,n,te,ne,fe,v1,ListOfLists[i].N))goto N2;
     }
     goto FINISHED;
     N2:
     CopyVector(List[te].p,v1);
     CopyVector(List[ne].p,v2);
     CopyVector(List[fe].p,v3);
     SubtractVectors(v2,v1,v2);
     SubtractVectors(v3,v1,v3);
     Cross(v2,v3,v1);
     Normalize(v1);
     // if the hole N is same direction as the outer N reverse hole
     if(Dot(v1,ListOfLists[ListOfLists[i].I].N) > 0.0){
       ReverseList(List,n);
       ScaleVector( -1.0,v1,ListOfLists[i].N);
     }
     else  CopyVector(v1,ListOfLists[i].N);
     // update the normals for this new hole curve
     UpdateOutlineNormals(i,ListOfLists);
   }
 }
//
// We must have some iteration here in case hole inside
// a piece cannot be attached due to another hole in the way, in
// such a case the intervening hole must be attached first
//
 for(k=0;k<2;k++)
 { // 2 iterations are probably enough
   for(i=0;i<N;i++)
   {
     if(ListOfLists[i].I >= 0 && ListOfLists[i].done == 0)
        UpdateOutlinePiece(i,N,ListOfLists);
   }
 }
//
// now fill the curves  - this is the main part of the algorithm
//
 for(k=0;k<N;k++)
 {
   n=ListOfLists[k].n;
//
// "n" will have been set to 0 for holes when they were attached to
// an outline curve.
//
   if(n >= 4)  // at least 4 points required
   {
     List=ListOfLists[k].List;
     CopyVector(ListOfLists[k].N,PieceNormal);
     nTotal=n;
     te=0;
     while(n > 3)
     {
       ne=List[te].next; fe=List[ne].next;
       if(IsEdgeInside(List,nTotal,te,ne,fe,EdgeNormal,PieceNormal)){
         // add the edge and face to chip a bit off the polygon
         newE(List[te].v,List[fe].v);
         newF(List[te].v,List[ne].v,List[fe].v);
         List[ne].inc=0;
         List[te].next=fe;
         List[fe].last=te;
         CopyVector(EdgeNormal,List[te].n);
         n--;
         te=fe;
       }
       else{ // move to next edge to try again
         te=ne;
       }
     }
     for(j=0,i=0;i<n;i++)    // Make the last face. Only 3 edges will be left.
     {                       // Get it. They are the ones not excluded.
       if(List[i].inc == 1){
         LastFace[j++]=List[i].v;
       }
     }
     newF(LastFace[0],LastFace[1],LastFace[2]);
   }
 }
 FINISHED:
 //
 //  This function does no error checking  - nor does it release
 //  resources for the ListOfLists or its data members.
 //
 return;
}

static void UpdateOutlineNormals(long ID, LISTOFLISTS *ListOfLists)
//
// make a normal for each edge of outline piece ID by taking cross product
// with edge direction and  normal to piece ID.
//
{
 long j,k,n;
 LIST *List;
 VECTOR v1,v2,v3;
 n=ListOfLists[ID].n;
 List=ListOfLists[ID].List;
 //
 // For every edge of piece ID.
 //
 for(j=0;j<n;j++)
 {
   CopyVector(List[j].p,v1);
   k=List[j].next;
   CopyVector(List[k].p,v2);
   SubtractVectors(v2,v1,v3);
   Normalize(v3);
   Cross(v3,ListOfLists[ID].N,List[j].n);
   Normalize(List[j].n);
 }
}

static void  ReverseList(LIST *List, long n)
//
// Reverse the list "List" of "n" items so that we have a consistent
// direction round the piece of the outline.
// Copy last to first  etc. etc.
//
{
  long i;
  VERTEX *temp;
  VECTOR vtemp;
  for(i=0;i<n/2;i++)
  {
     temp=List[i].v; List[i].v = List[n-1-i].v; List[n-1-i].v=temp;
     CopyVector(List[i].p,vtemp);
     CopyVector(List[n-1-i].p,List[i].p);
     CopyVector(vtemp,List[n-1-i].p);
  }
}

static long IsEdgeInside(LIST *List, long n,
                              long i0, long i1, long i2,
                              VECTOR new_normal, VECTOR N)
//
// Check whether an edge i0 to i2 is inside the outline piece recorded
// in List().
// Return the normal to the new edge in new_normal.
// N is the normal for the plane in which the outline piece "List()" lies.
//
{
 VECTOR p0,p1,pa,pb,p10,pba,p_centre;
 VECTOR pte,pne,pac,edge_normal,pi,da,db,p01,vilast;
 VECTOR vj,pn,pnb;
 long i,j,k,im1,number_of_crosses=0,excluded_id;
 float mu,mu1,denominator,d0,d1,d01;
 //
 // get normals for edges i0 -> i1  and  i0 -> i2 so that
 // check can be made to see if the new edge is internal
 //
 CopyVector(List[i0].p,pa);
 CopyVector(List[i1].p,pb);
 SubtractVectors(pb,pa,pte);
 Normalize(pte);
 CopyVector(List[i1].p,pa);
 CopyVector(List[i2].p,pb);
 SubtractVectors(pb,pa,pne);
 Normalize(pne);
 if((d0=Dot(pte,pne)) > 0.9999)
 {
 //
 // New edge (i0 - i2) is co-linear with i0 - i1
 //
   return 0;
 }
 CopyVector(List[i0].p,p0);
 CopyVector(List[i2].p,p1);
 SubtractVectors(p1,p0,p10);
 CopyVector(p10,p01);
 Normalize(p01);
 Cross(p01,N,new_normal);   // normal for edge between i0 - i2
 Normalize(new_normal);
 ScaleVector(0.5,p10,p_centre);
 AddVectors(p0,p_centre,p_centre);      //  point half way along i0 - i2
 //
 // flag to prevent testing edges that would give false count
 //
 excluded_id = -1;
 //
 // check to see if mid point is internal
 //
 for(i=0;i<n;i++)
 {
   if(List[i].inc == 1)  // check if edge i blocks line from  "p_centre" in
   {                     // direction "new_normal"
     CopyVector(List[i].p,pa);            // Position of vertex at one side
     j=List[i].next;                      // of the fencing plane.
     CopyVector(List[j].p,pb);            // Position of vertex at other.
     CopyVector(List[i].n,edge_normal);
     SubtractVectors(pa,p_centre,pac);
     mu1=Dot(edge_normal,new_normal);
     if(fabs(mu1) > 1.0e-4) // it will intesect
     {
       mu=Dot(pac,edge_normal)/mu1;
       if(mu > 0.0){                    // look for intersections on one side
         ScaleVector(mu,new_normal,pi); // intersection to plane of test edge
         AddVectors(p_centre,pi,pi);    // point of intersection on plane
         SubtractVectors(pb,pa,pba);
         k = -1; denominator = 1.0e-6;  // to prevent divide by zero
         if     (fabs(pba[0]) > denominator){k=0; denominator =fabs(pba[0]);}
         else if(fabs(pba[1]) > denominator){k=1; denominator =fabs(pba[1]);}
         else if(fabs(pba[2]) > denominator){k=2; denominator =fabs(pba[2]);}
         if(k >= 0){                   // "mu1" is fractional distance of
           mu1=(pi[k]-pa[k])/pba[k];   // interection between L and P (pi)
           if(fabs(mu1) < 1.0e-7)      // along pa -> pb
           {                           // at start of edge
             k=List[i].last;
             CopyVector(List[k].p,pn);
             SubtractVectors(pa,pn,pnb); // vector leading up to edge i0
             d0=Dot(pba,p01); d1=Dot(pnb,p01);
             if(d0*d1 > 0.0 && k != excluded_id){
               number_of_crosses++;
             }
           }
           else if(fabs(1.0-mu1) < 1.0e-7) // at end of edge
           {
             k=List[j].next;
             CopyVector(List[k].p,pn);
             SubtractVectors(pn,pb,pnb); // vector leading away from i1
             d0=Dot(pba,p01); d1=Dot(pnb,p01);
             if(d0*d1 > 0.0){
               number_of_crosses++; excluded_id=i;
               // excluded_id is used so that this edge is not checked
               // when processing the start of the next edge
             }
           }
           else if(mu1 > 0.0 && mu1 < 1.0){ // in centre of edge
             number_of_crosses++;
           }
         }
       }
     }
   }
 }
 if(number_of_crosses%2 == 0)return 0; // external to curve
 //
 //check for another edge crossing the potential internal edge i0 - i2
 //
 im1 = List[i0].last;  // edge prior to i0
 for(i=0;i<n;i++){
   if( !( i == im1 || i == i0 || i == i1 || i == i2)  && List[i].inc == 1){
     CopyVector(List[i].p,pa);
     j=List[i].next;
     CopyVector(List[j].p,pb);
     CopyVector(List[i].n,edge_normal);
     SubtractVectors(pa,p0,da);  d0=Dot(da,edge_normal);
     SubtractVectors(pa,p1,db);  d1=Dot(db,edge_normal);
     d01=d0*d1;
     if(d01 < 0.0){ // new edge crosses other edge plane, is it inside ?
       mu1=Dot(p10,edge_normal);
       if(fabs(mu1) > 1.0){
         mu=d0/mu1;
         ScaleVector(mu,p10,pi);
         AddVectors(p0,pi,pi);
         SubtractVectors(pb,pa,pba);
         k = -1; denominator = 1.0e-6;         // to prevent divide by zero
         if     (fabs(pba[0]) > denominator){k=0; denominator =fabs(pba[0]);}
         else if(fabs(pba[1]) > denominator){k=1; denominator =fabs(pba[1]);}
         else if(fabs(pba[2]) > denominator){k=2; denominator =fabs(pba[2]);}
         if(k >= 0){
           mu1=(pi[k]-pa[k])/pba[k];
           if(mu1 >= -0.0001 && mu1 < 1.0001){
             //
             // edge blocks potential edge
             //
             return 0;
           }
         }
       }
     }
   }
 }
 return 1;
}

static long IsCurveInside(LISTOFLISTS *ListOfLists,
                          long list_id, long n_lists)
//
// Determine whether the outline piece ListOfLists(list_id) lies inside
// any other outline piece. If it does return the identity of the other
// piece that is closest to it at the start of piece  list_id.
// Otherwise return -1
//
// Basically these tests are similar in many respects to those
// performed by function IsEdgeInside
//
{
 VECTOR p_centre,p0,p1,p10,p01,new_normal,pa,pb,old_normal,pac,pn,pnb,pba,pi;
 long i,j,k,l,number_of_crosses,n,identity = -1;
 float mu,mu1,denominator,mumin,mumumin=1.0e35;
 LIST *List;
 List=ListOfLists[list_id].List;
 CopyVector(List[0].p,p0);
 CopyVector(List[1].p,p1);
 SubtractVectors(p1,p0,p10);
 CopyVector(p10,p01);
 Normalize(p01);
 Cross(p01,ListOfLists[list_id].N,new_normal);
 Normalize(new_normal);
 ScaleVector(0.5,p10,p_centre);
 AddVectors(p0,p_centre,p_centre);            // mid point of first edge
 for(l=0;l<n_lists;l++)
 {
   if(l == list_id)continue;
   number_of_crosses=0;
   List=ListOfLists[l].List;
   n=ListOfLists[l].n;
   mumin=1.0e35;
   for(i=0;i<n;i++)
   { /* check to see if new_normal crosses edges in path */
     CopyVector(List[i].p,pa);
     j=List[i].next;
     CopyVector(List[j].p,pb);
     CopyVector(List[i].n,old_normal);
     SubtractVectors(pa,p_centre,pac);
     mu1=Dot(old_normal,new_normal);
     if(fabs(mu1) > 0.0001)
     {
       mu=Dot(pac,old_normal)/mu1;
       if(mu > 0.0000)
       {   // look for intersections on one side
         ScaleVector(mu,new_normal,pi);
         AddVectors(p_centre,pi,pi);    // intersection to plane of test edge
         SubtractVectors(pb,pa,pba);
         k = -1; denominator = 1.0e-6;
         if     (fabs(pba[0]) > denominator){k=0; denominator =fabs(pba[0]);}
         else if(fabs(pba[1]) > denominator){k=1; denominator =fabs(pba[1]);}
         else if(fabs(pba[2]) > denominator){k=2; denominator =fabs(pba[2]);}
         if(k >= 0){
           mu1=(pi[k]-pa[k])/pba[k];
           if(fabs(mu1) < 1.0e-7)
           {                        // at start of segment
             k=List[i].last;
             CopyVector(List[k].p,pn);
             SubtractVectors(pa,pn,pnb);
             if(Dot(pba,p01)*Dot(pnb,p01) > 0.0)number_of_crosses++;
           }
           else if(fabs(1.0-mu1) < 1.e-7)
           {                                 // at end of segment
             k=List[j].next;
             CopyVector(List[k].p,pn);
             SubtractVectors(pn,pb,pnb);
             if(Dot(pba,p01)*Dot(pnb,p01) > 0.0)number_of_crosses++;
           }
           else if(mu1 > 0.0 && mu1 < 1.0)
           {
             number_of_crosses++;
           }
           if(mu < mumin)mumin=mu;
         }
       }
     }
   }
   if(number_of_crosses%2 == 1 && mumin <= mumumin){identity=l; mumumin=mumin;}
 }
 return identity;
}

static long IsEdgeNotInWay(long n, LIST *List, VECTOR p0, VECTOR p1,
                        long e1, long e2)
//
// check whether the vector po->p1 crosses any edges in the List
// with the exception of edges e1 and e2
// returns 1 if NO edges are in the way otherwise returns 0
//
{
 VECTOR N,pa,pb,pba,pi,da,db,p10;
 float d0,d1,d01,mu,mu1,denominator;
 long i,j,k;
 SubtractVectors(p1,p0,p10);
 for(i=0;i<n;i++)
 {
   if(i != e1 && i != e2)
   {
     CopyVector(List[i].p,pa);   // First test if "p0" and "p1" are
     j=List[i].next;             // on the same side of the plane.
     CopyVector(List[j].p,pb);
     CopyVector(List[i].n,N);
     SubtractVectors(pa,p0,da);  // Form vectors from point in plane
     d0=Dot(da,N);               // to p0 and p1. Then if the dot product
     SubtractVectors(pa,p1,db);  // is > 0 they both lie on the same side.
     d1=Dot(db,N);
     d01=d0*d1;
     if(d01 < 0.0)
     { // edge joining p0 -> p1  crosses edge plane, is it with boundary ?
       mu1=Dot(p10,N);
       if(fabs(mu1) > 1.0)
       {
         mu=d0/mu1;
         ScaleVector(mu,p10,pi);
         AddVectors(p0,pi,pi);
         SubtractVectors(pb,pa,pba);
         k = -1; denominator = 1.0e-6;
         if     (fabs(pba[0]) > denominator){k=0; denominator = fabs(pba[0]);}
         else if(fabs(pba[1]) > denominator){k=1; denominator = fabs(pba[1]);}
         else if(fabs(pba[2]) > denominator){k=2; denominator = fabs(pba[2]);}
         if(k >= 0){
           mu1=(pi[k]-pa[k])/pba[k];
           if(mu1 >= -1.0e-4 && mu1 < 1.0001)return 0;
         }
       }
     }
   }
 }
 return 1;
}

static void UpdateOutlinePiece(long attach_id, long n,
                               LISTOFLISTS *ListOfLists)
//
// Attach piece "attach_id" which is a hole to the best bounding curve.
//
// First find a point of attachment for the hole to the outline piece
// this is the edge of closest approach that does not cross the hole or
// the outline of any hole that is a hole of the outline piece
// Input:
// attach_id  = identity of hole to be included
// n          = number of outline pieces in array  ListOfLists()
//
{
 long i,j,k,
      np_hole,
      np_outline,
      new_n,
      container_id,
      hole_attach_point = -1,
      outline_attach_point = -1,
      je1,je2,
      j0,j1,j2,j3;
 LIST *List_h,*List_o,*List_new;
 VECTOR ph,po,delta_ho;
 float length,distance,smallest_distance=1.e35;
 //
 np_hole=ListOfLists[attach_id].n;      // number of edges in hole curve
 List_h=ListOfLists[attach_id].List;    // list of hole edges
 container_id=ListOfLists[attach_id].I; // identity of piece bounding hole
 np_outline=ListOfLists[container_id].n;// number of edges in bounding curve
 List_o=ListOfLists[container_id].List; // list of edges in bounding curve
 for(i=0;i<np_hole;i++)
 {                                      // check if any edges get in the way
    for(j=0;j<np_outline;j++)           // compare all pairs
    {
       CopyVector(List_h[i].p,ph);
       CopyVector(List_o[j].p,po);
       //
       // Would an edge linking vertex "i" (on outline) to "j" (on hole)
       // intersect any edge in outline or hole.
       //
       if(IsEdgeNotInWay(np_hole,List_h,ph,po,i,List_h[i].last) &&
          IsEdgeNotInWay(np_outline,List_o,ph,po,j,List_o[j].last) ){
         //
         // check the other holes that lie inside "container_id"
         //
         for(k=0;k<n;k++) // loop all curves
         {
           if(k != attach_id && ListOfLists[k].I == container_id)
           {              // this is inside
             long np_temp;
             LIST *List_temp;
             np_temp=ListOfLists[k].n;
             List_temp=ListOfLists[k].List;
             if( ! IsEdgeNotInWay(np_temp,List_temp,ph,po,-1,-1))goto NOGO;
           }
         }
         distance=(po[0]-ph[0])*(po[0]-ph[0]) +  // find closest
                  (po[1]-ph[1])*(po[1]-ph[1]) +
                  (po[2]-ph[2])*(po[2]-ph[2]);
         if(distance < smallest_distance)
         {
           smallest_distance=distance;
           // record index of vertices in the hole and outline Lists
           hole_attach_point=i, outline_attach_point=j;
         }
         NOGO:;
       }
    }
 }
 //
 //  "outline_attach_point" and hole_attach_point"
 //   identify indices in boundary and hole lists
 //   of entries conecting the hole to the outline.
 //
 if(hole_attach_point == -1 || outline_attach_point == -1)
   return;                      // defer or can't do attachment
 ListOfLists[attach_id].done=1; // this hole has been processed
 new_n=np_hole+np_outline+2;    // number of edges in the replacement outline piece
 //
 // Make the new curve to replace the outline + hole. Do it in 6 bits
 // 1) up to join in outline
 // 2) the join
 // 3) to end of hole
 // 4) from start of hole to join
 // 5) join
 // 6) form join to end of outline
 //
 if((List_new=(LIST *)malloc(new_n*sizeof(LIST))) == 0)return;
 //
 // Start index in the new array
 //
 j=0;
 //
 // bit (1)
 if(outline_attach_point > 0)for(i=0;i<outline_attach_point;i++)
 {
   List_new[j].v=List_o[i].v;
   CopyVector(List_o[i].n,List_new[j].n);
   CopyVector(List_o[i].p,List_new[j].p);
   j++;
 }
 //
 // bit (2)
 List_new[j].v=List_o[outline_attach_point].v;
 je1=j; // keep a record of its index
 CopyVector(List_o[outline_attach_point].p,List_new[j].p);
 j++;
 //
 // bit(3)
 if(hole_attach_point < np_hole)for(i=hole_attach_point;i<np_hole;i++)
 {
   List_new[j].v=List_h[i].v;
   CopyVector(List_h[i].n,List_new[j].n);
   CopyVector(List_h[i].p,List_new[j].p);
   j++;
 }
 //
 // bit(4)
 if(hole_attach_point > 0)for(i=0;i<hole_attach_point;i++)
 {
   List_new[j].v=List_h[i].v;
   CopyVector(List_h[i].n,List_new[j].n);
   CopyVector(List_h[i].p,List_new[j].p);
   j++;
 }
 //
 // bit(5)
 List_new[j].v=List_h[hole_attach_point].v;
 je2=j;
 CopyVector(List_h[hole_attach_point].p,List_new[j].p);
 j++;
 //
 // bit(6)
 if(outline_attach_point < np_outline)
   for(i=outline_attach_point;i<np_outline;i++)
 {
   List_new[j].v=List_o[i].v;
   CopyVector(List_o[i].n,List_new[j].n);
   CopyVector(List_o[i].p,List_new[j].p);
   j++;
 }
 //
 // now update the conectivity information - next / last and included
 //
 for(j=0;j<new_n;j++)
 {
   List_new[j].inc=1;
   if     (j == 0)       {List_new[j].last=new_n-1;  List_new[j].next=1;  }
   else if(j == new_n-1) {List_new[j].last=j-1;      List_new[j].next=0;  }
   else                  {List_new[j].last=j-1;      List_new[j].next=j+1;}
 }
//
// Separate the curve at the join by
// adding a small displacement that pulls away the edges attached to
// the vertices at both ends of the edge linking hole to its outline.
//
 for(j1=0;j1<new_n;j1++)
 {
   //
   // only apply to edges adjacent to edges joining hole
   //
   if(j1 == je1 || j1== je2)
   {
     j0=List_new[j1].last;
     j2=List_new[j1].next;
     j3=List_new[j2].next;
     SubtractVectors(List_new[j1].p,
                     List_new[j0].p,
                     delta_ho); // up to edge j
     //
     // unless the edge is longer than a mimimum size don't pull it away
     //
     length=(delta_ho[0]*delta_ho[0])
           +(delta_ho[1]*delta_ho[1])
           +(delta_ho[2]*delta_ho[2]);
     if(length > MINIMUM_LENGTH){
       ScaleVector((1.0 - MINIMUM_SCALE/sqrt(length)),delta_ho,delta_ho);
       AddVectors(List_new[j0].p,delta_ho,List_new[j1].p);
     }
     SubtractVectors(List_new[j3].p,List_new[j2].p,delta_ho);
     length=(delta_ho[0]*delta_ho[0])
           +(delta_ho[1]*delta_ho[1])
           +(delta_ho[2]*delta_ho[2]);
     if(length > MINIMUM_LENGTH)
     {
       ScaleVector((1.0 - MINIMUM_SCALE/sqrt(length)),delta_ho,delta_ho);
       SubtractVectors(List_new[j3].p,delta_ho,List_new[j2].p);
     }
     //
     // Make the normals for the additional edges.
     //
     SubtractVectors(List_new[j2].p,List_new[j1].p,delta_ho);
     Cross(delta_ho,ListOfLists[container_id].N,List_new[j1].n);
     Normalize(List_new[j1].n);
   }
 }
 //
 // make an edge to join the hole to its bounding curve
 //
 newE(List_h[hole_attach_point].v,List_o[outline_attach_point].v);
 //
 // free up the old memory and put new pointers into the ListOfLists
 //
 free(List_o);
 ListOfLists[container_id].List=List_new;
 ListOfLists[container_id].n=n;
 free(List_h);
 ListOfLists[attach_id].List=0;   // this is now empty : hole is attached
 ListOfLists[attach_id].n=0;
 ListOfLists[attach_id].I = -1;
 //
 // update the new piece (boundary  + hole) normals
 //
 UpdateOutlineNormals(container_id,ListOfLists);
 return;
}
