//
// Function to subdivide a connected network of triangular facets so that
// each edge is broken in two at its mid point and every facet is divided
// into 4 smaller facets.

//  The data types (VERTEX, EDGE and FACET) will need to be modified
//  to fit whatever framework you are using.

typedef float VECTOR[3];

typedef struct tagVERTEX {
 // add other members as required
 VECTOR p;
} VERTEX;

typedef struct tagEDGE {
 long Vid[2];     // identity of vertices to which the edge is attached
 long Fid[2];     // list to identify the facets on either side of edge
 //
 // add other members as required
} EDGE;

typedef struct tagFACET {
 long Vid[3];     // identity of vertices to which facet is attached
 long Fid[3];     // list to indentity facets with an edge in common
 long Eid[3];     // list of the identity of edges surrounding the facet
 //
 // add other members as required
} FACET;


///////// The following data type is used internally by the function
///////// to record temporary information

typedef struct tagTEMPF {
 long V01;     // identity list of facets adjacent to this vertex
 long V12;
 long V20;
} TEMPF;

////////// The following functions will need to be provided to fulfil the
////////// described action.

// create an edge in database returns the identity of the edge
extern long   newE(VERTEX *, VERTEX *);
// create a facet in database return the identity of the facet
extern long   newF(VERTEX *, VERTEX *, VERTEX *);
// create a vertex in the database return the identity of the vertex
extern long  newV(VECTOR );
// scale the vector A with the scalar s,   R = sA
extern void   ScaleVector(float s,VECTOR A, VECTOR R);
// copy a vector from Source to Destination
extern void   CopyVector(VECTOR Source, VECTOR Destination);
// add two vectors R = A + B
extern void AddVectors(VECTOR A, VECTOR B, VECTOR R);
// subtract two vectors R = A - B
extern void SubtractVectors(VECTOR A, VECTOR B, VECTOR R);

///////// The following are standard functions and definitions that
///////// are usually provided as part of a compilers library.
///////// Substitute then with the ones you wish to use.

#define NULL 0
extern void *malloc(long X);      // allocate X bytes, return pointer to it.
// update memory block pointed to by P so that it has X bytes.
extern void *realloc(void *P, long X);
extern void free(void *P);        // release memory block pointed to by P

// Function to subdivide a connected network of facets. This function
// requires that the adjacency information "Fid[]" for each edge is known.

void SubdivideFacets(
  long   nVertices,         // Number of vertices
  VERTEX *VertexList,       // List of vertex structures
  long   nEdges,            // Number of edges
  EDGE   *EdgeList,         // List of edge structures
  long   nFacets,           // Number of facets
  FACET  *FacetList)        // List of facet structures
{
 long i,j,v0,v1,f,New_Vertex;
 TEMPF *flist=NULL;
 VECTOR p0,p1,dp,pc;
 //
 // Initialize the temporary data structure
 //
 if((flist=(TEMPF *)malloc(nFacets*sizeof(TEMPF))) == NULL)return;
 //
 // For each edge divide it in two by creating an additional edge. As each
 // edge is processed put a flag into the adjacent faces so that when we
 // come to divide the faces we will know to which vertices to attach the
 // new facets and internal edges. It is assumed that new vertices, edges
 // and facets are added at the end of the lists so that they will NOT
 // be subdivided.
 //
 for(i=0;i<nEdges;i++)
 {
   v0=EdgeList[i].Vid[0];  // vertices at either end of edge to be divided
   v1=EdgeList[i].Vid[1];
   CopyVector(VertexList[v0].p,p0);
   CopyVector(VertexList[v1].p,p1);
   SubtractVectors(p1,p0,dp);
   ScaleVector(0.5,dp,p1);
   AddVectors(p0,p1,pc);   // mid point of edge i
   New_Vertex=newV(pc);    // make new vertex and return its identity
   //
   // Make the new edge
   //
   newE((VertexList+New_Vertex),(VertexList+v1));
   //
   // modify the end point of the old one
   //
   EdgeList[i].Vid[1]=New_Vertex;
   //
   // place markers in adjacent faces, if there are adjacent faces
   //
   if((f=EdgeList[i].Fid[0]) >= 0)
   {
     if((FacetList[f].Vid[0] == v0 && FacetList[f].Vid[1] == v1) ||
        (FacetList[f].Vid[0] == v1 && FacetList[f].Vid[1] == v0))
     { // edge "i" occurs in facet "f" along side "01"
       // record the identity of the vertex added to the mid point of "i"
       flist[f].V01 = New_Vertex;
     }
     if((FacetList[f].Vid[1] == v0 && FacetList[f].Vid[2] == v1) ||
        (FacetList[f].Vid[1] == v1 && FacetList[f].Vid[2] == v0))
     { // edge "i" occurs in facet "f" along side "12"
       // record the identity of the vertex added to the mid point of "i"
       flist[f].V12 = New_Vertex;
     }
     if((FacetList[f].Vid[2] == v0 && FacetList[f].Vid[0] == v1) ||
        (FacetList[f].Vid[2] == v1 && FacetList[f].Vid[0] == v0))
     { // edge "i" occurs in facet "f" along side "20"
       // record the identity of the vertex added to the mid point of "i"
       flist[f].V20 = New_Vertex;
     }
   }
   if((f=EdgeList[i].Fid[1]) >= 0)  // the other side of edge "i"
   {
     if((FacetList[f].Vid[0] == v0 && FacetList[f].Vid[1] == v1) ||
        (FacetList[f].Vid[0] == v1 && FacetList[f].Vid[1] == v0))
     { // edge "i" occurs in facet "f" along side "01"
       // record the identity of the vertex added to the mid point of "i"
       flist[f].V01 = New_Vertex;
     }
     if((FacetList[f].Vid[1] == v0 && FacetList[f].Vid[2] == v1) ||
        (FacetList[f].Vid[1] == v1 && FacetList[f].Vid[2] == v0))
     { // edge "i" occurs in facet "f" along side "12"
       // record the identity of the vertex added to the mid point of "i"
       flist[f].V12 = New_Vertex;
     }
     if((FacetList[f].Vid[2] == v0 && FacetList[f].Vid[0] == v1) ||
        (FacetList[f].Vid[2] == v1 && FacetList[f].Vid[0] == v0))
     { // edge "i" occurs in facet "f" along side "20"
       // record the identity of the vertex added to the mid point of "i"
       flist[f].V20 = New_Vertex;
     }
   }
 }
 //
 // Now add the internal facets and edges for each subdivided facet.
 //
 for(j=0;j<nFacets;j++)
 {
   //
   // create the internal edge between V01 - V12  V12 - V20  V20 - V01
   //
   newE((VertexList+flist[j].V01),(VertexList+flist[j].V12));
   newE((VertexList+flist[j].V12),(VertexList+flist[j].V20));
   newE((VertexList+flist[j].V20),(VertexList+flist[j].V01));
   //
   // create the new facets V01-V12-V20  V01-V1-V12  V12-V2-V20
   //
   newF((VertexList+flist[j].V01),
        (VertexList+flist[j].V12),
        (VertexList+flist[j].V20));
   newF((VertexList+flist[j].V01),
        (VertexList+FacetList[j].Vid[1]),
        (VertexList+flist[j].V20));
   newF((VertexList+flist[j].V01),
        (VertexList+flist[j].V12),
        (VertexList+FacetList[j].Vid[2]));
   //
   // modify the original face to have vertices  V0-V01-V20
   //
   FacetList[j].Vid[1] = flist[j].V01;
   FacetList[j].Vid[2] = flist[j].V20;
 }
 //
 // Release temporary resources
 //
 free(flist);
 //
 // Note that when this procedure is complete the "adjacency" information is
 // corrupt and will need to be re-created before it is used again.
 //
 return;
}
