//
// Algorithm to loft a series of curves by joining them together
//

//////////  The data types (VECTOR and VERTEX) will need to be
//////////  modified to fit whatever framework you are using.

typedef float VECTOR[3];

typedef struct tagVERTEX {
 VECTOR p;
 // add other members as required
} VERTEX;

////////// The following functions will need to be provided to fulfil the
////////// described action.

// create an edge in database returns the identiry of new edge
extern long   newE(VERTEX *, VERTEX *);
// create a facet in database  returns the identity of new edge
extern long   newF(VERTEX *, VERTEX *, VERTEX *);
//  determine the euclidian distance between two point specified as vectors
extern float  Distance(VECTOR , VECTOR);

///////// The following are standard functions and definitions
///////// that are usually part of a compilers library.
///////// Substitute then with the ones you wish to use.

#define MAX_FLOAT 1.e30
#define NULL 0
extern void *malloc(long X);      // allocate X bytes, return pointer to it.
// update memory block pointed to by P so that it has X bytes.
extern void *realloc(void *P, long X);
extern void free(void *P);        // release memory block pointed to by P

/////////////////////// The following structures are used by the function
/////////////////////// and will be used by caller to set up the data.

typedef struct tagCURVE {
  struct tagVERTEX* v; // pointer to vertex associated with each edge in
                       // piece of outline.
  long   attached;     // identifier that records to which element of another
                       // curve this element is attached to.
} CURVE;

typedef struct tagLISTOFCURVES {
  struct tagCURVE* List; // array of structures with information about each
                         // edge it is assumed that the edges are ordered
                         // with the vertex at the start and then the edge
                         // joins with the vertex in the next structure in
                         // the list.
  long   n;              // number of edges in this piece of the outline
} LISTOFCURVES;

//
// functions uses internally by this algorithm
//

static void ReListCurve(CURVE *, long, long);

// LoftCurves is the entry point of the algorithm. The N curves must be
// ordered because they will be joined 0->1 1->2  ... (N-2)->(N-1) and
// each curve must > 3 vertices. It will be assumed that the last vertex
// is linked by an edge back to the first. It is also assumed that
// edges alrady exist between the vertices in the curves.

void LoftCurves(
     long         N,             // Number of curves forming the spec of loft
     LISTOFCURVES *ListOfCurves) // Array giving vertices
{
 long k;
 //
 // Take curves in pairs  0-1 1-2   (N-2)-(N-1)
 //
 for(k=0;k<N-1;k++)
 {
   int    swapped=0,switched;
   CURVE *Pn,*Pm;
   long   i,j,Nn,Nm,i0,i1,i2;
   float  distance,min_distance=MAX_FLOAT;
   //
   // Pick the appropriate pairs of curves
   //
   Pn=ListOfCurves[k  ].List; Nn=ListOfCurves[k  ].n;
   Pm=ListOfCurves[k+1].List; Nm=ListOfCurves[k+1].n;
   //
   // Find the vertices in curve 1 and 2 that is closest together
   //
   for(i=0;i<Nn;i++)for(j=0;j<Nm;j++)
   {
     if((distance=Distance((Pn[i].v)->p,(Pm[j].v)->p)) < min_distance)
     {
       min_distance=distance; i0=i; i1=j;
     }
   }
   //
   // Re-make the curves so that i0 and i1 are at the start of the list
   //
   if(i0 > 0)ReListCurve(Pn,Nn,i0);
   if(i1 > 0)ReListCurve(Pm,Nm,i1);
   //
   // Make curve Pn the curve with the fewest vertices/edges (elements)
   //
   if(Nm > Nn)
   {
     long t;
     CURVE *Pt;
     t=Nn;  Nn=Nm; Nm=t;
     Pt=Pn; Pn=Pm; Pm=Pt;
     swapped=1;
   }
   //
   // Test to see if path Pm should be reversed
   //
   if(Distance((Pn[1].v)->p,(Pm[1].v)->p) >
      Distance((Pn[1].v)->p,(Pm[Nm-2].v)->p))
   {
     //
     // If it should do it,  0 -> Nm-1      Nm-1 -> 0 etc.
     //
     VERTEX *Vt;
     for(i=0;i<Nm/2;i++)
     {
       Vt=Pm[i].v;
       Pm[i].v=Pm[Nm-1-i].v;
       Pm[Nm-1-i].v=Vt;
     }
   }
   //
   // Link the first vertices in Pn and Pm
   //
   newE(Pn[0].v,Pm[0].v);  Pn[0].attached=0;
   //
   // Join the remaining vertices in Pn to nearest vertex in Pm so that at
   // each stage no edges will cross back over the last one created.
   // Use the flag "i0" to limit the range of the test. At this stage
   // i0=0 because we may join vertex Pn(1)  to Pm(0)
   i0=0;
   for(i=1;i<Nn;i++)
   {
     //
     // Attach the remaining elements of Pn.  If i1 >= 0 then it is the ID
     // of the vertex in Pm to attach Pn(i) to.
     //
     min_distance=MAX_FLOAT;
     i1 = -1;
     //
     // Check all valid elements of Pm
     //
     if(i0 < Nm)for(j=i0;j<Nm;j++)
     {
       if((distance=Distance((Pn[i].v)->p,(Pm[j].v)->p)) < min_distance)
       {
         min_distance=distance; i1=j;  // Record closest
       }
     }
     //
     // When found make an edge and record connectivity. Restrict the
     // search for the next so that crossings are prevented.
     // Connect "i" in Pn to "i1" in Pm
     //
     if(i1 >= 0)
     {
       i0=i1;
       newE(Pn[i].v,Pm[i1].v);  Pn[i].attached=i1;
     }
   }
   //
   // Go round curve Pn again joining in any as yet unconnected vertices
   // in Pm make the necessary facets and add edges as required.
   //
   for(i=0;i<Nn;i++)
   {
     VERTEX *V0,*V1,*V2;     // These will be the vertices for new facets
     switched=0;             // This becomes set when unconnected vertices
                             // flip from being connected to "i" to being
                             // connected to "i+1".
     V0=Pn[i].v;             // Pointer to vertex at "i" in Pn
     i0=Pn[i].attached;      // ID of vertex in Pm connected to "i"
     V2=Pm[i0].v;            // Pointer to vertex in Pm to which "i" is
                             // connected. It will be updated to keep the
                             // clockwise ordering of facets correct.
     if(i == Nn-1)           // At the last vertex in Pn
     {
       V1=Pn[0].v;           // The next vertex in the loop is "0"
       i1=Nm;                // Beyond the end of the list
       i2=0;
     }
     else
     {
       V1=Pn[i+1].v;         // Pointer to vertex "i+1" in Pn
       i1=Pn[i+1].attached;  // Entry in Pm to which "i+1" in Pn is attached
       i2=i+1;               // Next element of Pn
     }
     if(i1 == i0)
     {
       //
       // No edges required only one facet
       //
       newF(V0,V1,V2);
     }
     else
     {
       //
       // At least 1 edge and 2 faces must be added
       //
       if(i1 > i0+1)
       {
         //
         // There is at least one unattached vertex in this section of Pm
         // at least 2 edges and 3 facets will be needed.
         // Start attaching them to "i" until they get
         // nearer to "i+1" then flip the attachment.
         //
         for(j=i0+1;j<i1;j++)
         {
           V1=Pm[j].v;
           if(!switched)
           {
             //
             //  Switch has not occured so attach edge from "j" in Pm
             //  to "i" in Pn and make facet.
             //
             newE(V0,V1);
             newF(V0,V1,V2);
             //
             // Check to see if vertex at "i+1" is closer to V1 than
             // vertex at "i", it is then switch.
             //
             if(Distance(V1->p,(Pn[i2].v)->p) > Distance(V1->p,V0->p))
             {
               newE(V1,Pn[i2].v);
               newF(V0,Pn[i2].v,V1);
               switched=1;
               V2=V1;
               V0=Pn[i2].v;
             }
             else
             {
               V2=V1;
             }
           }
           else   // switch has occurred so just go on adding
           {
             newE(V0,V1);
             newF(V0,V1,V2);
             V2=V1;
           }
         }
       }
       if(!switched)
       {
         //
         // The run has been completed but since a switch was not done
         // two faces and an edge must be added to tidy it up. This will
         // also cover the case where there are no unattached vertices in
         // this section of Pm.
         //
         newE(V2,Pn[i2].v);
         newF(V0,Pn[i2].v,V2);
         if(i1 >= Nm)i1=0;   // Link to vertex at start of curve
         newF(Pn[i2].v,Pm[i1].v,V2);
       }
       else
       {
         //
         // The switch has been done - all the edges are in place it
         // only remains to put the final facet in place.
         //
         if(i1 >= Nm)i1=0;
         newF(Pn[i2].v,Pm[i1].v,V2);
       }
     }
   }
 }
 return;
}


static void ReListCurve(CURVE* P, long n, long s)
{
 //
 // Re-order the path so that element "s > 0" is the first. Remember
 // that the path is closed so after re-ordering the last element is "s-1"
 //
 long i,j;
 CURVE *Temp;
 //
 // Make a temporary copy starting at "s" there is no need to copy the
 // attached member of the CURVE structure at this stage in the algorithm
 // they all all the same "undefined".
 //
 if((Temp=(CURVE *)malloc(n*sizeof(CURVE))) == NULL)return;
 for(j=0,i=s;i<n;i++,j++)
 {
   Temp[j].v = P[i].v;
 }
 for(i=0;i<s;i++,j++)
 {
   Temp[j].v = P[i].v;
 }
 //
 // Put copy back into P
 //
 for(i=0;i<n;i++)
 {
   P[i].v = Temp[i].v;
 }
 free(Temp);
 return;
}
