//
// Algorithm to produce a suface of revolution constructed from a network
// of connected edges and triangular facets
//

//////////  The data types (VECTOR, MATRIX and VERTEX) will need to be
//////////  modified to fit whatever framework you are using.

typedef float VECTOR[3];
typedef float MATRIX[4][4];

typedef struct tagVERTEX {
 VECTOR p;
 // add other members as required
} VERTEX;

typedef struct tagEDGE {
 ;
 // this will need at least some information about the two vertices
 // to which the edge is attached.
} EDGE;

typedef struct tagFACET {
 ;
 // this will need at least some information about the three vertices
 // to which the edge is attached.
} FACET;

////////// The following functions will need to be provided to fulfil the
////////// described action.

// create an edge in database returns a pointer to the new edge
extern EDGE*    newE(VERTEX *, VERTEX *);
// create a facet in database  returns the identity of new facet
extern FACET*   newF(VERTEX *, VERTEX *, VERTEX *);
// create a vertex in the database returns pointer to the new vertex
extern VERTEX*  newV(VECTOR );
// multiply two 4x4 matrices together Mr = Ma x Mb
extern void  MultiplyMatrix(MATRIX Ma, MATRIX Mb, MATRIX Mr);
// transform vector by multiplying it with a matrix  Vr = M x Va
extern void  TransformVector(MATRIX M, VECTOR Va, VECTOR Vr);
//  Scale the vector A with the scalar s,   R = sA
extern void   ScaleVector(float s,VECTOR A, VECTOR R);

///////// The following are standard functions and definitions
///////// that are usually part of a compilers library.
///////// Substitute then with the ones you wish to use.

#define MAX_FLOAT 1.e30
#define NULL 0
// allocate X bytes, return pointer to it.
extern void *malloc(long X);
// update memory block pointed to by P so that it has X bytes.
extern void *realloc(void *P, long X);
// release memory block pointed to by P
extern void free(void *P);
// return the sin of the angle T (radians)
extern float sin(float T);
// return the cos of the angle T (radians)
extern float cos(float T);
// return the inverse tangent of dy/dx in range (-pi, pi)
extern float atan2(float dy, float dx);
// return the square root of X
extern float sqrt(float X);

/////////////////////// The following structures are used by the function
/////////////////////// and will be used by caller to set up the data.

typedef struct tagCURVE {
  struct tagVERTEX* v; // pointer to vertex associated with each edge in
                       // the curve to be rotated.
} CURVE;

//
// functions uses internally by this algorithm
//

static void Translate(MATRIX , VECTOR );
static void RotateRoundZ(MATRIX , float );
static void RotateRoundX(MATRIX , float );
static void RotateRoundY(MATRIX , float );
static void RotateRoundAxis(MATRIX ,  VECTOR ,  VECTOR , float );

// Entry function SurfaceOfRevolution
//
// Function to generate a surface of revolution given an axis to revolve
// about and a list of vertices forming the curve "C" to be revolved. It is
// assumed that the vertices in "C" are already joined with edges and if "C"
// is closed then vertex (n-1) is joined to vertex 0.
//

void SurfaceOfRevolution(
     long         n,             // Number of vertices in curve array
     CURVE        *List,         // Array giving vertices
     long         loop,          // =1 if curve is closed loop 0 otherwise
     VECTOR       p,             // point on the axis of revolution
     VECTOR       d,             // direction of the axis of revolution
     long         Nsteps,        // number of steps to rotate
     float        theta          // angular rotation (degrees)
                                 //+ve=clockwise -ve=anti_c
 )
{
 int full_rev=0;
 float delta_theta;
 long i,j;
 CURVE *tList,*nList,*pList;  // temporary list
 MATRIX M_rotation;
 VECTOR pp;
 //
 // Can only work with <= one revolution
 //
 if((long)theta > 360 || (long)theta < -360)return;
 //
 // Need to make two lists of vertices to be joined as well as keeping the
 // original list if it is to be joined to complete the surface of revolution.
 // The pList can be thought of as the list of vertices in the previous
 // section. The nList is built and then joined to the pList.
 //
 if((pList=(CURVE *)malloc(n*sizeof(CURVE))) == NULL)return;
 if((nList=(CURVE *)malloc(n*sizeof(CURVE))) == NULL)return;
 //
 // delta_theta is the angular displacement of each section (in radians)
 // so do the conversion as well
 //
 delta_theta=theta/(float)Nsteps*3.1415926/180.0;
 //
 // Calculate the rotation matrix to rotate by delta_theta about axis
 // through "p" in direction "d".
 //
    RotateRoundAxis(M_rotation,p,d,delta_theta);
 //
 // If the surface is closed (revolution is 360 degrees) then 1 fewer steps
 // in the rotation is required but the vertices in the last step must be
 // joined with edges and facets back to the original curve.
 //
 if((long)theta == 360 || (long)theta == -360)
 {
   full_rev=1;
   Nsteps--;
 }
 //
 // pList starts at the original curve - so copy it
 //
 for(i=0;i<n;i++)pList[i].v = List[i].v;
 //
 // Make each section by creating vertices at the previous section, rotating
 // them into position by(delta_theta) and then building edges and facets to
 // join the new section on to the last section.
 //
 for(j=0;j<Nsteps;j++)
 {
   for(i=0;i<n;i++) // make the new vertices in position rotated
                    // from last step
   {
     TransformVector(M_rotation,(pList[i].v)->p,pp);
     nList[i].v=newV(pp);
     //
     // make the edges and facets
     //
     newE(nList[i].v,pList[i].v);
     if( i > 0)
     {
       // add the copy of the edge between list entry vertices  i and  i-1
       //
       newE(nList[i].v,nList[i-1].v);
       //
       // make the cross edge and add the facets. At the first vertex there is
       // no previous one to link to. Make facet order anti-clock to ensure
       // that the normal points outwards.
       //
       newE(pList[i-1].v,nList[i].v);
       newF(pList[i-1].v,nList[i].v,nList[i-1].v);
       newF(pList[i-1].v,pList[i].v,nList[i].v);
     }
     if( loop && i == n-1 )
     {
       // add the copy of the edge between list entry vertices n-1 and 0
       //
       newE(nList[i].v,nList[0].v);
       //
       // if the curve being revolved is closed edges and faces making the
       // link between entries (n-1) and (0) in the lists need to be created
       //
       newE(pList[i].v,nList[0].v);
       newF(pList[i].v,nList[0].v,nList[i].v);
       newF(pList[i].v,pList[0].v,nList[0].v);
     }
   }
   //
   // before moving on to the next section set the curve list so we start
   // working from step "j" i.e. swap the  "n" and "p" lists.
   //
   tList=pList;
   pList=nList;
   nList=tList;
 }
 if(full_rev)
 {
   //
   // If a full revolution then the last step has to be linked back to the
   // first. Its basically the same as the last step except that nList
   // is replaced by the original list.
   //
   for(i=0;i<n;i++)
   {
     newE(nList[i].v,List[i].v);
     if( i > 0)
     {
       newE(pList[i-1].v,List[i].v);
       newF(pList[i-1].v,List[i].v,List[i-1].v);
       newF(pList[i-1].v,pList[i].v,List[i].v);
     }
     if( loop && i == n-1 )
     {
       newE(pList[i].v,List[0].v);
       newF(pList[i].v,List[0].v,List[i].v);
       newF(pList[i].v,pList[0].v,List[0].v);
     }
   }
 }
 free(nList);
 free(pList);
 return;
}

static void RotateRoundZ(MATRIX M, float theta)
//
// build transformation matrix for rotation round the Z axis
//
{
 long i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  M[i][j]=0.0;
  if(i == j)M[i][j]=1.0;
 }
  M[0][0]= cos(theta);
  M[0][1]=-sin(theta);
  M[1][0]= sin(theta);
  M[1][1]= cos(theta);
  return;
}

static void RotateRoundY(MATRIX M, float theta)
//
// build transformation matrix for rotation round the Y axis
//
{
 short i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  M[i][j]=0.0;
  if(i == j)M[i][j]=1.0;
 }
  M[0][0]= cos(theta);
  M[0][2]=-sin(theta);
  M[2][0]= sin(theta);
  M[2][2]= cos(theta);
  return;
}

static void RotateRoundX(MATRIX M, float theta)
//
// build transformation matrix for rotation round the X axis
//
{
 long i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  M[i][j]=0.0;
  if(i == j)M[i][j]=1.0;
 }
  M[1][1]= cos(theta);
  M[1][2]=-sin(theta);
  M[2][1]= sin(theta);
  M[2][2]= cos(theta);
  return;
}

static void Translate(MATRIX M, VECTOR d)
//
// Return translation matrix for translation by  d
//
{
 long i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  M[i][j]=0.0;
  if(i == j)M[i][j]=1.0;
 }
 M[0][3]=d[0];
 M[1][3]=d[1];
 M[2][3]=d[2];
 M[3][3]=1.0;
 return;
}

static void RotateRoundAxis(
  MATRIX M,       // transformation matrix (returned)
  VECTOR p,       // point on axis of rotation
  VECTOR d,       // direction of axis to rotate round
  float angle     // angle to rotate +ve =clockwise (radians)
                  // round the direction d
)
{
  float x,y,z,phi,theta,xy;
  VECTOR pm;
  MATRIX M1,M2,M3;
  x=d[0]; y=d[1]; z=d[2];
  xy=x*x+y*y;
  ScaleVector(-1.0,p,pm);
  Translate(M1,pm);
  if(xy < 1.0e-5)
  {
    //
    // vertical so just rotate about z  and return
    //
    if(z > 0.0)
      RotateRoundZ(M2, angle);
    else
      RotateRoundZ(M2,-angle);
    MultiplyMatrix(M2,M1,M3);
    Translate(M1,p);
    MultiplyMatrix(M1,M3,M);
    return;
  }
  xy=sqrt(xy);
  if     (x == 0.0 && y > 0.0)phi =  1.57079633;
  else if(x == 0.0 && y < 0.0)phi = -1.57079633;
  else phi = atan2(y,x);
  theta = atan2(z,xy);
  //
  // Align the axis of rotation to the X axis with p at (0,0,0)
  //
  RotateRoundZ(M2, -phi);
  MultiplyMatrix(M2,M1,M3);
  RotateRoundY(M2, -theta);
  MultiplyMatrix(M2,M3,M1);
  //
  // Rotate by the required ammount round the X axis
  //
  RotateRoundX(M2, angle);
  MultiplyMatrix(M2,M1,M3);
  //
  // Apply the inverse of the alignment transformation
  //
  RotateRoundY(M2,theta);
  MultiplyMatrix(M2,M3,M1);
  RotateRoundZ(M2,phi);
  MultiplyMatrix(M2,M1,M3);
  Translate(M1,p);
  MultiplyMatrix(M1,M3,M);
}
