//
// Algorithm to bevel an outline by adding an outline to it and
// joining them up with edges and facets. There is some commonality
// with the procedure to cap an outline (fill polygon wiht triangular
// facets).
// However the LIST structure has an extra vector to accomodate the
// average vector between edge "i" and edge "i-1" needed by the algorithm.
// One normal vector "Normal" specifies the plane in which all the pieces
// of the outline lies and the order in which the items occur in the LIST is
// assumed to be anti-clockwise round "Normal" i.e. according to the vector
// product.
//

//////////  The data types (VECTOR and VERTEX) will need to be modified
//////////  to fit whatever framework you are using.

typedef float VECTOR[3];

typedef struct tagVERTEX {
 // add other members as required
 VECTOR p;
} VERTEX;

typedef struct tagEDGE {
 ;
 // this will need at least some information about the two vertices
 // to which the edge is attached.
} EDGE;

typedef struct tagFACET {
 ;
 // this will need at least some information about the three vertices
 // to which the edge is attached.
} FACET;

////////// The following functions will need to be provided to fulfil the
////////// described action.

// create an edge in database returns a pointer to the new edge
extern EDGE*  newE(VERTEX *, VERTEX *);
// create a facet in database  returns the identity of new facet
extern FACET* newF(VERTEX *, VERTEX *, VERTEX *);
// create a vertex in the database returns pointer to the new vertex
extern VERTEX*  newV(VECTOR );
// make vector V unit length
extern void   Normalize(VECTOR V);
// R = vector cross product of U x V
extern void   Cross(VECTOR U, VECTOR V, VECTOR R);
// return the dot product of A.B
extern float  Dot(VECTOR A, VECTOR B);
//  Scale the vector A with the scalar s,   R = sA
extern void   ScaleVector(float s,VECTOR A, VECTOR R);
// copy a vector from Source to Destination
extern void   CopyVector(VECTOR Source, VECTOR Destination);
// add two vectors R = A + B
extern void AddVectors(VECTOR A, VECTOR B, VECTOR R);
// subtract two vectors R = A - B
extern void SubtractVectors(VECTOR A, VECTOR B, VECTOR R);

///////// The following are standard functions and definitions that
///////// are usually part of a compilers library.
///////// Substitute then with the ones you wish to use.

#define NULL 0
extern void *malloc(long X);      // allocate X bytes, return pointer to it.
extern void free(void *P);        // release memory block pointed to by P
extern float fabs(float D);       // return floating point absolute value of D
extern float sqrt(float D);       // return floating point square root of  D
extern float tan(float A);        // return tangent of angle A (radians)

///////////// Define the angle between two adjacent edges beyond which
///////////// an additional vertex+edge will need to be added

#define OBTUSE_A -0.7071

/////////////////////// The following structures are used by the function
/////////////////////// and will be used by caller to set up the data.

// Each element in the list describing a part of the outline of the
// curve to be bevelled is described in a "LIST" structure.

typedef struct tagLIST {
  struct tagVERTEX* v; // pointer to the vertex associated with this edge
                       // of a piece of the outline.
  long   last,         // last entry in the list that is currently used
         next,         // next entry in the list that is currently used
         inc;          // this edge is included in current piece.
  VECTOR n;    // A vector perpendicular to this edge pointing in the
               // direction the bevel is to be made. It should also be
               // perpendicular to a vector that is normal to the plane in
               // which this piece of the outline lies.
               // (n will be calculated !! in the bevel algorithm
               // NB. it was NOT calculated in the capping procedure)
  VECTOR p;    // Vector giving position of vertex v (can be copy of v->p)
  VECTOR nav;  // Average of n for this edge and previous edge. (Added for
               // bevelling.  (Will be calculated !)
} LIST;

//
// The curve may consist of several pieces. Each piece is described by
// a "LISTOFLISTS" data structure.  The full curve will be given by
// an array of "LISTOFLISTS" structures.
//

typedef struct tagLISTOFLISTS {
  struct tagLIST* List; // Array of structures with information about a vertex/
                        // edge pair. It is assumed that the vertex is at the
                        // start of the edge which then joins it
                        // to the vertex in the next entry in the List.
  long   n,             // Number of edges in this piece of the outline
         I,             // inside / outside flag
         done;          // Finished processing this piece
} LISTOFLISTS;

////
//// functions uses internally by this algorithm
////

static long  IsCurveInside(LISTOFLISTS *, long , long );

// BevelPolygon is the entry function of the algorithm.

void BevelPolygon(
 long N,                    // number of pieces in outline
 LISTOFLISTS *ListOfLists,  // structure describing the outlines
 VECTOR Normal,             // A vector normal to the plane containing the
                            // set of curves. "Normal" and the order in
                            // which the entries in "ListOfLists[i].List"
                            // should be recorded should confrorm to the
                            // vector product convention.
 float  depth,              // depth of the bevel (always +ve)
 float  alpha               // angle of the bevel (degrees)
 )
//
//                            Normal
//                              ^
//                              |
//                              |
//                              |
//                          /--------\
//                         /          \
//                        /       alpha\
//                       ----------------
//                                 ->    <-  depth
//

// The function adds a bevel to an outline in N closed pieces. The pieces
// are given in an array of LISTOFLISTS structures that are filled before
// calling to function BevelPolygon.
//
{
 long i,j,k,n,h_count;
 LIST *List;
 VECTOR p0,p1,p10;
 VECTOR v1,v2,v3,vab,Nav,N_disp;
 float dd,dplane;
 long inside,add_point,first_vertex,new_count;
 VERTEX *nv,*lv,*fv,**Vlist;
 ScaleVector(depth*tan(alpha/180.0*3.1415926),Normal,N_disp);
 //
 // Make the normals  "n" and "nav" for each edge in each piece of the
 // outline to be bevelled. These normals point  "outwards" from all curves.
 //
 for(i=0;i<N;i++)
 {
   List=ListOfLists[i].List;
   for(j=0;j<ListOfLists[i].n;j++)  // first do the  "n"
   {
     k=List[j].next;
     CopyVector(List[j].p,p0);
     CopyVector(List[k].p,p1);
     SubtractVectors(p1,p0,p10);
     Cross(p10,Normal,List[j].n);
     Normalize(List[j].n);
   }
   for(j=0;j<ListOfLists[i].n;j++)  // then do the   "nav"
   {
     k=List[j].last;
     AddVectors(List[j].n,List[k].n,List[j].nav);
     Normalize(List[j].nav);
   }
 }
 //
 // is curve inside another curve (we don't  really need to know which
 // curve curve "i" is inside)
 // function returns -1 if not or index of curve it is inside
 //
 for(i=0;i<N;i++)
 {
   ListOfLists[i].I=IsCurveInside(ListOfLists,i,N);
   ListOfLists[i].done=0;
 }
 //
 // is it a hole or inside a hole ?
 //
 for(i=0;i<N;i++)
 {
   h_count=0;
   //
   // if inside work outwards to see how may curves it lies inside
   //
   if((j=ListOfLists[i].I) >= 0)while(j >= 0){
     h_count++;
     j=ListOfLists[j].I;
   }
   //
   // if a curve is inside an even number of curves then it is NOT the
   // outline of a hole and should be identified as a normal outline
   //
   if(h_count %2 == 0)ListOfLists[i].I = -1;
 }
 //
 // Having determined which pieces are the outlines of holes we need to
 // make sure that the normal associated with every edge in a piece
 // points to the side on which we wish to build the bevel this means
 // reversing the normals that are NOT the outlines of holes.
 //
 for(i=0;i<N;i++)
 {
   if(ListOfLists[i].I < 0)
   {
     List=ListOfLists[i].List;
     for(j=0;j<ListOfLists[i].n;j++)
     {
       ScaleVector(-1.0,List[j].n,List[j].n);
       ScaleVector(-1.0,List[j].nav,List[j].nav);
     }
   }
 }
 //
 // Now make the bevel - one curve at a time. The bevelled edges are
 // kept in the plane of the curve until the full bevel has been built.
 //
 for(i=0;i<N;i++)
 {
   n=ListOfLists[i].n;
   List=ListOfLists[i].List;
   //
   // Create a temporary list to record the identity of the the
   // vertices added to make the bevel. Make it bigger than will ever be
   // required so that it doesn't have to be extended (its unlikely to ever
   // be so big as to fall foul of resource limitations).
   //
   if((Vlist=(VERTEX **)malloc(n*2*sizeof(VERTEX *))) == NULL)goto FINISHED;
   //
   // set flags to keep track of where we are
   //
   lv = NULL;
   fv = NULL;
   first_vertex=1;
   new_count=0;
   for(j=0;j<n;j++)
   {
     CopyVector(List[j].p,v1);
     k=List[j].next;
     CopyVector(List[k].p,v2);
     SubtractVectors(v2,v1,vab);
     Normalize(vab);
     k=List[j].last;
     if(Dot(vab,List[j].nav) < OBTUSE_A)
     {
       //
       // large obtuse angle add vertex
       //
       ScaleVector(depth,List[k].n,v2);
       AddVectors(v1,v2,v3);
       nv=newV(v3);
       Vlist[new_count++]=nv;
       if(fv == NULL)fv=nv;
       newE(List[j].v,nv);
       if(lv != NULL){
         newE(nv,lv);
         newE(nv,List[k].v);
         newF(nv,List[k].v,List[j].v);
         newF(nv,lv,List[k].v);
       }
       lv=nv;
       ScaleVector(depth,List[j].n,v2);
       AddVectors(v1,v2,v3);
       nv=newV(v3);
       Vlist[new_count++]=nv;
       if(fv == NULL)fv=nv;
       newE(List[j].v,nv);
       newE(lv,nv);
       newF(lv,nv,List[j].v);
       lv=nv;
     }
     else{
       //
       // first check to see if a vertex should be added
       //
       dd=fabs(Dot(List[j].nav,List[j].n));
       //
       // get angular distance
       // to keep bevel same size
       //
       if(dd < 1.e-10)dd=depth;
       else           dd=depth/dd;
       //
       // Nav is a vector normal to a plane in which the vector of the "nav"
       // of previous edge lies. The basic idea of deciding whether an a
       // vertex is to be added is to see whether the previous vertex is
       // on the far side of this plane. If it is then the vertex is added.
       //
       CopyVector(List[k].p,v2);
       Cross(Normal,List[k].nav,Nav);
       Normalize(Nav);
       dplane=Dot(List[j].nav,Nav);
       if(fabs(dplane) < 1.e-10)add_point=1;
       else
       {
         SubtractVectors(v2,v1,v3);
         dplane=Dot(v3,Nav)/dplane;
         if(dplane < 0.0 || dplane >= dd)add_point=1;
         else                            add_point=0;
       }
       if(add_point == 1)
       {
         ScaleVector(dd,List[j].nav,v2);
         AddVectors(v1,v2,v3);
         nv=newV(v3);
         Vlist[new_count++]=nv;
         if(fv == NULL)fv=nv;
         if(lv >= 0)
         {
           newE(lv,nv);
           newE(nv,List[k].v);
           newF(nv,List[k].v,List[j].v);
           newF(nv,lv,List[k].v);
         }
         lv=nv;
         newE(List[j].v,nv);
       }
       else
       {
         //
         // don't add point
         //
         if(lv != NULL)
         {
           newE(List[j].v,lv);
           newF(List[j].v,lv,List[k].v);
           //
           // update the average normal for edge j to take account of
           // fact that it must be shifted forward to line up with last
           //
           SubtractVectors(lv->p,List[j].p,List[j].nav);
           Normalize(List[j].nav);
         }
         if(j == 0)first_vertex=0;
       }
     }
   }
   //
   // fix up the loop end->start
   //
   if(fv != NULL && lv != NULL)
   {
     newE(lv,fv);
     newE(lv,List[0].v);
     newF(lv,fv,List[0].v);
     newF(lv,List[0].v,List[n-1].v);
     if(first_vertex == 0)
     {
       newE(fv,List[0].v);
       newF(fv,List[1].v,List[0].v);
     }
   }
   //
   // Move the vertices in the bevel in the direction of "Normal".
   // This forms the bevel.
   //
   for(j=0;j<new_count;j++)
   {
     AddVectors(Vlist[j]->p,Nav,Vlist[j]->p);
   }
   free(Vlist);
 }
 FINISHED:
 return;
}

static long IsCurveInside(LISTOFLISTS *ListOfLists,
                          long list_id,
                          long n_lists)
//
// Determine whether the outline piece ListOfLists(list_id) lies inside
// any other outline piece. If it does return the identity of the other
// piece that is closest to it at the start of piece  list_id.
// Otherwise return -1
//
// Basically these tests are similar in many respects to those
// performed by function IsEdgeInside
//
{
 VECTOR p_centre,p0,p1,p10,p01,test_normal,pa,pb,edge_normal,pac,pn,pnb,pba,pi;
 long i,j,k,l,number_of_crosses,n,identity = -1;
 float mu,mu1,denominator,mumin,mumumin=1.0e35;
 LIST *List;
 List=ListOfLists[list_id].List;
 CopyVector(List[0].p,p0);
 CopyVector(List[1].p,p1);
 SubtractVectors(p1,p0,p10);
 CopyVector(p10,p01);
 Normalize(p01);
 CopyVector(List[0].n,test_normal);
 ScaleVector(0.5,p10,p_centre);
 AddVectors(p0,p_centre,p_centre);            // mid point of first edge
 //
 // check all other curves to see if vector from p_centre in direction
 // test_normal crosses any of the other curves and if so make a count
 // and find the identity of the curve that is closest to "list_id"
 //
 for(l=0;l<n_lists;l++)
 {
   if(l == list_id)continue;
   number_of_crosses=0;
   List=ListOfLists[l].List;
   n=ListOfLists[l].n;
   mumin=1.0e35;
   for(i=0;i<n;i++)
   { /* check to see if test_normal crosses edges in path */
     CopyVector(List[i].p,pa);
     j=List[i].next;
     CopyVector(List[j].p,pb);
     CopyVector(List[i].n,edge_normal);
     SubtractVectors(pa,p_centre,pac);
     mu1=Dot(edge_normal,test_normal);
     if(fabs(mu1) > 0.0001)
     {
       mu=Dot(pac,edge_normal)/mu1;
       if(mu > 0.0000)
       {   // look for intersections on one side
         ScaleVector(mu,test_normal,pi);
         AddVectors(p_centre,pi,pi);    // intersection to plane of test edge
         SubtractVectors(pb,pa,pba);
         k = -1; denominator = 1.0e-6;
         if     (fabs(pba[0]) > denominator){k=0; denominator =fabs(pba[0]);}
         else if(fabs(pba[1]) > denominator){k=1; denominator =fabs(pba[1]);}
         else if(fabs(pba[2]) > denominator){k=2; denominator =fabs(pba[2]);}
         if(k >= 0){
           mu1=(pi[k]-pa[k])/pba[k];
           if(fabs(mu1) < 1.0e-7)
           {                        // at start of segment
             k=List[i].last;
             CopyVector(List[k].p,pn);
             SubtractVectors(pa,pn,pnb);
             if(Dot(pba,p01)*Dot(pnb,p01) > 0.0)number_of_crosses++;
           }
           else if(fabs(1.0-mu1) < 1.e-7)
           {                                 // at end of segment
             k=List[j].next;
             CopyVector(List[k].p,pn);
             SubtractVectors(pn,pb,pnb);
             if(Dot(pba,p01)*Dot(pnb,p01) > 0.0)number_of_crosses++;
           }
           else if(mu1 > 0.0 && mu1 < 1.0)
           {
             number_of_crosses++;
           }
           if(mu < mumin)mumin=mu;
         }
       }
     }
   }
   if(number_of_crosses%2 == 1 && mumin <= mumumin){identity=l; mumumin=mumin;}
 }
 return identity;
}
