//
// Function to orientate consistently a set of networks of triangular facets
//

///////// The following are standard functions and definitions
///////// that are usually part of a compilers library.
///////// Substitute then with the ones you wish to use.

#define MAX_FLOAT 1.e30

//  The data types (EDGE and FACET) will need to be modified
//  to fit whatever framework you are using.

typedef float VECTOR[3];

typedef struct tagVERTEX {
 VECTOR p;          // position of the vertex in 3D space
 //
 // add other members as required
} VERTEX;

typedef struct tagEDGE {
 long Vid[2];     // identity of vertices to which the edge is attached
 long Fid[2];     // list to identify the facets on either side of edge   *
 //
 // add other members as required
} EDGE;

typedef struct tagFACET {
 long Vid[3];     // identity of vertices to which facet is attached
 long Fid[3];     // list to indentity facets with an edge in common      *
 long Eid[3];     // list of the identity od edges surrounding the facet  *
 //
 // add other members as required
 //
 long Sid;        // identifies to which piece in the moded this facet belongs
} FACET;

//
// The following external function is required to calculate the adjacency
// data, it's source is given elsewhere.
//

extern void MakeAdjacencyData(long, long, EDGE  *, long, FACET *);

////////// The following functions will need to be provided to fulfil the
////////// described action.

// make vector V unit length
extern void   Normalize(VECTOR V);
// R = vector cross product of U x V
extern void   Cross(VECTOR U, VECTOR V, VECTOR R);
// return the dot product of A.B
extern float  Dot(VECTOR A, VECTOR B);
// add two vectors R = A + B
extern void AddVectors(VECTOR A, VECTOR B, VECTOR R);
// subtract two vectors R = A - B
extern void SubtractVectors(VECTOR A, VECTOR B, VECTOR R);

//
// functions used internally by this algorithm
//

static void MakeSameOrientation(long , long);
static void IdentifyConnected(long, long);

//
// The following global variables are used to communicate to the
// recursive function. Using global variables keeps the size of the
// stack to a minimum.
//

static long  count_of_pieces;
static FACET *LocalFacetList;

// Function to fill orientate the networks.

void OrientateNetworks(    // Parameters
  long   nVertex,          // Number of vertices
  VERTEX *VertexList,      // List of vertex structures
  long   nEdges,           // Number of edges
  EDGE   *EdgeList,        // List of edge structures
  long   nFacets,          // Number of facets
  FACET  *FacetList)       // List of facet structures
{
 int i,j;
 count_of_pieces=0;        // initialize the count of number of pieces
 LocalFacetList=FacetList; // local pointer to list of facets.
 //
 // Fill in the adjacency information in the FACET data structure,
 // particularly the "Fid" entries that for each facet record its nearest
 // neighbours.
 //
 MakeAdjacencyData(nVertex,nEdges,EdgeList,nFacets,FacetList);
 //
 // For each facet record that it is "unidentified" i.e. not part of any
 // pieces and not yet processed.
 //
 for(i=0;i<nFacets;i++)FacetList[i].Sid = -1;
 //
 // Process all the facets in the model. Start with the first and work
 // through the lsit skipping over any that are identified. Some  will
 // be identified during the recursve use of the  "IdentifyConnected"
 // function.
 //
 for(i=0;i<nFacets;i++)
 {
   if(FacetList[i].Sid < 0)IdentifyConnected(-1,i);
 }
 //
 // At this point each piece of the model that forms a connected fragment
 // have consistent ordering. The remaining code checks whether the
 // orientation is appropriate and if not reverses the order in which the
 // vertex information is recorded.
 //
 if(count_of_pieces > 0)for(i=0;i<count_of_pieces;i++)
 {
   VECTOR p,p10,p20,n;
   float d,d_min =  MAX_FLOAT;
   long  closest_id = -1;
   //
   // Find the facet with the leftmost centroid X coordinate and the
   // vertex ordering consistent with a normal points to the left. You may
   // wish to substitute a better test that is more careful because this
   // one can be upset on a few occasions.
   //
   for(j=0;j<nFacets;j++)
   {
     if(FacetList[j].Sid == i)   // this piece
     {
       if((d=(VertexList[FacetList[j].Vid[0]].p[0]*0.3333 +
              VertexList[FacetList[j].Vid[1]].p[0]*0.3333 +
              VertexList[FacetList[j].Vid[2]].p[0]*0.3333)) < d_min)
       {
         d_min=d;
         closest_id=j;
       }
     }
   }
   //
   // If there is a closest facet then calculate its surface normal by
   // taking the vector cross product of vectors formed between vertices
   // 0 - 1  and 0 - 2.  If the scalar product with vector in +ve x direction
   // is > 0 then all the facets in this piece must have their vertex
   // ordering reversed.
   //
   if(closest_id >= 0)
   {
     SubtractVectors(VertexList[FacetList[closest_id].Vid[1]].p,
                     VertexList[FacetList[closest_id].Vid[0]].p,p10);
     SubtractVectors(VertexList[FacetList[closest_id].Vid[2]].p,
                     VertexList[FacetList[closest_id].Vid[0]].p,p20);
     Cross(p10,p20,n);
     if(n[0] > 0.0)
     {
       long temp_v;
       //
       // the current order of the facets in piece "i" is towards the
       // the right and all those facets must be reversed;
       //
       for(j=0;j<nFacets;j++)
       {
         if(FacetList[j].Sid == i)   // reverse this piece
         {
           temp_v=FacetList[j].Vid[1];
           FacetList[j].Vid[1]=FacetList[j].Vid[2];
           FacetList[j].Vid[2]=temp_v;
         }
       }
     }
   }
 }
 return;
}

static void MakeSameOrientation(long last_one_set, long set_this_one)
{
 //
 // Check whether this facet has the same orientation as it's neighhbour.
 // If not then reverse two of the vertices.
 //
 FACET *f;
 long i0,i1,i2,l0,l1,l2,a0,a1,a2;
 f=(LocalFacetList+last_one_set);
 l0=f->Vid[0]; l1=f->Vid[1]; l2=f->Vid[2];
 f=(LocalFacetList+set_this_one);
 i0=f->Vid[0]; i1=f->Vid[1]; i2=f->Vid[2];
 a0=f->Fid[0]; a1=f->Fid[1]; a2=f->Fid[2];
 if((i2 == l2 && i1 == l1) ||
    (i2 == l1 && i1 == l0) ||
    (i2 == l0 && i1 == l2))
{
   f->Vid[2]=i1; f->Vid[1]=i2; f->Fid[2]=a0; f->Fid[0]=a2;
   return;
 }
 if((i0 == l2 && i2 == l1) ||
    (i0 == l1 && i2 == l0) ||
    (i0 == l0 && i2 == l2))
 {
   f->Vid[0]=i2; f->Vid[2]=i0; f->Fid[0]=a1; f->Fid[1]=a0;
   return;
 }
 if((i1 == l1 && i0 == l0) ||
    (i1 == l2 && i0 == l1) ||
    (i1 == l0 && i0 == l2))
 {
   f->Vid[1]=i0; f->Vid[0]=i1; f->Fid[1]=a2; f->Fid[2]=a1;
   return;
 }
 return;
}

static void IdentifyConnected(long last_one_set, long set_this_one)
{
 if((LocalFacetList+set_this_one)->Sid >= 0)return;
 (LocalFacetList+set_this_one)->Sid=count_of_pieces;
 //
 // If this is not the first facet in a piece then force it to take
 // the same orientation.
 //
 if(last_one_set >= 0)MakeSameOrientation(last_one_set,set_this_one);
 else count_of_pieces++;
 last_one_set=set_this_one;
 //
 // Proceed to repeat this process in the neighbouring facets
 //
 if((LocalFacetList+set_this_one)->Fid[0] >= 0)
   IdentifyConnected(last_one_set,(LocalFacetList+set_this_one)->Fid[0]);
 if((LocalFacetList+set_this_one)->Fid[1] >= 0)
   IdentifyConnected(last_one_set,(LocalFacetList+set_this_one)->Fid[1]);
 if((LocalFacetList+set_this_one)->Fid[2] >= 0)
   IdentifyConnected(last_one_set,(LocalFacetList+set_this_one)->Fid[2]);
 return;
}
