
// Function to build a Delaunay triangulation from a collection of
// vertices.

//  The data types (VERTEX, EDGE and FACET) will need to be modified
//  to fit whatever framework you are using.

typedef float VECTOR[3];

typedef struct tagVERTEX {
 // add other members as required
 VECTOR p;
} VERTEX;

typedef struct tagEDGE {
 long Vid[2];     // identity of vertices to which the edge is attached
 long Fid[2];     // list to identify the facets on either side of edge
 //
 // add other members as required
} EDGE;

typedef struct tagFACET {
 long Vid[3];     // identity of vertices to which facet is attached
 long Fid[3];     // list to indentity facets with an edge in common
 long Eid[3];     // list of the identity of edges surrounding the facet
 //
 // add other members as required
} FACET;


////////// The following functions will need to be provided to fulfil the
////////// described action.

// create an edge in database returns the identity of the edge
extern long   newE(long , long );
// create a facet in database return the identity of the facet
extern long   newF(long , long , long );
// create a vertex in the database return the identity of the vertex
extern long  newV(VECTOR );
// remove the vertex from the vertex list at element "i"
extern void  eraseV(long i);
// scale the vector A with the scalar s,   R = sA
extern void   ScaleVector(float s,VECTOR A, VECTOR R);
// copy a vector from Source to Destination
extern void   CopyVector(VECTOR Source, VECTOR Destination);
// add two vectors R = A + B
extern void AddVectors(VECTOR A, VECTOR B, VECTOR R);
// subtract two vectors R = A - B
extern void SubtractVectors(VECTOR A, VECTOR B, VECTOR R);
// make vector V unit length
extern void   Normalize(VECTOR V);
// R = vector cross product of U x V
extern void   Cross(VECTOR U, VECTOR V, VECTOR R);
// return the dot product of A.B
extern float  Dot(VECTOR A, VECTOR B);

// Provided in file CHAP1_0.CPP
long PointInsidePolygon(VECTOR vi,  // point to test
                        VECTOR p0,  // polygon vertex 1
                        VECTOR p1,  // polygon vertex 2
                        VECTOR p2,  // polygon vertex 2
                        long *code, // intersect where
                        long *id);  // edge id - if intersect edges

///////// The following are standard functions and definitions that
///////// are usually provided as part of a compilers library.
///////// Substitute then with the ones you wish to use.

#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
#define MAX_FLOAT 1.e30
#define NULL 0
extern void *malloc(long X);      // allocate X bytes, return pointer to it.
// update memory block pointed to by P so that it has X bytes.
extern void *realloc(void *P, long X);
extern void free(void *P);        // release memory block pointed to by P
extern float sqrt(float F);       // return the square root of F

////
//// functions uses internally by this algorithm
////

static long TestForSwap(VECTOR, long, long, long);
static long IdentifyEdge(long , long ,
                         long ,
                         long *, long *, long *, long *);
static void UpdateDelaunayTriangulation(VECTOR);

//
// Global variables:  The following variable are defines globally because
// the number of facets and edges are changed by calls to the newE, and newF
// functions, and after each call to UpdateDelaunayTriangulation the
// variables nFacets and nEdges will be different. Note that it is also
// possible that the pointers to the "edge" and "facet" lists may have
// been relocated after each call to newE and newF
//
// Initially nVertices and VertexList should contain the points to be
// inserted into the triangulation. nEdges and nFacets should both be zero.
//

extern  long   nVertices;         // Number of vertices
extern  VERTEX *VertexList;       // List of vertex structures
extern  long   nEdges;            // Number of edges
extern  EDGE   *EdgeList;         // List of edge structures
extern  long   nFacets;           // Number of facets
extern  FACET  *FacetList;        // List of facet structures



// Function to build a Delaunay triangulation from a colection of
// vertices.

void BuildDelaunayTriangulation(
  long   plane)             // which plane l=2 -> xy plane
{
 VERTEX *vpp;
 long w0,w1,w2;
 long i,j,k,l,m,f;
 VECTOR du,p0,p1,p2,normal;
 float xc, d_max_x = -MAX_FLOAT, d_min_x = MAX_FLOAT;
 float yc, d_max_y = -MAX_FLOAT, d_min_y = MAX_FLOAT;
 float zc, d_max_z = -MAX_FLOAT, d_min_z = MAX_FLOAT;
 float radius;
 //
 // Do not do it if there are no vertices or any facets already exist
 //
 if(nVertices == 0)return;
 if(nFacets > 0)return;
 //
 // Make the supertriangle
 //
 l=plane;
 vpp=VertexList;
 for(i=0;i<nVertices;i++)
 {
   if(l == 2){j=0; k=1; m=2;}
   if(l == 0){j=1; k=2; m=0;}
   if(l == 1){j=2; k=0; m=1;}
   d_min_x=min(d_min_x,vpp->p[j]);
   d_max_x=max(d_max_x,vpp->p[j]);
   d_min_y=min(d_min_y,vpp->p[k]);
   d_max_y=max(d_max_y,vpp->p[k]);
   d_min_z=min(d_min_z,vpp->p[m]);
   d_max_z=max(d_max_z,vpp->p[m]);
 }
 vpp++;
 xc = (d_min_x/2 + d_max_x/2);
 yc = (d_min_y/2 + d_max_y/2);
 zc = (d_min_z/2 + d_max_z/2);
 radius = max((d_max_x - d_min_x),(d_max_y - d_min_y));
 //
 // "normal" is the normal to the plane in which the points lie
 //
 normal[0]=normal[1]=normal[2]=0.0;
 if(l == 2){i=0; j=1; k=2; normal[2]=1.0;}
 if(l == 0){i=1; j=2; k=0; normal[0]=1.0;}
 if(l == 1){i=2; j=0; k=1; normal[1]=1.0;}
 p0[i]=xc;
 p0[j]=yc+2.0*radius;
 p0[k]=zc;
 p1[i]=xc-1.732*radius;
 p1[j]=yc-radius;
 p1[k]=zc;
 p2[i]=xc+1.732*radius;
 p2[j]=yc-radius;
 p2[k]=zc;
 w0=newV(p0);
 w1=newV(p1);
 w2=newV(p2);
 newE(w0,w1);
 newE(w1,w2);
 newE(w2,w0);
 f=newF(w0,w1,w2);
 //
 // The supertriangle has no adjacent faces
 //
 for(i=0;i<3;i++){
   FacetList[f].Fid[i] = -1;
   FacetList[f].Eid[i] =  i;
 }
 //
 // insert all the vertices (note that "nVertices" now included
 // the three vertices added to make the supertriangle) at the end of the
 // list the index to those vertices is stored in w0, w1, and w2
 // So don't try to insert those!
 //
 for(i=0;i<nVertices-3;i++)
 {
   double mu;
   SubtractVectors(VertexList[i].p,p0,du);
   mu=Dot(du,normal);
   ScaleVector(mu,normal,du);
   AddVectors(VertexList[i].p,du,p2);
   UpdateDelaunayTriangulation(p2);
 }
 //
 // Erase the vertices of the supertriangle, this will automatically
 // remove an edges or facets attached to them as well.
 //
 eraseV(w0);
 eraseV(w1);
 eraseV(w2);
 return;
}

static long TestForSwap(VECTOR p, long v1l, long v2l, long v3l)
//
// test to see if edge common to triangle pair  v1,v2,v3  and v1,v2,p
// should be swapped so as to keep the triangles as near equilateral
// as possible.
//
// It returns "1" if the swap should be made, "0" otherwise
//
{
 VERTEX *v1,*v2,*v3;
 VECTOR vp3,v13,v23,va,vb,v1p,v2p;
 float cosa,cosb,sina,sinb,sinab;
 v1=(VertexList+v1l); v2=(VertexList+v2l); v3=(VertexList+v3l);
 SubtractVectors(v1->p,v3->p,v13); Normalize(v13);
 SubtractVectors(v2->p,v3->p,v23); Normalize(v23);
 SubtractVectors(v1->p,p,v1p); Normalize(v1p);
 SubtractVectors(v2->p,p,v2p); Normalize(v2p);
 cosa=Dot(v13,v23);
 cosb=Dot(v1p,v2p);
 if(cosa >= 0.0 && cosb >= 0.0)return 0;
 if(cosa <  0.0 && cosb <  0.0)return 1;
 Cross(v13,v23,va); sina=va[0]*va[0]+va[1]*va[1]+va[2]*va[2];
 if(sina < 1.e-10)return 0; sina = sqrt(sina);
 Cross(v2p,v1p,vb); sinb=vb[0]*vb[0]+vb[1]*vb[1]+vb[2]*vb[2];
 if(sinb < 1.e-10)return 0; sinb = sqrt(sinb);
 sinab=sina*cosb+sinb*cosa;
 if(sinab < 0.0)return 1;
 return 0;
}

static long IdentifyEdge(
 long v1, long v2,  // vertex indicies that identify the common edge for
 long k,            // facet (k) and the facet with the point of insertion
                    // as one of its vertices.
 long *fa1,         // fa1 and fa2 are returned values that identify the
 long *fa2,         // facets adjacent to facet k along the "other" edges
                    // i.e. not the edge between v1 - v2.
 long *e1,          // e1 and e2 are returned values that identify the edges
 long *e2)          // common between facets k and *fa1 and between k and *fa2
//
// This function takes a the "opposite" facet "k" and returns the ID of
// the vertex in "k" which is neither "v1" or "v2". The function also returns
// some adjacency information about facet "k" which is required by the
// caller to update the adjacency data should a swap of the common edge be
// necessary.
//
// For a definition of the term "opposite" see the comments in the function
// UpdateDelaunayTriangulation().
//
{
 FACET *f;
 f=(FacetList+k);
 if((f->Vid[0] == v1 && f->Vid[1] == v2) ||
    (f->Vid[0] == v2 && f->Vid[1] == v1))
 {
   *fa1=FacetList[k].Fid[1]; *fa2=FacetList[k].Fid[2];
   *e1=FacetList[k].Eid[1];  *e2=FacetList[k].Eid[2];
   return f->Vid[2];
 }
 else if((f->Vid[1] == v1 && f->Vid[2] == v2) ||
         (f->Vid[1] == v2 && f->Vid[2] == v1))
 {
   *fa1=FacetList[k].Fid[2]; *fa2=FacetList[k].Fid[0];
   *e1=FacetList[k].Eid[2];  *e2=FacetList[k].Eid[0];
   return f->Vid[0];
 }
 else if((f->Vid[2] == v1 && f->Vid[0] == v2) ||
         (f->Vid[2] == v2 && f->Vid[0] == v1))
 {
   *fa1=FacetList[k].Fid[0]; *fa2=FacetList[k].Fid[1];
   *e1=FacetList[k].Eid[0];  *e2=FacetList[k].Eid[1];
   return f->Vid[1];
 }
 return -1;
}

static void UpdateDelaunayTriangulation(
 VECTOR p2)    // point to be inserted in existing Delaunay triangulation
{
 long code,id;
 long j,k,l,m,t,nn,*Dstack=NULL,Dss=0,fa1,fa2;
 long vp,w0,w1,w2,vo,vt;
 long e0,e1,e2,ep;
 //
 // Adding new facets will change the number of facets "nFacets" but this
 // is only done when facet in which the point "p2" lies is found. When the
 // intersection point has been processed the function returns without
 // working on the rest of the ffacet list. So although it is not good
 // practice to have the upper limit of a loop adjusted within the loop
 // in this case it does no harm.
 //
 for(j=0;j<nFacets;j++)
 {
   //
   // find which facet the point "p2" lies inside
   //
   w0=FacetList[j].Vid[0];
   w1=FacetList[j].Vid[1];
   w2=FacetList[j].Vid[2];
   if(PointInsidePolygon(p2,
                         (VertexList+w0)->p,
                         (VertexList+w1)->p,
                         (VertexList+w2)->p,
                         &code,&id))
   {
     //
     // "p2" lies inside facet "j" so we need to add at vertex at "p2"
     // make two new facets, three new edges and update facet "j" itself
     // to reflect the fact that one of its vertices is now the vew vertex
     // at "p2".
     //
     vp=newV(p2);
     e0=newE(vp,w0);
     e1=newE(vp,w1);
     e2=newE(vp,w2);
     newF(vp,w0,w1);
     newF(vp,w2,w0);
     FacetList[j].Vid[0]=vp;              // modify facet "j"
     nn = nFacets;
     //
     // Now the adjacency information needs to be built for the new
     // facets and the old adjacency information needs to be modified
     // to reflect the new network topology.
     //
     // Adjacency in new facets first.
     //
     FacetList[nn-2].Fid[0]=nn-1;
     FacetList[nn-2].Fid[1]=FacetList[j].Fid[0];
     FacetList[nn-2].Fid[2]=j;
     FacetList[nn-1].Fid[0]=j;
     FacetList[nn-1].Fid[1]=FacetList[j].Fid[2];
     FacetList[nn-1].Fid[2]=nn-2;
     //
     //  Change adjacency in the facets adjacent to facet "j"
     //
     if((k=FacetList[j].Fid[0]) >= 0)
     {
       for(l=0;l<3;l++)if(FacetList[k].Fid[l] == j)FacetList[k].Fid[l]=nn-2;
     }
     if((k=FacetList[j].Fid[2]) >= 0)
     {
       for(l=0;l<3;l++)if(FacetList[k].Fid[l] == j)FacetList[k].Fid[l]=nn-1;
     }
     //
     //  Change adjacency in facet "j"
     //
     FacetList[j].Fid[0]=nn-2;
     FacetList[j].Fid[2]=nn-1;
     //
     // And the edge adjacency too
     //
     FacetList[nn-2].Eid[0]=e0;
     FacetList[nn-2].Eid[1]=FacetList[j].Eid[0];
     FacetList[nn-2].Eid[2]=e1;
     FacetList[nn-1].Eid[0]=e2;
     FacetList[nn-1].Eid[1]=FacetList[j].Eid[2];
     FacetList[nn-1].Eid[2]=e0;
     FacetList[j].Eid[0]=e1;
     FacetList[j].Eid[2]=e2;
     //
     // Now the Delaunay updating can begin - push the new + modified
     // facets on to the stack. Do this if there are any facets "opposite"
     // the inserted point "p0". A facet is "opposite" facet "j" if it shares
     // a comon edge (equivalently has two vertices the same as facet "j")
     // that is NOT attached to the vertex at "p0".
     // Since a consistent labelling is being used: when a point is inserted
     // the facets that are built the record of which vertices a facet is
     // attached to always starts with the vertex associated with the
     // point of insertion. The first edge (edge 0) then joins vertices
     // FacetList[j].Vid[0] to FacetList[j].Vid[1]. Thus the entry in the
     // list of edges round facet "j" FacetList[j].Eid[0] is the edge
     // joining FacetList[j].Vid[0] to FacetList[j].Vid[1].
     // The facets adjacent to facet "j" are listed in the same order thus,
     // the neighbouring facet FacetList[j].Fid[0] is the one whose common
     // edge is the one between FacetList[j].Vid[0] to FacetList[j].Vid[1].
     //
     // The implication of this is that the facet "opposite" the insertion
     // point is always FacetList[j].Fid[1] which joins vertices
     // FacetList[j].Vid[1] to FacetList[j].Vid[2]
     //
     if(Dstack == NULL)
     {
       if((Dstack=(long *)malloc(sizeof(long)*(3)))
                 == NULL)return;
     }
     else {
       if((Dstack=(long *)realloc(Dstack,sizeof(long)*(Dss+3)))
                 == NULL)return;
     }
     if(FacetList[j].Fid[1] >= 0)   { Dstack[Dss] = j;    Dss++;}
     if(FacetList[nn-2].Fid[1] >= 0){ Dstack[Dss] = nn-2; Dss++;}
     if(FacetList[nn-1].Fid[1] >= 0){ Dstack[Dss] = nn-1; Dss++;}
     //
     // With the stack initialized the "shuffling" process can begin
     // continue with it until the stack is empty (until the number of
     // elements on it Dss = 0)
     //
     while(Dss > 0)
     {
       //
       // pop a facet off the stack for consideration and get the identity
       // of the facet "opposite" to it. "j" is the ID of the facet being
       // considered and "k" is the ID of the "opposite" facet.
       //
       Dss--;
       j=Dstack[Dss]; k=FacetList[j].Fid[1];
       if(Dss > 0)
       {
         if((Dstack=(long *)realloc(Dstack,sizeof(long)*(Dss))) == NULL)
         {
           return;
         }
       }
       //
       // Determine some information about the "opposite" facet. To be
       // able to test whether the common edge between facets "j" and "k"
       // should be swapped we need to know the ID of the vertex in "k"
       // that is NOT common to facet "j". If a swap occurs we will need to
       // fix up adjacency in facets adjacent to facet "k" as well so
       // function IdentifyEdge() provides this information. It return the
       // ID of the required vertex and the ID's of the facets (fa1,fa2)
       // which are adjacent to "k" but are NOT facet "j". (e1 and e2 are
       // the edges common to facet "k" and facets "fa1" and "fa2".
       /*
                 ___________
               /!\   fa1   !
             /  !  \      !
           /    !    \   !
     p0 ./   j  !      \. vo      ( facet "k" is adjacent to facet "j" and
        !\      !  k   /!         ( opposite vertex at p0
       !   \    !    /   !
      !   m  \  !  / fa2  !
     !_________\!/________ !

       */
       //
       if((vo=IdentifyEdge(FacetList[j].Vid[1],
                           FacetList[j].Vid[2],
                           k,
                           &fa1,&fa2,
                           &e1,&e2)) < 0)return;
       //
       // Do we need to swap the edge of facet "j" that is opposite "p0"
       //
       if(TestForSwap(p2,
                      FacetList[j].Vid[1],
                      FacetList[j].Vid[2],
                      vo))
       {
         //
         // YES !  "e0" is that edge. There is quite a bit of tidying up!
         //
         e0=FacetList[j].Eid[1];
         //
         // Swapping to the scheme below means that facets "j", "k", "fa1"
         // and  "m" all have to have their adjacency fixed up.
         /*
                 ___________
               / \   fa1   !
             /     \      !
           /    j    \   !
     p0 ./_____________\. vo      ( facet "k" is adjacent to facet "j" and
        !\             /!         ( opposite vertex at p0
       !   \    k    /   !
      !   m  \     / fa2  !
     !_________\ /________ !

         */
         //
         // change adjacency in adjacent facets "m" and "fa1"
         //
         if((m=FacetList[j].Fid[2]) >= 0)
         {
           for(l=0;l<3;l++)
             if(FacetList[m].Fid[l] == j)FacetList[m].Fid[l]=k;
         }
         if(fa1 >= 0)
         {
           for(l=0;l<3;l++)
             if(FacetList[fa1].Fid[l] == k)FacetList[fa1].Fid[l]=j;
         }
         //
         //  reverse edge
         //
         EdgeList[e0].Vid[0]=vp;
         EdgeList[e0].Vid[1]=vo;
         //
         //  update facet j
         //
         vt=FacetList[j].Vid[2];
         FacetList[j].Vid[2]=vo;
         FacetList[j].Fid[1]=fa1;
         t=FacetList[j].Fid[2];
         FacetList[j].Fid[2]=k;
         FacetList[j].Eid[1]=e1;
         ep=FacetList[j].Eid[2];
         FacetList[j].Eid[2]=e0;
         //
         // update facet k
         //
         FacetList[k].Vid[0]=vp;
         FacetList[k].Vid[1]=vo;
         FacetList[k].Vid[2]=vt;
         FacetList[k].Fid[0]=j;
         FacetList[k].Fid[1]=fa2;
         FacetList[k].Fid[2]=t;
         FacetList[k].Eid[0]=e0;
         FacetList[k].Eid[1]=e2;
         FacetList[k].Eid[2]=ep;
         //
         // Since a swap has been done p0 is now a vertex of two modified
         // facets "j" and "k" which have "opposite" facets "fa1" and "fa2".
         // This new configuration itself needs to be updated to a Delaunay
         // triangulation, so push facets "j" and "k" onto the stack and
         // go back and pop another off until the stack is empty.
         if((Dstack=(long *)realloc(Dstack,sizeof(long)*(Dss+2))) == NULL)
         {
           return;
         }
         //
         // puson onto stack if there are any "opposite" facets to consider
         //
         if(FacetList[j].Fid[1] >= 0){Dstack[Dss] = j; Dss++;}
         if(FacetList[k].Fid[1] >= 0){Dstack[Dss] = k; Dss++;}
       }
     }
     //
     // At this point the stack is now empty!
     //
     if(Dstack != NULL)free(Dstack);
     return;
   }
 }
 //
 // If execution gets here then "p0" was not inserted in any triangle.
 //
 return;
}
