//
// These functions implement the Boolean algorithm for two data sets. Each vertex
// data structure contains a flag which is set to either TRUE or FALSE.
// If it is TRUE then any edges or facets attached to those vertices are regarded
// as part of the tool. If it is FALSE then any edges or facets attached to those
// vertices are the work edges and work facets.
//

// The data types (VERTEX, EDGE and FACET) will need to be modified
// to fit whatever framework you are using.

typedef float VECTOR[3];

typedef struct tagVERTEX {
 // add other members as required
 VECTOR p;	  // position of vertex
 int flag;	  // flag to indicate which to data set the vertex belongs
 int id;	  // identity tag - uses by the "adjacency finding" algorithm
} VERTEX;

typedef struct tagEDGE {
 long Vid[2];     // identity of vertices to which the edge is attached
 //
 // add other members as required
} EDGE;

typedef struct tagFACET {
 long Vid[3];     // identity of vertices to which facet is attached
 //
 // add other members as required
} FACET;

////////// The following functions will need to be provided to fulfil the
////////// described action.

// create an edge in database returns the identity of the edge
extern long   newE(long , long );
// create a facet in database return the identity of the facet
extern long   newF(long , long , long );
// create a vertex in the database return the identity of the vertex
extern long  newV(VECTOR );
// remove the vertex from the vertex list at element "i"
extern void  eraseV(long i);
// remove the vertices from the list that have their flag value set to TRUE;
extern void  eraseFlaggedVertices();
// scale the vector A with the scalar s,   R = sA
extern void   ScaleVector(float s,VECTOR A, VECTOR R);
// copy a vector from Source to Destination
extern void CopyVector(VECTOR Source, VECTOR Destination);
// add two vectors R = A + B
extern void AddVectors(VECTOR A, VECTOR B, VECTOR R);
// subtract two vectors R = A - B
extern void SubtractVectors(VECTOR A, VECTOR B, VECTOR R);
// make vector V unit length return 1 if OK return 0 if vector is zero
extern int  Normalize(VECTOR V);
// R = vector cross product of vectors   U x V
extern void   Cross(VECTOR U, VECTOR V, VECTOR R);
// return the dot product of vectors     A.B
extern float  Dot(VECTOR A, VECTOR B);


// Provided in file CHAP1_0.CPP
long PointInsidePolygon(VECTOR vi,  // point to test
                        VECTOR p0,  // polygon vertex 1
                        VECTOR p1,  // polygon vertex 2
                        VECTOR p2,  // polygon vertex 2
                        long *code, // intersect where
                        long *id);  // edge id - if intersect edges

///////// The following are standard functions and definitions that
///////// are usually provided as part of a compilers library.
///////// Substitute then with the ones you wish to use.

#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
#define MAX_FLOAT 1.e30
#define NULL      0

// allocate X bytes, return pointer to it.
extern void *malloc(long X);
// update memory block pointed to by P so that it has X bytes.
extern void *realloc(void *P, long X);
// release memory block pointed to by P
extern void free(void *P);
// return the square root of F
extern float sqrt(float F);
// return the floating point absolute value	of  "x"
extern float fabs(float x);

//
// Global variables:  The following variable are defines globally because
// the number of facets and edges are changed by calls to the newV, newE,
// and newF, and after calls to various functions in the algorithm
// variables nFacets and nEdges will be different. Note that it is also
// possible that the pointers to the "edge" and "facet" lists may have
// been relocated after each call to newE and newF
//
// Initially nVertices and VertexList should contain the points to be
// inserted into the triangulation. nEdges and nFacets should both be zero.
//

extern  long   nVertices;         // Number of vertices
extern  VERTEX *VertexList;       // List of vertex structures
extern  long   nEdges;            // Number of edges
extern  EDGE   *EdgeList;         // List of edge structures
extern  long   nFacets;           // Number of facets
extern  FACET  *FacetList;        // List of facet structures

//
// Internal data structures used to to partition the model into two
// data sets, that acting as the part of the model being cut and the
// part doing the cutting.
//

typedef struct tagWORKEDGE {
  long e;     // Identily which edge in the main data structures this edge is
  long ea[2]; // faces adjacent to this edge (assuming no bifurcation)
  long  t;    // flag to id whether the edge is an original or has been added
              // by the boolean procedure  0=original 1=added in pass
  long  b;    // is this edge a boundary where a cut will need to be made
              // that is does it lie in the plane of tool face
} WORKEDGE;

typedef struct tagWORKFACE {
  long f;
  long ea[3]; // id of edges round face                                  
  long fa[3]; // id if faces adjacent to "f"  -1 => no adjacent face     
  int  id;    // identify which of the cut pieces this face belongs to   
  VECTOR n;
} WORKFACE;

typedef struct tagTOOLEDGE {
  long e;        // Identity of the edge in the main data structure
  VECTOR p0,p1;  // Vectors specifying  the points at the ends of the edge
} TOOLEDGE;

typedef struct tagTOOLFACE {
  long f;
  VECTOR p0,p1,p2;
  VECTOR n;
} TOOLFACE;

//
// This structure is used to record information on which edges and which faces are
// adajent to a vertex.
//

typedef struct tagVTXADJ {
  long nf,ne;
  long *flist;
  long *elist;
} VTXADJ;

//
// The following numbers are factors that are required to avoid numerical problems
// due to the "inaccuracy" of commputer arithmetic.
//

#define NUMERIC_FACTOR       2.0
#define NUMERIC_FACTOR1      0.0005
#define NUMERIC_FACTOR2      0.9995
#define NUMERIC_FACTOR3     -1.00000
#define NUMERIC_FACTOR8      1.00000
#define NUMERIC_FACTOR5     -0.001
#define NUMERIC_FACTOR6      1.001
#define NUMERIC_FACTOR9    100.0
#define NUMERIC_FACTORA    100.0
#define NUMERIC_FACTORB     10.0
#define NUMERIC_FACTORC    100.0

////
//// The following are functions and definitions used internally by this algorithm
////

#define FAIL      0
#define FALSE     0
#define TRUE      1
#define UNDECIDED 2

static long TestForSwap(VECTOR, long, long, long);
static long IdentifyEdge(long , long ,
                         long ,
                         long *, long *, long *, long *);
static int FindAdjFaces(WORKFACE *, long , WORKEDGE *, long);		
static long EdgeThroughFace(VECTOR, VECTOR, VECTOR, VECTOR,
                            VECTOR, VECTOR, VECTOR, long);
static long BoolAddVertex(VECTOR, long);
static WORKFACE *MakeFacet(long, long, long, long, 
                           WORKFACE *, long *);
static WORKFACE *UpdateDelaunayTriangulation(VECTOR, WORKFACE *, long, long,
                                             long *, WORKFACE *, long *, WORKEDGE **);
static int TestForFacetSwitch(long, WORKFACE *, WORKEDGE *);
static WORKFACE *DivideEdge(VECTOR, long, WORKFACE *, WORKEDGE **,
                            long *, long *, long *);
static void FlipCommonEdge(long, WORKFACE *, long,
                                  WORKEDGE *, long);
static void SplitAlongBorder(WORKFACE *, long, WORKEDGE *, long);

//
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
// This is the main entry point of the algorithm
//

void BooleanCutAndSplit(void)
{
 //
 // The first action is to make lists of vertices and edges that are used
 // by the algorithm. These lists are independent of the main data structures
 // because  1) The main data structures are updated as the new vertices
 //             edges and facets are created.
 //          2) The second phase of the algorithm requires that the TOOL and
 //             WORK lists are swapped. Thhey must be swapped in their
 //             original form.
 //          3) You might want to carry out the boolean operation on only
 //             a subset of the full data structures.
 //
 // The list are the  WORKFACE list
 //                   WORKEDGE list
 //
 //                   TOOLFACE list
 //                   TOOLEDGE list
 //
 // At any instant on the algorithm the WORKFACES and WORKEDGES are being
 // modified WORKFACES and WORDEDGES are being divided due to either.
 //    1) A TOOL EDGE penetrates a WORK FACE
 //    2) A WORK EDGE penetrates a TOOL FACE
 //
 WORKEDGE *we,*wee;
 TOOLEDGE *te,*xe;
 WORKFACE *w;
 TOOLFACE *t,*x;
 FACET *ff1;
 long f1;
 VERTEX *vpp,*v0,*v1,*v2;
 long vp,w0,w1,w2;
 EDGE *ep;
 long xn;
 long e0,e1,e2;
 long c,i,j,k,l,nwf,nns,passID,LastVtxID,StartVertex;
 VECTOR du,dv,n,pa,pb,pc,p0,p1,p2;
 long n_tool_face=0,
      n_work_face=0,
      n_tool_edge=0,
      n_work_edge=0,
      x_tool_face=0,
      x_tool_edge=0;
 unsigned char first_time=1;
 if((ff1=FacetList) == NULL)return;
 StartVertex=nVertices;
 vpp=VertexList; for(i=0;i<nVertices;i++)
 {
   vpp->id=0; vpp++;
 }
 t=NULL; w=NULL; te=NULL; we=NULL; x=NULL; xe=NULL;
 for(c=0;c<nFacets;c++)
 { 
   //
   // find then number of work and tool facets
   //
   v0=(VertexList+ff1->Vid[0]); 
   v1=(VertexList+ff1->Vid[1]); 
   v2=(VertexList+ff1->Vid[2]);
   if     (v0->flag == TRUE &&
           v1->flag == TRUE &&
           v2->flag == TRUE )n_tool_face++;
   else if(v0->flag == FALSE &&
           v1->flag == FALSE &&
           v2->flag == FALSE)n_work_face++;
   ff1++;
 }
 //
 //	Allocate memory for the work and tool facets and space to 
 // put a copy of the original work facets for use in the second
 // pass when they become the tool facets
 //
 if((t=(TOOLFACE *)malloc(n_tool_face*(long)sizeof(TOOLFACE)))
       == NULL)goto ERROR_1;
 if((w=(WORKFACE *)malloc(n_work_face*(long)sizeof(WORKFACE)))
       == NULL)goto ERROR_1;
 //
 //	This is the copy
 //
 x_tool_face=n_work_face;
 if((x=(TOOLFACE *)malloc(x_tool_face*(long)sizeof(TOOLFACE)))
       == NULL)goto ERROR_1;
 //
 //	Build the data structures for the tool and work facet lists
 // facet normals are calculated at this time to save on 
 // execution time later.
 //
 n_work_face=0; 
 n_tool_face=0; 
 x_tool_face=0;
 for(ff1=FacetList,c=0;c<nFacets;c++)
 {
  v0=(VertexList+ff1->Vid[0]); 
  v1=(VertexList+ff1->Vid[1]); 
  v2=(VertexList+ff1->Vid[2]);
  CopyVector(v0->p,p0);
  CopyVector(v1->p,p1);
  CopyVector(v2->p,p2);
  SubtractVectors(p1,p0,dv);
  SubtractVectors(p2,p0,du);
  if(v0->flag == TRUE &&
     v1->flag == TRUE &&
     v2->flag == TRUE)
  {
	//
	// tool facets
	//
    Cross(dv,du,n);
    if(Normalize(n) == FAIL)
    {
      SubtractVectors(p0,p1,dv);
      SubtractVectors(p2,p1,du);
      Cross(dv,du,n);
      if(Normalize(n) == FAIL){
        SubtractVectors(p0,p2,dv);
        SubtractVectors(p1,p2,du);
        Cross(dv,du,n);
        if(Normalize(n) == FAIL)goto ERROR_2;
      }
    }
    t[n_tool_face].f=c;
    CopyVector(v0->p,t[n_tool_face].p0);
    CopyVector(v1->p,t[n_tool_face].p1);
    CopyVector(v2->p,t[n_tool_face].p2);
    CopyVector(n,t[n_tool_face].n);
    n_tool_face++;
  }
  else if(v0->flag == FALSE &&
          v1->flag == FALSE &&
          v2->flag == FALSE)
  {
  	//
	// work facets
	//
    w[n_work_face].f=c;
    for(i=0;i<3;i++){
      w[n_work_face].fa[i] = -1;
      w[n_work_face].ea[i] = -1;
    }
    Cross(dv,du,n);
    if(Normalize(n) == FAIL)
    {
      SubtractVectors(p0,p1,dv);
      SubtractVectors(p2,p1,du);
      Cross(dv,du,n);
      if(Normalize(n) == FAIL)
      {
        SubtractVectors(p0,p2,dv);
        SubtractVectors(p1,p2,du);
        Cross(dv,du,n);
        if(Normalize(n) == FAIL)goto ERROR_2;
      }
    }
    CopyVector(n,w[n_work_face].n);
	//
	// and the copy 
	//
    CopyVector(v0->p,x[x_tool_face].p0);  
    CopyVector(v1->p,x[x_tool_face].p1);  
    CopyVector(v2->p,x[x_tool_face].p2);
    CopyVector(n,x[x_tool_face].n);
    x[x_tool_face].f=c; 
    x_tool_face++;
    n_work_face++;
  }
  ff1++;
 }
 //
 // Make the list of edges for the tool and work lists. First count them
 //
 for(ep=EdgeList,c=0;c<nEdges;c++)
 {
   v0=(VertexList+ep->Vid[0]); 
   v1=(VertexList+ep->Vid[1]);
   if     (v0->flag == TRUE && v1->flag == TRUE)n_tool_edge++;
   else if(v0->flag == FALSE && v1->flag == FALSE){
     n_work_edge++;
     x_tool_edge++;
   }
   ep++;
 }
 //
 //	Allocate memory for the tool and work edges and a copy of the
 // original work edges to be used in the secon pass when they 
 // become part of the tool.
 //
 if((te=(TOOLEDGE *)malloc(n_tool_edge*(long)sizeof(TOOLEDGE)))
      == NULL)goto ERROR_1;
 if((we=(WORKEDGE *)malloc(n_work_edge*(long)sizeof(WORKEDGE)))
      == NULL)goto ERROR_1;
 if((xe=(TOOLEDGE *)malloc(x_tool_edge*(long)sizeof(TOOLEDGE)))
      == NULL)goto ERROR_1;
 n_tool_edge=0; 
 n_work_edge=0; 
 x_tool_edge=0;
 //
 //	Build the internal edge data structures
 //
 for(ep=EdgeList,c=0;c<nEdges;c++)
 { 
   v0=(VertexList+ep->Vid[0]); 
   v1=(VertexList+ep->Vid[1]);
   if(v0->flag == TRUE && v1->flag == TRUE)
   {
	 //
	 // tool edges
	 //
     CopyVector(v0->p,te[n_tool_edge].p0);
     CopyVector(v1->p,te[n_tool_edge].p1);
     te[n_tool_edge].e=c;
     n_tool_edge++;
   }
   else if(v0->flag == FALSE && v1->flag == FALSE)
   {
	 //
	 // work edges	(many new work edges will be created by the first pass
	 //
     we[n_work_edge].e=c;    
     we[n_work_edge].t=0;    
     we[n_work_edge].b=0;   
     we[n_work_edge].ea[0] = we[n_work_edge].ea[1] = -1;
     n_work_edge++;
     //
	 //	and the copy for later use
	 //
     CopyVector(v0->p,xe[x_tool_edge].p0);
     CopyVector(v1->p,xe[x_tool_edge].p1); 
     xe[x_tool_edge].e=c;                  
     x_tool_edge++;
   }
   ep++;
 }
 //
 // At this point the appropriate lists have been build and the
 // algorithm can proceed as described in the text provided there
 // are some facets in the WORK and TOOL lists.
 //
 if(n_work_face == 0 || n_tool_face == 0)goto FINISHED;
 passID=1;
 //
 //	Return to this point when the tool and work lists are swapped so
 // that the work facets will then cut into the tool
 //
 //
 PHASE_POINT:
 //
 //	Start the first pass: Make intersection in WORK facets due to TOOL edges
 //
 nwf=n_work_face;
 for(i=0;i<nwf;i++)
 { 
   //
   // Consider each work facet
   //
   f1=w[i].f;
   w0=(FacetList+f1)->Vid[0]; 
   w1=(FacetList+f1)->Vid[1]; 
   w2=(FacetList+f1)->Vid[2];
   CopyVector((VertexList+w0)->p,pa);
   CopyVector((VertexList+w1)->p,pb);
   CopyVector((VertexList+w2)->p,pc);
   nns=n_work_face;
   //
   // Check each edge "j" into facet "i" (part of original workfacets)
   //
   for(j=0;j<n_tool_edge;j++)
   {
     if(EdgeThroughFace(pa,pb,pc,w[i].n,te[j].p0,te[j].p1,p2,0))
     {
       //
	   // Edge "j" pierces facet "i"
       if(nns == n_work_face)
       {
         //
		 //	It occurs in one of the original work facets (not added
		 // by Delaunay procedure)
		 //
         //
		 //	Add a vertex at the point of intersection
		 //
         vp=BoolAddVertex(p2,FALSE);
         //
		 //	Allow for an extra three edges
		 //
         if((wee=(WORKEDGE *)realloc(we,(n_work_edge+3)*(long)sizeof(WORKEDGE)))
               == NULL)goto ERROR_1;
         we=wee;
		 //
		 //	Create the three edges and update the adjacency information
		 //
         newE(vp,w0); 
         e0=n_work_edge;   
         we[e0].e=nEdges-1;  
         we[e0].t = -1;  
         we[e0].b=0;
         we[e0].ea[0] = we[e0].ea[1] = -1;
         newE(vp,w1);
         e1=n_work_edge+1; 
         we[e1].e=nEdges-1;  
         we[e1].t = -1;  
         we[e1].b=0;
         we[e1].ea[0] = we[e1].ea[1] = -1;
         newE(vp,w2);
         e2=n_work_edge+2; 
         we[e2].e=nEdges-1;  
         we[e2].t = -1;  
         we[e2].b=0;
         we[e2].ea[0] = we[e2].ea[1] = -1;
         n_work_edge+=3;
		 //
		 //	Make two new facets and modify the original facet so that
		 // we have three facets where there was originally one
		 //
         if((w=MakeFacet(vp,w0,w1,i,w,&n_work_face))==NULL)goto ERROR_1;
         if((w=MakeFacet(vp,w2,w0,i,w,&n_work_face))==NULL)goto ERROR_1;
         //
		 //	Modify original facet
		 //
         (FacetList+f1)->Vid[0]=vp;
         //
         // Modify the adjacency data  - see the example on Delaunay 
		 // triangulation
		 //
         w[n_work_face - 2].fa[0]=n_work_face-1;
         w[n_work_face - 2].fa[1]=w[i].fa[0];
         w[n_work_face - 2].fa[2]=i;
         w[n_work_face - 1].fa[0]=i;
         w[n_work_face - 1].fa[1]=w[i].fa[2];
         w[n_work_face - 1].fa[2]=n_work_face-2;
         //
         // change adjacency in adjacent 
		 //
         if((k=w[i].fa[0]) >= 0)
         {
           for(l=0;l<3;l++)if(w[k].fa[l] == j)w[k].fa[l]=n_work_face-2;
         }
         if((k=w[i].fa[2]) >= 0)
         {
           for(l=0;l<3;l++)if(w[k].fa[l] == j)w[k].fa[l]=n_work_face-1;
         }
         w[i].fa[0]=n_work_face-2;
         w[i].fa[2]=n_work_face-1;
         w[n_work_face - 2].ea[0]=e0;
         w[n_work_face - 2].ea[1]=w[i].ea[0];
         w[n_work_face - 2].ea[2]=e1;
         w[n_work_face - 1].ea[0]=e2;
         w[n_work_face - 1].ea[1]=w[i].ea[2];
         w[n_work_face - 1].ea[2]=e0;
         w[i].ea[0]=e1;
         w[i].ea[2]=e2;
       }
       else
       {
       	 //
		 //	Complete the Delaunay procedure - updating - via recursion
		 // the triangulation as appropriate (see the Delaunay example)
		 //
         // A vertex is inserted at p2 in wf(i) or any of it's subdivisions.
		 //
         if((w=UpdateDelaunayTriangulation(p2,&w[i],i,nns,&n_work_face,w,
                &n_work_edge,&we)) == NULL)goto ERROR_1;
       }
     }
   }
 }
 //
 //	This completes the first pass!
 //
 //
 //	Before the first pass determine the adjaceny information for facets and
 // edges.  This procedure  
 //
 LastVtxID=FindAdjFaces(w,n_work_face,we,n_work_edge);
 //
 //	Begin the second pass which has an interative element which attempts to 
 // keep to a minimum the number edges added in this phase. This phase 
 // checks the intersection of edges of the work facets into tool facets.
 //	The first thing that must be done is to check and label the WORKFACE edges
 // since they cannot be moved by the interative procedure. (Moving them would
 // destroy the "shape" of the model.
 //
 for(i=0;i<n_tool_face;i++)
 { 
   //
   // check the original WORKFACE edges that pierce tool facet "i"
   //
   CopyVector(t[i].p0,pa);
   CopyVector(t[i].p1,pb);
   CopyVector(t[i].p2,pc);
   CopyVector(t[i].n,n);
   nns=n_work_edge;
   //
   // Do each work edge "j"
   //
   for(j=0;j<nns;j++)
   {
     //
	 //	only original WORKFACE edges
	 //
     if(we[j].t != 0)continue;  
     ep=(EdgeList+we[j].e);
     CopyVector((VertexList+(ep->Vid[0]))->p,p0);
     CopyVector((VertexList+(ep->Vid[1]))->p,p1);
     //
	 //	Check whether and where the edge pierces the facet and
	 // specifically whether it lies in the plane of the facet,
	 // if it does then it is a boundary edge.
	 //
     if((xn=EdgeThroughFace(pa,pb,pc,n,p0,p1,p2,2)) == 1)
     {
       //
	   // The edge pierces the tool facet and therefore has to be 
	   // split into with a vertex at the point where it 
	   //
       if((w=DivideEdge(p2,j,w,&we,&n_work_face,&n_work_edge,&LastVtxID))
             == NULL || we == NULL)goto ERROR_1;
     }
     else if(xn == 2)
     { 
       //
       // "EdgeThroughFace" indicated that edge "j" lies in 
       // the plane of facet "i" and so it is a boundary edge -
	   // raise the boundary flag!
	   //
       we[j].b=1;
     }
   }
 }
 //
 // Go through again all intersections between TOOL facets and WORK
 // edges. This time consider any edges created by the "DivideEdge"
 // function. They can be swapped and the structure iteratively 
 // refined as explained in the text.
 //
 for(i=0;i<n_tool_face;i++)
 { 
   CopyVector(t[i].p0,pa);
   CopyVector(t[i].p1,pb);
   CopyVector(t[i].p2,pc);
   CopyVector(t[i].n,n);
   //
   // Return to this point if iterating, set nns=0 to indicate no 
   // iteration is necessary, if it is nns will be set to 1
   //
   ITERATE: 
   nns=0;
   for(j=0;j<n_work_edge;j++)
   {
     //
     //skip any original or boundary edges which must remain unchanged
     //
     if(we[j].t == 0)continue;  
     if(we[j].b)continue;       
     ep=(EdgeList+we[j].e);
     CopyVector((VertexList+(ep->Vid[0]))->p,p0);
     CopyVector((VertexList+(ep->Vid[1]))->p,p1);
	 //
	 //	Does this edge pierce the facet, again check to see if
	 // it is a boundary.
	 //
     if((xn=EdgeThroughFace(pa,pb,pc,n,p0,p1,p2,2)) == 1)
     {
	   //
	   // The edge pierces the facet but we know that a suitable
	   // iteration we can rearrange the edges so that the edge
	   // does NOT pierce the facet. It either meets the facet at
	   // one end or lies in the plane and is therefore a boundary
	   // edge.	Test to see if the edge can be switched to join 
	   // the other two vertices in the two adjacent facets.
	   //
       if(TestForFacetSwitch(j,w,we))
       {
         //
		 //	Switch the facets and make sure the adjaceny information
		 // is updated accordingly.
		 //
         FlipCommonEdge(j,w,n_work_face,we,n_work_edge);
         CopyVector((VertexList+(ep->Vid[0]))->p,p0);
         CopyVector((VertexList+(ep->Vid[1]))->p,p1);
         //
		 //	Now test the new edge to see it it is a boundary
		 //
         if(EdgeThroughFace(pa,pb,pc,n,p0,p1,p2,2) == 2)
         {
		   //
		   // yes it is! so flag this fact
		   //
           we[j].b=1; 
         }
       }
       else
       {
		 //
		 //	Swapping is not possible at this time so raise a
		 // flag to tell the algorithm that an iteration must be
		 // performed.
		 //
         nns=1; 
       }
     }
     else if(xn == 2)
     { 
       //
       // This is a boundary edge so flag it
	   //
       we[j].b=1;
     }
   }
   //
   // If the local flag "nns" is raised then at least one edge could
   // not be swapped - however others will and so we go back and see
   // if the edge can then be swapped.
   //
   if(nns)goto ITERATE;
 }
 //
 // That the end of the basic Boolean algorithm. The function 
 // "SplitAlongBorder" splits the WORK facets along the boundary
 // which has been now determined.  It does this by making
 // local copies of each piece and then deleting the modified 
 // original.
 //
 SplitAlongBorder(w,n_work_face,we,n_work_edge);
 //
 //	If this is the first pass then swap the TOOL and WORK pieces
 // and repeat the procedure.
 //
 if(first_time)
 {
   //
   // In this second pass through the algorithm the WORK and TOOL lists
   // must be swapped and partially rebuilt 
   //
   first_time=0;
   free(we); 
   we=NULL; 
   n_work_edge=0;
   n_work_edge=n_tool_edge;  
   //
   // make new work edges from the tool edges 
   //
   if((we=(WORKEDGE *)malloc(n_work_edge*(long)sizeof(WORKEDGE)))
           == NULL)goto ERROR_1;
   for(i=0;i<n_tool_edge;i++)
   {
       we[i].e=te[i].e;
       we[i].t=0; we[i].b=0;
       we[i].ea[0] = we[i].ea[1] = -1;
   }
   free(te); 
   te=xe;xe=NULL; 
   n_tool_edge=x_tool_edge; 
   //
   // new tool edges 
   //
   free(w); w=NULL; n_work_face=0;
   n_work_face=n_tool_face;
   if((w=(WORKFACE *)malloc(n_work_face*(long)sizeof(WORKFACE)))
          == NULL)goto ERROR_1;
   //
   // new work faces from the tool faces 
   //
   for(i=0;i<n_tool_face;i++)
   { 
     w[i].f=t[i].f;
     for(j=0;j<3;j++)
     {
       w[i].fa[j] = -1;
       w[i].ea[j] = -1;
     }
     CopyVector(t[i].n,w[i].n);
   }
   free(t);
   //
   // new tool faces 
   //
   t=x;  
   x=NULL;  
   n_tool_face=x_tool_face; 
   vpp=VertexList; for(c=0;c<nVertices;c++)
   {
     if(vpp->flag == TRUE){
       vpp->flag=FALSE;
     }
     else if(vpp->flag == FALSE)
     {
       vpp->flag=TRUE;
     }
     vpp++;
   }
   passID=2;
   //
   // Go back and make the second cut
   //
   goto PHASE_POINT;
 }
 //
 // The Algorithm is complete.
 //
 goto FINISHED;
 ERROR_2:
 //
 // There is a problem with the facet normals - probably too small a triangle
 //
 ERROR_1:
 //
 // Out of memory
 //
 FINISHED:
 //
 // Free up any memory used
 //
 if(t  != NULL)free(t);
 if(w  != NULL)free(w);
 if(te != NULL)free(te);
 if(we != NULL)free(we);
 if(x  != NULL)free(x);
 if(xe != NULL)free(xe);
 return;
}

////////////////////////////////////////////////////////////////////////////////////
///////////// These are the local functions used by the main routine ///////////////

static long EdgeThroughFace(
  VECTOR pa, VECTOR pb, VECTOR pc,
  VECTOR n,
  VECTOR p0, VECTOR p1,
  VECTOR p2,
  long full)
{
//
// Determine whether an edge from "p0" to "p1" passes through a triangular
// facet "pa" "pb" "pc" according to the code "full" - if an intersection
// occurs then "p2" is the point of intersection.
//
// Returns  1 if intersection occurs given a number of different
// conditions, so if:
// full = 0 implies that the edge intersects the tool face internally.
//          At least one vertex of the edge does not lie in the plane
//          of the tool face face.
// full = 1 implies that the edge crosses the tool face but it may cross
//          the face at one of it's edges, however neither of the vertices
//          of the edge lie in/near the plane of the tool face.
// full = 2 implies that if both points lie in plane of tool face then
//          it is checked to see if it is a bounding edge.
//
 long code,id;
 VECTOR da,db,dba;
 double d,d0,d1,d01,mu,mu1;
 SubtractVectors(pa,p0,da);
 SubtractVectors(pa,p1,db);
 d0=Dot(da,n);
 d1=Dot(db,n);
 d01 =d0*d1;
 if(fabs(d0) > NUMERIC_FACTORA || fabs(d1) > NUMERIC_FACTORA)
 {
   //
   // At least one vertex lies out of the tool plane
   //
   if(d01 <= NUMERIC_FACTOR8)
   {
     //
     // the edge crosses the tool face
     //
     SubtractVectors(p1,p0,dba);
     mu1=Dot(dba,n);
     if(fabs(mu1) > NUMERIC_FACTOR9)
     {
       mu=d0/mu1;
       if(full > 0)
       { //
         // not near either end of edge
         //
         if(fabs(d0) < NUMERIC_FACTORA || fabs(d1) < NUMERIC_FACTORA)return 0;
       }
       ScaleVector(mu,dba,dba);
       AddVectors(p0,dba,p2);
       //
       // p2 is the intersection point
       //
       if(PointInsidePolygon(p2,pa,pb,pc,&code,&id))
       {
         //
         // intersection occurs even at the end of the edge
         // i.e. p2  may lie in the plane of the polygon
         //
         if(code == 0 || full > 0)return 1;
       }
     }
   }
 }
 else if(full == 2)
 {
   //
   // both points lie in plane of tool face so
   // check to see if this edge is a boundary edge of the facet
   //
   AddVectors(p0,p1,dba);
   ScaleVector(0.5,dba,dba);
   SubtractVectors(pa,dba,da);
   if(fabs(Dot(da,n)) < NUMERIC_FACTORC)
   {
     //
     //  use the mid-point to see if it is inside the facet
     //
     if(PointInsidePolygon(dba,pa,pb,pc,&code,&id))
     {
       //
       // is inside
       //
       return 2;
     }
   }
 }
 return 0;
}

/////////////////////////////////////////////////////////////
// The following functions provide the Delaunay functionality

static long TestForSwap(VECTOR p, long v1l, long v2l, long v3l)
{
 //
 // For a descripton of the actions of this function see the Delaunay
 // triangulation example code.
 //
 VERTEX *v1,*v2,*v3;
 float cosa,cosb,sina,sinb,sinab;
 VECTOR vp3,v13,v23,va,vb,v1p,v2p;
 v1=(VertexList+v1l); v2=(VertexList+v2l); v3=(VertexList+v3l);
 SubtractVectors(v1->p,v3->p,v13);
 Normalize(v13);
 SubtractVectors(v2->p,v3->p,v23);
 Normalize(v23);
 SubtractVectors(v1->p,p,v1p);
 Normalize(v1p);
 SubtractVectors(v2->p,p,v2p);
 Normalize(v2p);
 cosa=Dot(v13,v23);
 cosb=Dot(v1p,v2p);
 if(cosa >= 0.0 && cosb >= 0.0)return 0;     
 if(cosa <  0.0 && cosb <  0.0)return 1;
 Cross(v13,v23,va);
 sina=va[0]*va[0]+va[1]*va[1]+va[2]*va[2];
 if(sina < 1.e-10)return 0; sina = sqrt(sina);
 Cross(v2p,v1p,vb);
 sinb=vb[0]*vb[0]+vb[1]*vb[1]+vb[2]*vb[2];
 if(sinb < 1.e-10)return 0; sinb = sqrt(sinb);
 sinab=sina*cosb+sinb*cosa;
 if(sinab < 0.0)return 1;
 return 0;
}

static long IdentifyEdge(
  long v1,
  long v2,
  WORKFACE *w,
  long k,
  long *fa1, long *fa2,
  long *e1, long *e2)
{
 //
 // For a descripton of the actions of this function see the Delaunay
 // triangulation example code.
 //
 FACET *f;
 f=(FacetList+w[k].f);
 if((f->Vid[0] == v1 && f->Vid[1] == v2) ||
    (f->Vid[0] == v2 && f->Vid[1] == v1))
 {
   *fa1=w[k].fa[1]; *fa2=w[k].fa[2];
   *e1=w[k].ea[1];  *e2=w[k].ea[2];
   return f->Vid[2];
 }
 else if((f->Vid[1] == v1 && f->Vid[2] == v2) ||
         (f->Vid[1] == v2 && f->Vid[2] == v1))
 {
   *fa1=w[k].fa[2]; *fa2=w[k].fa[0];
   *e1=w[k].ea[2];  *e2=w[k].ea[0];
   return f->Vid[0];
 }
 else if((f->Vid[2] == v1 && f->Vid[0] == v2) ||
         (f->Vid[2] == v2 && f->Vid[0] == v1))
 {
   *fa1=w[k].fa[0]; *fa2=w[k].fa[1];
   *e1=w[k].ea[0];  *e2=w[k].ea[1];
   return f->Vid[1];
 }
 //
 // This point should not be reached - if it is an error condition
 // should be raised.
 //
 return -1;
}

static WORKFACE *UpdateDelaunayTriangulation(VECTOR p2, 
                                             WORKFACE *bf1, long bi,
                                             long nns,
                                             long *nf, WORKFACE *w,
                                             long *nwe, WORKEDGE **we)
{
 //
 //	Insert a vertex into an existing triangle and perform the
 // Delaunay procedure to update the triangular mesh according
 // to the "rules". This procedure is explained in the example
 // code on Delaunay triangulation, the code in this function
 // is basically the same but it uses slightly different notation.
 //
 // 
 long code,id;
 long i,j,k,l,m,t,n,nn,*Dstack=NULL,Dss=0,fa1,fa2;
 WORKEDGE *wee;
 WORKFACE *bf;
 VECTOR pa,pb,pc;
 long vp,w0,w1,w2,vo,vt;
 long f1;
 long e0,e1,e2,ep;
 n= *nf;
 //
 // "nns" is pointer to original top of work facet list before adding new
 //  facets caused by point "p2" being inserted in a work facet.
 //
 //
 // find which face this intersects  
 //
 for(i=nns;i<=n;i++)
 { 
   if(i == n)
     {bf=bf1;   j=bi;} /* use an extra face for the original */
   else      
     {bf=&w[i]; j=i; }
   f1=bf->f;
   w0=(FacetList+f1)->Vid[0]; 
   w1=(FacetList+f1)->Vid[1]; 
   w2=(FacetList+f1)->Vid[2];
   CopyVector((VertexList+w0)->p,pa);
   CopyVector((VertexList+w1)->p,pb);
   CopyVector((VertexList+w2)->p,pc);
   if(PointInsidePolygon(p2,pa,pb,pc,&code,&id) && code == 0)
   {
     vp=BoolAddVertex(p2,FALSE);
     if((wee=(WORKEDGE *)realloc(*we,(*nwe + 3)*(long)sizeof(WORKEDGE)))
               == NULL)return NULL;
     *we=wee;
     //
	 //	When point is inserted add three edges and two facets
	 //
     newE(vp,w0);
     e0 = *nwe;     
     wee[e0].e=nEdges-1; 
     wee[e0].t = -1; wee[e0].b=0;
     wee[e0].ea[0] = wee[e0].ea[0] = -1;
     newE(vp,w1);
     e1 = *nwe + 1; 
     wee[e1].e=nEdges-1; 
     wee[e1].t = -1; wee[e1].b=0;
     wee[e1].ea[0] = wee[e1].ea[1] = -1;
     newE(vp,w2);
     e2 = *nwe + 2; 
     wee[e2].e=nEdges-1; 
     wee[e2].t = -1; wee[e2].b=0;
     wee[e2].ea[0] = wee[e2].ea[1] = -1;
     *nwe = *nwe + 3;
     //
	 //	The additional facets
	 //
     if((w=MakeFacet(vp,w0,w1,j,w,nf))==NULL)return NULL;
     if((w=MakeFacet(vp,w2,w0,j,w,nf))==NULL)return NULL;
     //
	 //	Modify the original facet and update adjacency information
	 //
     (FacetList+f1)->Vid[0]=vp;       
     nn = *nf;
     w[nn-2].fa[0]=nn-1;         
     w[nn-2].fa[1]=w[j].fa[0];
     w[nn-2].fa[2]=j;
     w[nn-1].fa[0]=j;
     w[nn-1].fa[1]=w[j].fa[2];
     w[nn-1].fa[2]=nn-2;
     //
	 //	change the adjacency information in the adjacent
	 //
     if((k=w[j].fa[0]) >= 0)
     {  
       for(l=0;l<3;l++)if(w[k].fa[l] == j)w[k].fa[l]=nn-2;
     }
     if((k=w[j].fa[2]) >= 0)
     {
       for(l=0;l<3;l++)if(w[k].fa[l] == j)w[k].fa[l]=nn-1;
     }
     //
	 //	Adjacency in original facet
	 //
     w[j].fa[0]=nn-2;           
     w[j].fa[2]=nn-1;
     w[nn-2].ea[0]=e0;
     w[nn-2].ea[1]=w[j].ea[0];
     w[nn-2].ea[2]=e1;
     w[nn-1].ea[0]=e2;
     w[nn-1].ea[1]=w[j].ea[2];
     w[nn-1].ea[2]=e0;
     w[j].ea[0]=e1;
     w[j].ea[2]=e2;
     //
	 //	Use the Delaunay stacking procedure to update the facet
	 // network according to the "Delaunay" rules.
	 //
     if(Dstack == NULL)
     {
       if((Dstack=(long *)malloc(sizeof(long)*(3)))
                 == NULL)return NULL;
     }
     else 
     {
       if((Dstack=(long *)realloc(Dstack,sizeof(long)*(Dss+3)))
                 == NULL)return NULL;
     }
     if(w[j].fa[1] >= 0)   { Dstack[Dss] = j;    Dss++;}
     if(w[nn-2].fa[1] >= 0){ Dstack[Dss] = nn-2; Dss++;}
     if(w[nn-1].fa[1] >= 0){ Dstack[Dss] = nn-1; Dss++;}
     while(Dss > 0)
     {
       Dss--;
       j=Dstack[Dss]; k=w[j].fa[1];
       if(Dss > 0)
       {
         if((Dstack=(long *)realloc(Dstack,sizeof(long)*(Dss))) == NULL){
           return NULL;  
         }
       }
       if((vo=IdentifyEdge((FacetList+(w[j].f))->Vid[1],
                         (FacetList+(w[j].f))->Vid[2],w,k,
                         &fa1,&fa2,&e1,&e2)) < 0)return NULL;
       if(TestForSwap(p2,(FacetList+(w[j].f))->Vid[1],(FacetList+(w[j].f))->Vid[2],vo))
       {
         e0=w[j].ea[1];
         //
		 //	change the adjacency information in the adjacent 
		 //
         if((m=w[j].fa[2]) >= 0)
         {  
           for(l=0;l<3;l++)if(w[m].fa[l] == j)w[m].fa[l]=k;
         }
         if(fa1 >= 0)
         {
           for(l=0;l<3;l++)if(w[fa1].fa[l] == k)w[fa1].fa[l]=j;
         }
         //
		 //	reverse the edge
		 //
         (EdgeList+wee[e0].e)->Vid[0]=vp;
         (EdgeList+wee[e0].e)->Vid[1]=vo; 
         //
		 //	facet "j"
		 //
         vt=(FacetList+(w[j].f))->Vid[2];
         (FacetList+(w[j].f))->Vid[2]=vo;   
         w[j].fa[1]=fa1; t=w[j].fa[2]; w[j].fa[2]=k;
         w[j].ea[1]=e1; ep=w[j].ea[2]; w[j].ea[2]=e0;
         //
		 //	facet "k"
		 //
         (FacetList+(w[k].f))->Vid[0]=vp;
         (FacetList+(w[k].f))->Vid[1]=vo;
         (FacetList+(w[k].f))->Vid[2]=vt;
         w[k].fa[0]=j; w[k].fa[1]=fa2; w[k].fa[2]=t;
         w[k].ea[0]=e0; w[k].ea[1]=e2; w[k].ea[2]=ep;
         if((Dstack=(long *)realloc(Dstack,sizeof(long)*(Dss+2))) == NULL)
         {
           return NULL;  
         }
		 //
		 //	Push both onto stack
		 //
         if(w[j].fa[1] >= 0){Dstack[Dss] = j; Dss++;}
         if(w[k].fa[1] >= 0){Dstack[Dss] = k; Dss++;}
       }
     }
     if(Dstack != NULL)free(Dstack);
	 //
	 //	point "p2" has been inserted in the triangular network
	 // return pointer to updated workface and workedge lists.
	 //
     return w; 
   }
 }
 return w;
}

// End of Delaunay procedures
/////////////////////////////////////////////////////////////


static WORKFACE *DivideEdge(VECTOR p2, 				   // split at this point
                            long edgeID,			   // edge to be split
                            WORKFACE *w, WORKEDGE **e, // lists of WORK and 
                            long *nface, long *nedge,  // TOOL items
                            long *vid)			       // where to get ID from
 {
 //
 // Split faces attached to edge edgeID at position "p2" - keep 
 // the adjacency information correctly updated to reflect the new
 // structure.
 //
 long nf,i,j,k,l,m,fid,f1id,f2id,faid,e1id,e2id,a0,a1,a2,e0,e1,e2,fai[2],fx;
 long s,vp,q1,q2;
 FACET *f;
 WORKEDGE *be;
 long ep;
 nf=*nface;
 be = *e;
 ep=be[edgeID].e;
 //
 //	creater a vertex at "p2" (somewhere on the edge)
 //
 vp=BoolAddVertex(p2,FALSE);
 //
 // record its identity
 //
 (VertexList+vp)->id=(long)(*vid); 
 *vid = (*vid) + 1;
 q1=(EdgeList+ep)->Vid[0]; 
 q2=(EdgeList+ep)->Vid[1];
 //
 //	Make a new edge so that the original edge is now made up
 // from two pieces. 
 //
 newE(vp,(EdgeList+ep)->Vid[1]); 
 //
 //	Modify the connections of the original
 //
 (EdgeList+ep)->Vid[1]=vp;
 if((be=(WORKEDGE *)realloc(be,((*nedge)+1)*(long)sizeof(WORKEDGE)))
       == NULL)return NULL;
 //
 // The new edge is not an internal edge internal since it is derived from
 // one of the originals
 //
 be[*nedge].e=nEdges-1; 
 be[*nedge].t=0;  
 be[*nedge].b=0;
 *e=be; e1id = *nedge;
 (*nedge)++;
 f1id=be[edgeID].ea[0]; 
 f2id=be[edgeID].ea[1];
 fid=f1id;
 //
 //	now fix up at most two facets on either side of "edgeID" for each
 // facet one extra edge and one extra facet will be needed to maintain
 // triangular polygons
 // 
 //
 for(i=0;i<2;i++)
 {
   if(fid >= 0)
   {
     fx=w[fid].f; f=(FacetList+fx);
     if     (f->Vid[0] == q1 && f->Vid[1] == q2)
       { s=f->Vid[2]; k=1; l=0; m=2; }
     else if(f->Vid[0] == q2 && f->Vid[1] == q1)
       { s=f->Vid[2]; k=2; l=0; m=1; }
     else if(f->Vid[1] == q1 && f->Vid[2] == q2)
       { s=f->Vid[0]; k=2; l=1; m=0; }
     else if(f->Vid[1] == q2 && f->Vid[2] == q1)
       { s=f->Vid[0]; k=0; l=1; m=2; }
     else if(f->Vid[2] == q1 && f->Vid[0] == q2)
       { s=f->Vid[1]; k=0; l=2; m=1; }
     else if(f->Vid[2] == q2 && f->Vid[0] == q1)
       { s=f->Vid[1]; k=1; l=2; m=0; }
     else return NULL;
     a0=w[fid].fa[k]; a1=w[fid].fa[l]; a2=w[fid].fa[m];
     e0=w[fid].ea[k]; e1=w[fid].ea[l]; e2=w[fid].ea[m];
     //
	 //	Modify the original facet 
	 //
     f->Vid[0]=s; f->Vid[1]=vp; f->Vid[2]=q1; 
     //
	 //	make an new edge leading up to "p2"
	 //
     newE(s,vp);        
     if((be=(WORKEDGE *)realloc(be,sizeof(WORKEDGE)*((*nedge) + 1))) == NULL)
        return NULL;
     //
	 // This is an internal edge in phase 2  so set ".t" = 1
	 //
     be[*nedge].e=nEdges-1;
     be[*nedge].t=1; 
     be[*nedge].b=0; *e=be;
     e2id = *nedge;
     (*nedge)++;
     //
	 //	add an extra facet - modify the existing facet to meet up with
	 // the extra point - fix up the adjacency information.
	 //
     if((w=MakeFacet(s,q2,vp,fid,w,nface)) == NULL)return NULL;
     faid = (*nface) - 1;
     w[fid].fa[0]=faid; w[fid].fa[1]=a1;    w[fid].fa[2]=a2;
     w[fid].ea[0]=e2id; w[fid].ea[1]=e1;    w[fid].ea[2]=e2;
     w[faid].fa[0]=a0;  w[faid].fa[1]=a1;   w[faid].fa[2]=fid;
     w[faid].ea[0]=e0;  w[faid].ea[1]=e1id; w[faid].ea[2]=e2id;
     if(a0 >= 0)for(j=0;j<3;j++)
     { 
       //
       // fix up adjacent facet 
	   //
       if(w[a0].fa[j] == fid)w[a0].fa[j]=faid;
     }
     be[e2id].ea[0]=fid; be[e2id].ea[1]=faid;
     for(j=0;j<2;j++)
     { 
       //
       // fix up edge edjacency
	   //
       if(be[e0].ea[j] == fid)be[e0].ea[j]=faid;
     }
     be[e1id].ea[i]=faid;
     fai[i]=faid;
   }
   else
   {
   	 //
	 // the new piece of the edge that was split has no other side 
	 //
	 be[e1id].ea[i] = -1; 
   }
   fid=f2id;
 }
 if(f1id >= 0 && f2id >= 0)
 { //
   // fix adjacent information on facets in new facets   
   //
   w[fai[0]].fa[1] = fai[1];
   w[fai[1]].fa[1] = fai[0];
 }
 return w;
}

static long BoolAddVertex(VECTOR v, long flag)
{
 //
 //	Add a vertex at position "v" and assign it's flag
 //
 VERTEX *vv;
 newV(v);
 vv=(VertexList+nVertices-1);
 vv->id=0; 
 vv->flag=flag;
 return nVertices-1;
}

static WORKFACE *MakeFacet(long a, long b, long c,  // vertices of the facet
                           long i,                  // derived from facet "i"
                           WORKFACE *w,             // work facet list and
                           long *nface)             // number of entries
{
  //
  // Create a facet - update the WORKFACE data structures to
  // include it and copy the normal from facet "i"
  //
  long nf;
  nf = *nface;
  if((w=(WORKFACE *)realloc(w,(nf+1)*(long)sizeof(WORKFACE)))
       == NULL)return NULL;
  newF(a,b,c);
  w[nf].f=nFacets-1;
  CopyVector(w[i].n,w[nf].n);
  (*nface)++;
  return w;
}

static void FlipCommonEdge(long edgeID, 		   // this edge
                           WORKFACE *w, long nf,   // workface list	+ number
                           WORKEDGE *we, long ne)  // wirkedge list + number
{
 //
 // Switch diagonal edgeID in the two faces on either side so that it will
 // join the other two vertices. Take care to keep the adjacency information
 // correct so as to reflect the new structure
 //
 long i,fid1,fid2,k,l,m;
 long a0,a1,a2,e0,e1,e2,b0,b1,b2,d0,d1,d2;
 long q1,q2,s1=-1,s2=-1;
 FACET *f1,*f2;
 EDGE  *e;
 e=(EdgeList+we[edgeID].e);
 q1=e->Vid[0]; q2=e->Vid[1];
 fid1=we[edgeID].ea[0]; fid2=we[edgeID].ea[1];
 f1=(FacetList+w[fid1].f); f2=(FacetList+w[fid2].f);
 if     (f1->Vid[0] == q1 && f1->Vid[1] == q2)
   { s1=f1->Vid[2]; k=1; l=0; m=2; }
 else if(f1->Vid[0] == q2 && f1->Vid[1] == q1)
   { s1=f1->Vid[2]; k=2; l=0; m=1; }
 else if(f1->Vid[1] == q1 && f1->Vid[2] == q2)
   { s1=f1->Vid[0]; k=2; l=1; m=0; }
 else if(f1->Vid[1] == q2 && f1->Vid[2] == q1)
   { s1=f1->Vid[0]; k=0; l=1; m=2; }
 else if(f1->Vid[2] == q1 && f1->Vid[0] == q2)
   { s1=f1->Vid[1]; k=0; l=2; m=1; }
 else if(f1->Vid[2] == q2 && f1->Vid[0] == q1)
   { s1=f1->Vid[1]; k=1; l=2; m=0; }
 //
 // The numbering starts at s1 and goew clockwise keep it consistent
 //
 a0=w[fid1].fa[k]; a1=w[fid1].fa[l]; a2=w[fid1].fa[m]; 
 e0=w[fid1].ea[k]; e1=w[fid1].ea[l]; e2=w[fid1].ea[m]; 
 if     (f2->Vid[0] == q1 && f2->Vid[1] == q2){ s2=f2->Vid[2]; k=1; l=0; m=2; }
 else if(f2->Vid[0] == q2 && f2->Vid[1] == q1){ s2=f2->Vid[2]; k=2; l=0; m=1; }
 else if(f2->Vid[1] == q1 && f2->Vid[2] == q2){ s2=f2->Vid[0]; k=2; l=1; m=0; }
 else if(f2->Vid[1] == q2 && f2->Vid[2] == q1){ s2=f2->Vid[0]; k=0; l=1; m=2; }
 else if(f2->Vid[2] == q1 && f2->Vid[0] == q2){ s2=f2->Vid[1]; k=0; l=2; m=1; }
 else if(f2->Vid[2] == q2 && f2->Vid[0] == q1){ s2=f2->Vid[1]; k=1; l=2; m=0; }
 //
 // For the second facet use an opposite winding 
 //
 b0=w[fid2].fa[m]; b1=w[fid2].fa[l]; b2=w[fid2].fa[k]; 
 d0=w[fid2].ea[m]; d1=w[fid2].ea[l]; d2=w[fid2].ea[k]; 
 if(s1 < 0 || s2 < 0)
 {
   return;
 }
 
 e->Vid[0]=s1; e->Vid[1]=s2;    
 f1->Vid[0]=s1; f1->Vid[1]=s2; f1->Vid[2]=q1;
 w[fid1].fa[0]=fid2; w[fid1].fa[1]=b0; w[fid1].fa[2]=a2;
 w[fid1].ea[0]=edgeID; w[fid1].ea[1]=d0; w[fid1].ea[2]=e2;
 f2->Vid[0]=s1; f2->Vid[1]=q2; f2->Vid[2]=s2;
 w[fid2].fa[0]=a0; w[fid2].fa[1]=b2; w[fid2].fa[2]=fid1;
 w[fid2].ea[0]=e0; w[fid2].ea[1]=d2; w[fid2].ea[2]=edgeID;
 //
 // reset adjacency information in adjacent faces - a0 and b0 
 //
 if(a0 >= 0)for(i=0;i<3;i++)
 {
   if(w[a0].fa[i] == fid1)w[a0].fa[i]=fid2;
 }
 if(b0 >= 0)for(i=0;i<3;i++)
 {
   if(w[b0].fa[i] == fid2)w[b0].fa[i]=fid1;
 }
 //
 // and for the edges e0 and d0 
 //
 for(i=0;i<2;i++)
 {
   if(we[e0].ea[i] == fid1)we[e0].ea[i]=fid2;
   if(we[d0].ea[i] == fid2)we[d0].ea[i]=fid1;
 }
 return;
}

static int TestForFacetSwitch(long edgeID,  // this edge
                              WORKFACE *w,  // workface list
                              WORKEDGE *we) // workedge list
{
 //
 // test to see if faces adjacent to edge edgeID can be swapped      
 // they can be swapped if edge "edgeID" would remain internal to    
 // both faces after any swappping takes place - use vector cross   
 // product for edges adjacent to q1 and q2 (not e) and test that  
 // crosses point in same direction ie. c1 dot c2 > 0. Note all    
 // vectors lie in the same plane as they originated in the same   
 // original triangular face.
 //
 long fid1,fid2;
 long q1,q2,s1=-1,s2=-1;
 FACET *f1,*f2;
 EDGE *e;
 VECTOR vs1q1,vs2q1,vs1q2,vs2q2,va,vb;
 e=(EdgeList+we[edgeID].e);
 q1=e->Vid[0]; q2=e->Vid[1];
 fid1=we[edgeID].ea[0]; fid2=we[edgeID].ea[1];
 f1=(FacetList+w[fid1].f); f2=(FacetList+w[fid2].f);
 if     (f1->Vid[0] == q1 && f1->Vid[1] == q2)s1=f1->Vid[2];
 else if(f1->Vid[0] == q2 && f1->Vid[1] == q1)s1=f1->Vid[2];
 else if(f1->Vid[1] == q1 && f1->Vid[2] == q2)s1=f1->Vid[0];
 else if(f1->Vid[1] == q2 && f1->Vid[2] == q1)s1=f1->Vid[0];
 else if(f1->Vid[2] == q1 && f1->Vid[0] == q2)s1=f1->Vid[1];
 else if(f1->Vid[2] == q2 && f1->Vid[0] == q1)s1=f1->Vid[1];
 if     (f2->Vid[0] == q1 && f2->Vid[1] == q2)s2=f2->Vid[2];
 else if(f2->Vid[0] == q2 && f2->Vid[1] == q1)s2=f2->Vid[2];
 else if(f2->Vid[1] == q1 && f2->Vid[2] == q2)s2=f2->Vid[0];
 else if(f2->Vid[1] == q2 && f2->Vid[2] == q1)s2=f2->Vid[0];
 else if(f2->Vid[2] == q1 && f2->Vid[0] == q2)s2=f2->Vid[1];
 else if(f2->Vid[2] == q2 && f2->Vid[0] == q1)s2=f2->Vid[1];
 if(s1 < 0 || s2 < 0)
 {
   //
   // This should NOT happen
   //
   return 0; 
 }
 SubtractVectors((VertexList+s1)->p,(VertexList+q1)->p,vs1q1);
 Normalize(vs1q1);
 SubtractVectors((VertexList+s2)->p,(VertexList+q1)->p,vs2q1);
 Normalize(vs2q1);
 SubtractVectors((VertexList+s1)->p,(VertexList+q2)->p,vs1q2);
 Normalize(vs1q2);
 SubtractVectors((VertexList+s2)->p,(VertexList+q2)->p,vs2q2);
 Normalize(vs2q2);
 Cross(vs2q1,vs1q1,va);
 Cross(vs1q2,vs2q2,vb);
 //
 //	return 1 if edge can be switched
 //
 if(Dot(va,vb) > 0.0)return 1; 
 return 0;                     
}

//
// These global static variables are uses to help mimimize
// the depth of the stack during recursive assignment.
//
static int lastID,NF,NE;
static WORKFACE *WF;
static WORKEDGE *WE;

static void SetConnectingID(long id)
{
 //
 //	Set the identity of all faces that are neighbors 
 // of facet "id" and all their neighbors too until
 // either a boundary is reached or the facet has
 // no neighbours.
 //
 long eid;
 WF[id].id=lastID;
 if(WF[id].ea[0] >= 0 && WE[WF[id].ea[0]].b == 0 &&
   WF[id].fa[0] >= 0 && WF[WF[id].fa[0]].id < 0)
   SetConnectingID(WF[id].fa[0]);
 if(WF[id].ea[1] >= 0 && WE[WF[id].ea[1]].b == 0 &&
   WF[id].fa[1] >= 0 && WF[WF[id].fa[1]].id < 0)
   SetConnectingID(WF[id].fa[1]);
 if(WF[id].ea[2] >= 0 && WE[WF[id].ea[2]].b == 0 &&
   WF[id].fa[2] >= 0 && WF[WF[id].fa[2]].id < 0)
   SetConnectingID(WF[id].fa[2]);
}


static void SplitAlongBorder(WORKFACE *wf,   // list of WORK facets
                             long nf,        // number of WORK facets 
                             WORKEDGE *we,   // list of WORK edges
                             long ne)        // number of WORK edges
{
 //
 // This function splits the WORK facet mesh into as many
 // separate pieces are there "destinct" bounded regions.
 // The concept is explained in the text/
 //
 FACET *f;
 EDGE *e;
 VERTEX *vp,*Vp,*v0,*v1,*v2;
 long LastVp;
 long c,i,id,nvs,k,inc,nVertices1,Nface1,nEdges1;
 lastID=0;
 //
 // Assign the temporary global pointers to the lists
 //
 NF=nf; NE=ne;
 WF=wf; WE=we;
 //
 // indicate that no facets have been identified as part of
 // one of the pieces to be separated. "wf[i].id will be 
 // determined to specify to which piece each facet belongs.
 // "wf[i].id = -1 indicated the edge is not yet assigned
 //
 for(i=0;i<nf;i++)wf[i].id = -1;
 //
 // repeat until all the bits have been processed
 //
 while(1)
 { 
   for(id=0;id<nf;id++)
   {
     //
     // found the first unassigned vertex
     //
     if(wf[id].id == -1)goto GET_THE_REST;
   }
   //
   // no bits left so break out of the while loop
   //
   break;  
   GET_THE_REST:
   //
   // set all "wf[j].id" for facets that are adjacent to "id"
   // this process proceeds recursively until it can go no further
   // either because the edge of the mesh is reached or a boundary
   // edge is encountered
   //
   SetConnectingID(id);
   lastID++;
 }
 //
 // Split up the bits and pieces by copying the facets assigned to 
 // each piece, e.g. "wf[j].id = 1" and then deleting original
 //
 if(lastID < 1){
   //
   // there are no pieces
   //
   return;
 }
 LastVp=nVertices;
 for(id=0;id<lastID;id++)
 {
   vp=VertexList; 
   //
   // Zero out vertex flags so that they can be used for copy.
   //
   for(c=0;c<nVertices;c++)
   {
     vp->id=0; 
     vp++;
   }
   for(i=0;i<nf;i++)
   { 
     //
     // flag all vertices used by facets in piece "id" 
     //
     if(wf[i].id == id)
     { 
       f=(FacetList+wf[i].f);
       (VertexList+(f->Vid[0]))->id=1;
       (VertexList+(f->Vid[1]))->id=1;
       (VertexList+(f->Vid[2]))->id=1;
     }
   }
   //
   // Count the number of vertices edges and facets
   // in piece "id"
   //
   nvs=0; 
   vp=VertexList; 
   for(c=0;c<nVertices;c++)
   {
     if(vp->id == 1)
     {
       nvs++;
     }
     vp++;
   }
   for(nEdges1=0,i=0;i<ne;i++)
   { 
     inc=0;
     if((k=we[i].ea[0]) >= 0 && wf[k].id == id)inc=1;
     if((k=we[i].ea[1]) >= 0 && wf[k].id == id)inc=1;
     if(inc == 1)nEdges1++;
   }
   for(Nface1=0,i=0;i<nf;i++)
   {
     if(wf[i].id == id)Nface1++;
   }
   //
   // some structure to copy
   //
   if(nvs > 0 && nEdges1 > 0 && Nface1 > 0 )
   {
     nVertices1=nVertices; 
     nvs=0; 
     vp=VertexList; 
     //
     // duplicate vertices
     //
     for(c=0;c<nVertices1;c++)
     {
       if(vp->id == 1)
       {
         newV(vp->p);
         Vp=(VertexList+nVertices-1);
         vp->id=(long)nVertices-1;
         Vp->flag=FALSE;
         nvs++;
       }
       else vp->id = -1;
       vp++;
     }
     //
     // duplicate facets
     //
     for(i=0;i<nf;i++)
     { 
       if(wf[i].id == id)
       { 
         f=(FacetList+wf[i].f);
         v0=(VertexList+(f->Vid[0])); 
         v1=(VertexList+(f->Vid[1])); 
         v2=(VertexList+(f->Vid[2]));
         newF(v0->id,v1->id,v2->id);
       }
     }
     //
     // duplicate edges
     //
     for(i=0;i<ne;i++)
     {
       inc=0;
       if((k=we[i].ea[0]) >= 0 && wf[k].id == id)inc=1;
       if((k=we[i].ea[1]) >= 0 && wf[k].id == id)inc=1;
       if(inc == 1)
       {
         e=(EdgeList+we[i].e);
         v0=(VertexList+(e->Vid[0])); 
         v1=(VertexList+(e->Vid[1]));
         newE(v0->id,v1->id);
       }
     }
   }
 }
 vp=VertexList; for(c=0;c<LastVp;c++)
 {
   //
   // now flag the vertices that have to been removed
   // (they are the content of the WORKFACE list
   //
   if(vp->flag == TRUE)
   {
     //
     // These vertices should be left undisturbed and
     // so a temporary assignment is applied to them
     //
     vp->flag=UNDECIDED;
   }
   else if(vp->flag == FALSE)
   {
     //
     // These are the vertices to be removed so flag
     // them as true.
     //
     vp->flag=TRUE;
   }
   vp++;
 }
 //
 //	remove any vertex that has its flag value set to TRUE, any
 // edges or facets that are atttached to a removed vertex must 
 // also be removed
 //
 eraseFlaggedVertices();
 //
 //	restore the state of the flags so that the second pass
 // can proceed if necessary
 //
 vp=VertexList; 
 for(c=0;c<nVertices;c++)
 {
   if(vp->flag == UNDECIDED)
   {
     vp->flag=TRUE;
   }
   vp++;
 }
 return;
}

static int FindAdjFaces(
  WORKFACE *wf, 	  // WORK facet list
  long nf, 			  // number of entries in the work facet list
  WORKEDGE *we, 	  // WORK edge list
  long ne)			  // number of entries in the work edge list
{
 EDGE  *e;
 FACET *f;
 VERTEX *vp;
 VTXADJ *vl=NULL;
 long c,nvl=0;
 long *fl,i,j,k,id,id0,id1,id2,fjd,fkd,n,m,kk,nl[3],ml[3];
 //
 //	This function calculates and builds the adjacency information
 // for the data that is used in identifying the various pieces
 // that are separated by edges in the boolean structure.
 //
 //	The basic ideas of this procedure are discussed in the text.
 //
 //	The first thing to do is establish which edges and facets are
 // adjacent to which vertex. This information is recored in a
 // VTXADJ structure, one for each vertex.
 //
 //
 vp=VertexList; 
 //
 //	count the number of vertices associated with WORK items
 //
 for(c=0;c<nVertices;c++)
 {
   if(vp->flag == FALSE)
   { 
     vp->id=nvl;
     nvl++;
   }
   else vp->id = -2;
   vp++;
 }
 //
 //	Assign suitable storage.
 //
 if((vl=(VTXADJ *)malloc(nvl*sizeof(VTXADJ))) == NULL)return 0;
 //
 //	initialize the structures
 //
 for(i=0;i<nvl;i++)
 {
   vl[i].nf=0; vl[i].ne=0;
   vl[i].flist=NULL; vl[i].elist=NULL;
 }
 //
 //	Make list of WORK facets attached to each vertex. See the
 // text and listing CHAP2_6.CPP
 //
 for(i=0;i<nf;i++)
 { 
   f=(FacetList+wf[i].f);
   for(j=0;j<3;j++){
     id=(VertexList+(f->Vid[j]))->id;
     fl=vl[id].flist;
     n=vl[id].nf;
     if(fl == NULL){
       if((fl=(long *)malloc(sizeof(long))) == NULL)return 0;
     }
     else{
       if((fl=(long *)realloc(fl,(n+1)*sizeof(long))) == NULL)return 0;
     }
     fl[n]=i;
     vl[id].flist=fl;
     vl[id].nf=n+1;
   }
 }
 //
 //	Use the vertex list to find the facets that are adjacent to 
 // facet "i"
 //
 for(i=0;i<nf;i++)
 { 
   f=(FacetList+wf[i].f);
   id0=(VertexList+(f->Vid[0]))->id;
   id1=(VertexList+(f->Vid[1]))->id;
   id2=(VertexList+(f->Vid[2]))->id;
   nl[0]=id0; ml[0]=id1;
   nl[1]=id1; ml[1]=id2;
   nl[2]=id2; ml[2]=id0;
   for(kk=0;kk<3;kk++)
   {
     wf[i].fa[kk] = -1;
     if((n=vl[nl[kk]].nf) > 0 && (m=vl[ml[kk]].nf) > 0)
     {
       for(j=0;j<n;j++)
       {
         fjd=vl[nl[kk]].flist[j];
         for(k=0;k<m;k++)
         {
           fkd=vl[ml[kk]].flist[k];
           if(fkd == fjd && fkd != i)
           {
             wf[i].fa[kk]=fkd;
             goto END_FACETS;
           }
         }
       }
     }
     END_FACETS:;
   }
 }
 //
 //	Use the vertex list to find facets that are adjecent
 // to edge "i"
 //
 for(i=0;i<ne;i++)
 { 
   e=(EdgeList+we[i].e);
   id0=(VertexList+(e->Vid[0]))->id; id1=(VertexList+(e->Vid[1]))->id;
   kk=0; we[i].ea[0] = we[i].ea[1] = -1;
   if((n=vl[id0].nf) > 0 && (m=vl[id1].nf) > 0)
   {
     for(j=0;j<n;j++){
       fjd=vl[id0].flist[j];
       for(k=0;k<m;k++){
         fkd=vl[id1].flist[k];
         if(fkd == fjd){
           we[i].ea[kk]=fkd;
           kk++;
           //
		   // Finish when two have been found
		   //
           if(kk == 2)goto END_EDGES2; 
         }
       }
     }
     END_EDGES2:;
   }
 }
 //
 //	Make a list of the WORK edges attached to each vertex. 
 // The procedure is vey similar to the one above for making the
 // list of WORK facets attached to a vertex.
 //
 for(i=0;i<ne;i++)
 {
   e=(EdgeList+we[i].e);
   for(j=0;j<2;j++)
   {
     id=(VertexList+(e->Vid[j]))->id;
     fl=vl[id].elist;
     n=vl[id].ne;
     if(fl == NULL)
     {
       if((fl=(long *)malloc(sizeof(long))) == NULL)return 0;
     }
     else 
     {
       if((fl=(long *)realloc(fl,(n+1)*sizeof(long))) == NULL)return 0;
     }
     fl[n]=i;
     vl[id].elist=fl;
     vl[id].ne=n+1;
   }
 }
 //
 //	Use the list of EDGES attached to a vertex to find the EDGES that
 // lie on either side of an edge. 
 //
 for(i=0;i<nf;i++)
 { 
   f=(FacetList+wf[i].f);
   id0=(VertexList+(f->Vid[0]))->id;
   id1=(VertexList+(f->Vid[1]))->id;
   id2=(VertexList+(f->Vid[2]))->id;
   nl[0]=id0; ml[0]=id1;
   nl[1]=id1; ml[1]=id2;
   nl[2]=id2; ml[2]=id0;
   for(kk=0;kk<3;kk++)
   {
     wf[i].ea[kk] = -1;
     if((n=vl[nl[kk]].ne) > 0 && (m=vl[ml[kk]].ne) > 0)
     {
       for(j=0;j<n;j++)
       {
         fjd=vl[nl[kk]].elist[j];
         for(k=0;k<m;k++)
         {
           fkd=vl[ml[kk]].elist[k];
           if(fkd == fjd){
             wf[i].ea[kk]=fkd;
             goto END_EDGES1;
           }
         }
       }
     }
     END_EDGES1:;
   }
 }
 //
 //	Now that everything has been found free up the
 // memory allocated for the vertex lists.
 //
 for(i=0;i<nvl;i++)
 {
   if(vl[i].flist != NULL)free(vl[i].flist);
   if(vl[i].elist != NULL)free(vl[i].elist);
 }
 free(vl);
 return nvl;
return 0;
}
