//
// Illustrate some image processing functions such as blurring
// filtering and basic edge detection. The functions use the same
// calling parameters as used by Video processing effects so that
// if required they can be applied to a collection of images
// originating from a video signal.
//

//
// Include file that contains definitions for the PIXEL data structure
//

#include "chap8.h"

///////// The following are standard functions that are usually part
///////// of a compilers library.
///////// Substitute then with the ones you wish to use.


extern void *malloc(long X);      // allocate X bytes, return pointer to it.
extern void free(void *P);        // release memory block pointed to by P
extern float sqrt(float D);       // return floating point square root of  D
extern long labs(long a);         // return absolute value
//
// Modulo functions
//
double fmod(double, double); // floating point modulo
//
// integer modulo - implemented as macro
//
#define imod(z,a) ((z) - ((z)/(a))*(a))
//
// Define the NULL
//
#define NULL 0
//
// Maximum and Minimum operations
//
#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif

//////////////////////////////////////////////////////////////////
//////// Local function  /////////////////////////////////////////
//////////////////////////////////////////////////////////////////

static void Convolve(long Width,    // width and height of the the
                     long Height,   // image
                     lpPIXEL Si,    // pointer to input
                     lpPIXEL So,    // and output buffers
                     float scale,   // scaling factor to apply
                     long bias,     // bias to add after scaling
                     long filterX,  // width and height
                     long filterY,  // of the filter
                     float *filter) // pointer to the filter
{
 //
 // Perform "convolution" on the image using a given filter
 //
 lpPIXEL S;
 long i,j,k,l,c,r;
 float ar,ag,ab,weight;
 r=filterY/2;
 c=filterX/2;
 for(i=r;i<Height-r;i++)
 {
   for(j=c;j<Width-c;j++)
   {
     ar=ag=ab=0.0;
     for(l=0;l<filterY;l++)for(k=0;k<filterX;k++)
     {
       S=(Si+(Width*(i+(l-r))+(j+(k-c))));
       weight = *(filter+filterX*l+k);
       ar+=(float)S->Red  *weight;
       ag+=(float)S->Green*weight;
       ab+=(float)S->Blue *weight;
     }
     S=(So+(Width*i)+j);
     S->Red   = (unsigned char)max(0.0,(min(255.0,bias+ar*scale)));
     S->Green = (unsigned char)max(0.0,(min(255.0,bias+ag*scale)));
     S->Blue  = (unsigned char)max(0.0,(min(255.0,bias+ab*scale)));
   }
 }
}

//
//  Some basic image processing functions
//

void BasicSobelEdgeDetect(
     lpPIXEL I1,      // first input stream
     lpPIXEL I2,      // (not used)
     lpPIXEL O,       // pointer to frame buffer for result
     long    Width,   // width of images in framebuffers
     long    Height,  // height of images in frame buffers
     float   t)       // (not used)
{
 //
 // Perform the SOBEL edge detection algorithm
 //
 long i,j;
 long lA,lB,lC,lD,lE,lF,lG,lH,lI,MaxDif,threshold;
 long lineAEIbelow,lineAEIabove,lineAEIdif;
 long lineBEHbelow,lineBEHabove,lineBEHdif;
 long lineCEGbelow,lineCEGabove,lineCEGdif;
 long lineDEFbelow,lineDEFabove,lineDEFdif;
 lpPIXEL S;
 //
 // Threshold is a parameter in range [0-255] that controls how noticable
 // an edge has to be before it is represented in the output.
 //
 threshold=16;
 threshold *= 3;
 for(i=1;i<Height-1;i++)for(j=1;j<Width-1;j++)
 {
   S=(I1+(Width*(i-1))+(j-1));
   lA = (long)S->Red   + (long)S->Green + (long)S->Blue ;
   S=(I1+(Width*(i  ))+(j-1));
   lB = (long)S->Red   + (long)S->Green + (long)S->Blue ;
   S=(I1+(Width*(i+1))+(j-1));
   lC = (long)S->Red   + (long)S->Green + (long)S->Blue ;
   S=(I1+(Width*(i-1))+(j  ));
   lD = (long)S->Red   + (long)S->Green + (long)S->Blue ;
   S=(I1+(Width*(i  ))+(j  ));
   lE = (long)S->Red   + (long)S->Green + (long)S->Blue ;
   S=(I1+(Width*(i+1))+(j  ));
   lF = (long)S->Red   + (long)S->Green + (long)S->Blue ;
   S=(I1+(Width*(i-1))+(j+1));
   lG = (long)S->Red   + (long)S->Green + (long)S->Blue ;
   S=(I1+(Width*(i  ))+(j+1));
   lH = (long)S->Red   + (long)S->Green + (long)S->Blue ;
   S=(I1+(Width*(i+1))+(j+1));
   lI = (long)S->Red   + (long)S->Green + (long)S->Blue ;
   lineAEIabove=(lB+lC+lF)/3;
   lineAEIbelow=(lD+lG+lH)/3;
   lineAEIdif=labs(lineAEIbelow-lineAEIabove);
   lineBEHabove=(lC+lF+lI)/3;
   lineBEHbelow=(lA+lD+lG)/3;
   lineBEHdif=labs(lineBEHbelow-lineBEHabove);
   lineCEGabove=(lA+lB+lD)/3;
   lineCEGbelow=(lH+lF+lI)/3;
   lineCEGdif=labs(lineCEGbelow-lineCEGabove);
   lineDEFabove=(lA+lB+lC)/3;
   lineDEFbelow=(lG+lH+lI)/3;
   lineDEFdif=labs(lineDEFbelow-lineDEFabove);
   MaxDif=max(lineAEIdif,lineBEHdif);
   MaxDif=max(lineCEGdif,MaxDif);
   MaxDif=max(lineDEFdif,MaxDif);
   S=(O+(Width*i)+j);
   if(MaxDif > threshold)
   {
     S->Red   = (unsigned char)(255.0);
     S->Green = (unsigned char)(255.0);
     S->Blue  = (unsigned char)(255.0);
   }
 }
}

void EnhanceEdges(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Enhance the edges in the image by using the Laplace filter
 //
 float filter[]={0.0,-1.0,0.0,  -1.0,4.0,-1.0,  0.0,-1.0,0.0};
 Convolve(Width,Height,I1,O,1.0,128,3,3,filter);
}

void BasicBlur(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Blur the image by averaging a 5 x 5 block of pixels
 //
 float filter[]={1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};
 Convolve(Width,Height,I1,O,1.0/25.0,0,5,5,filter);
}

void HighPassFilter(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Apply a high pass filter to the image there are 3 types, choose the
 // one that gives the best result
 //
//   float filter[]={1.0,-2.0,1.0,  -2.0,5.0,-2.0,  1.0,-2.0,1.0};
//   float filter[]={0.0,-1.0,0.0,  -1.0,5.0,-1.0,  0.0,-1.0,0.0};
   float filter[]={-1.0,-1.0,-1.0,  -1.0,9.0,-1.0,  -1.0,-1.0,-1.0};
   Convolve(Width,Height,I1,O,1.0,128,3,3,filter);
}

void LowPassFilter(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Apply a low pass filter to the image
 //
 float filter[]={0.111111,0.111111,0.111111,
                 0.111111,0.111111,0.111111,
                 0.111111,0.111111,0.111111};
 Convolve(Width,Height,I1,O,1.0,0,3,3,filter);
}

void SharpenImage(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Give the illusion of a sharper clearer image by mixing the result of a
 // high pass filter with the original image
 //
 long i,j;
 float filter[]={-1.0,-1.0,-1.0,  -1.0,9.0,-1.0,  -1.0,-1.0,-1.0};
 Convolve(Width,Height,I1,O,1.00,128,3,3,filter);
 //
 // mix in high frequency component, subtract 128 to allow for +/-
 // bias centered on the original value.
 //
 for(i=1;i<Height-1;i++)for(j=1;j<Width-1;j++)
 {
   O->Red   = (unsigned char)max(0.0,min(255.0,0.80*(float)I1->Red
            + 0.40*((float)O->Red  -128)));
   O->Green = (unsigned char)max(0.0,min(255.0,0.80*(float)I1->Green
            + 0.40*((float)O->Green-128)));
   O->Blue  = (unsigned char)max(0.0,min(255.0,0.80*(float)I1->Blue
            + 0.40*((float)O->Blue -128)));
   I1++;
   O++;
 }
}
