//
// Basic mixing and wiping video processes. Each function is called
// in a standard way using parameters discussed in the text.
//
//
// Include file that contains definitions for the PIXEL data structure
//

#include "chap8.h"

///////// The following are standard functions that are usually part
///////// of a compilers library.
///////// Substitute then with the ones you wish to use.


extern void *malloc(long X);      // allocate X bytes, return pointer to it.
extern void free(void *P);        // release memory block pointed to by P
extern float sqrt(float D);       // return floating point square root of  D
//
//  Quick copy function to copy a block of memory to another location
//
extern void memcpy(void *Destination, void *Source, long NumberOfBytes);
//
// Random number functions
//
void srand(unsigned int);  // initialize the random number generator
int  rand(void);           // generate random number "C" language spec
                           // says this is in range [-32767 +32767]
//
// Modulo functions
//
double fmod(double, double); // floating point modulo
//
// integer modulo - implemented as macro
//
#define imod(z,a) ((z) - ((z)/(a))*(a))
//
// Define the NULL
//
#define NULL 0


//////////////////////////////////////////////////////////////////////////
//////// Now the effects /////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////



void BasicMix(
     lpPIXEL I1,      // first input stream  (pointer to framebuffer)
     lpPIXEL I2,      // second input stream
     lpPIXEL O,       // pointer to frame buffer for result
     long    Width,   // width of images in framebuffers
     long    Height,  // height of images in frame buffers
     float   t)       // if t=0 then  O=I1  if t=1 then O=I2
{
 //
 // The simplest effect possible!
 // Mix images from "input1" to "input2" as "tau" changes 0 -> 1
 //
 long i,j;
 //
 // Work throught the images mixing each pixel in turn
 //
 for(i=0;i<Height;i++)for(j=0;j<Width;j++)
 {
   O->Red   = (unsigned char)((float)(I1->Red)  *(1.0-t)+(float)I2->Red*t);
   O->Green = (unsigned char)((float)(I1->Green)*(1.0-t)+(float)I2->Green*t);
   O->Blue  = (unsigned char)((float)(I1->Blue) *(1.0-t)+(float)I2->Blue*t);
   I1++;
   I2++;
   O++;
 }
}

void BasicLeftToRightWipe(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Image "I2" appears from the left side of the screen as t goes 0 -> 1
 // Function is easily modifed for right to left mix
 //
 lpPIXEL p;
 long i,j;
 long w;
 //
 // "w" is the horizontal coordinate beyond which I1 appears.
 //
 w=(long)((float)Width*t)+1;
 for(i=0;i<Height;i++)for(j=0;j<Width;j++)
 {
   if(j < w)p=(I2+i*Width+j);
   else     p=(I1+i*Width+j);
   O->Red  =(unsigned char)p->Red  ;
   O->Green=(unsigned char)p->Green;
   O->Blue =(unsigned char)p->Blue ;
   O++;
 }
}

void LeftToRightWipeWithSoftEdge(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Image "I2" appears from the left side of the screen as t goes 0 -> 1
 // the edge between the two images appears soft as the two image blend
 // together over a small fraction of the width of the image.
 //
 long i,j;
 long w,dx;
 float f;
 //
 // dx is number of pixels over which the images blend together
 //
 dx=Width/5;
 w = -dx + (long)((float)(Width+dx)*t);
 for(i=0;i<Height;i++)
 {
   for(j=0;j<Width;j++)
   {
     if(j < w)
     {
       //
       // copy in image 2
       //
       O->Red  =(unsigned char)I2->Red  ;
       O->Green=(unsigned char)I2->Green;
       O->Blue =(unsigned char)I2->Blue ;
     }
     else if(j < w+dx)
     {
       //
       // mix the two images together
       //
       f = 1.0-(float)(j-w)/(float)dx;
       O->Red   = (unsigned char)((float)(I1->Red  )*(1.0-f)+(float)I2->Red  *f);
       O->Green = (unsigned char)((float)(I1->Green)*(1.0-f)+(float)I2->Green*f);
       O->Blue  = (unsigned char)((float)(I1->Blue )*(1.0-f)+(float)I2->Blue *f);
     }
     else
     {
       //
       // copy in image 1
       //
       O->Red  =(unsigned char)I1->Red  ;
       O->Green=(unsigned char)I1->Green;
       O->Blue =(unsigned char)I1->Blue ;
     }
     I1++;
     I2++;
     O++;
   }
 }
}

void RightToLeftWipeWithSoftEdge(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Image "I2" appears from the right side of the screen as t goes 0 -> 1
 // the edge between the two images appears soft as the two image blend
 // together over a small fraction of the width of the image.
 //
 long i,j;
 long w,dx;
 float f;
 //
 // dx is number of pixels over which the images blend together
 //
 dx=Width/5;
 w=Width-(long)((float)(Width+dx)*t);
 for(i=0;i<Height;i++)
 {
   for(j=0;j<Width;j++)
   {
     if(j >= w+dx)
     {
       O->Red  =(unsigned char)I2->Red  ;
       O->Green=(unsigned char)I2->Green;
       O->Blue =(unsigned char)I2->Blue ;
     }
     else if(j > w)
     {
       f = (float)(j-w)/(float)dx;
       O->Red   = (unsigned char)((float)(I1->Red  )*(1.0-f)+(float)I2->Red  *f);
       O->Green = (unsigned char)((float)(I1->Green)*(1.0-f)+(float)I2->Green*f);
       O->Blue  = (unsigned char)((float)(I1->Blue )*(1.0-f)+(float)I2->Blue *f);
     }
     else {
       O->Red  =(unsigned char)I1->Red  ;
       O->Green=(unsigned char)I1->Green;
       O->Blue =(unsigned char)I1->Blue ;
     }
     I1++;
     I2++;
     O++;
   }
 }
}

void HorizontalWipeTowardsCenterWithSoftEdge(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Image "I2" appears from the left and right edges as t goes 0 -> 1
 // Edges between the two images appears soft as the two image blend
 // together over a small fraction of the width of the image.
 //
 // nx is the percentage of the width of the image over which the blend
 // between the incoming and outgoing images are mixed.
 //
 long i,j,nx=10;
 long w1,w2,dx;
 float f;
 dx=(Width*nx)/100;
 w1 =  -dx + (long)((float)(Width/2+dx)*t);
 w2 = Width + dx - (long)((float)(Width/2+dx)*t);
 for(i=0;i<Height;i++)
 {
   for(j=0;j<Width;j++)
   {
     if(j < w1 || j >= w2)
     {
       //
       //  Copy in image 2
       //
       O->Red  =(unsigned char)I2->Red  ;
       O->Green=(unsigned char)I2->Green;
       O->Blue =(unsigned char)I2->Blue ;
     }
     else if(j < w1+dx || j >= w2-dx)
     {
       //
       //  mix the two images
       //
       if(j < Width/2)f = 1.0-(float)(j-w1)/(float)dx;
       else           f = (float)(j-(w2-dx))/(float)dx;
       O->Red   = (unsigned char)((float)(I1->Red  )*(1.0-f)+(float)I2->Red  *f);
       O->Green = (unsigned char)((float)(I1->Green)*(1.0-f)+(float)I2->Green*f);
       O->Blue  = (unsigned char)((float)(I1->Blue )*(1.0-f)+(float)I2->Blue *f);
     }
     else
     {
       //
       //  copy remainder of starting image
       //
       O->Red  =(unsigned char)I1->Red  ;
       O->Green=(unsigned char)I1->Green;
       O->Blue =(unsigned char)I1->Blue ;
     }
     I1++;
     I2++;
     O++;
   }
 }
}

void HorizontalWipeAwayFromCenterWithSoftEdge(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Image "I2" appears in the centre of screen and moves towards the
 // the left and right edges as t goes 0 -> 1
 // Edges between the two images appears soft as the two image blend
 // together over a small fraction of the width of the image.
 //
 // nx is the percentage of the width of the image over which the blend
 // between the incoming and outgoing images are mixed.
 //
 long i,j,nx=10;
 long w1,w2,dx;
 float f;
 dx=(Width*nx)/100;
 w1 = Width/2 - (long)((float)(Width/2+dx)*t);
 w2 = Width/2 + (long)((float)(Width/2+dx)*t);
 for(i=0;i<Height;i++)
 {
   for(j=0;j<Width;j++)
   {
     if(j > w1+dx && j < w2-dx)
     {
       O->Red  =(unsigned char)I2->Red  ;
       O->Green=(unsigned char)I2->Green;
       O->Blue =(unsigned char)I2->Blue ;
     }
     else if(j < w1 || j >= w2)
     {
       O->Red  =(unsigned char)I1->Red  ;
       O->Green=(unsigned char)I1->Green;
       O->Blue =(unsigned char)I1->Blue ;
     }
     else
     {
       if(j < Width/2)f = (float)(j-w1)/(float)dx;
       else           f = 1.0-(float)(j-(w2-dx))/(float)dx;
       O->Red   = (unsigned char)((float)(I1->Red  )*(1.0-f)+(float)I2->Red  *f);
       O->Green = (unsigned char)((float)(I1->Green)*(1.0-f)+(float)I2->Green*f);
       O->Blue  = (unsigned char)((float)(I1->Blue )*(1.0-f)+(float)I2->Blue *f);
     }
     I1++;
     I2++;
     O++;
   }
 }
}

void BasicTopToBottomWipe(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Image "I2" appears from the top of the screen as t goes 0 -> 1
 // Function is easily modifed for BottomToTopWipe
 //
 long i,j;
 long w;
 w=(long)((float)Height*t)+1;
 for(i=0;i<Height;i++)for(j=0;j<Width;j++)
 {
   if(i < w)
   {
     O->Red  =(unsigned char)I2->Red  ;
     O->Green=(unsigned char)I2->Green;
     O->Blue =(unsigned char)I2->Blue ;
   }
   else
   {
     O->Red  =(unsigned char)I1->Red  ;
     O->Green=(unsigned char)I1->Green;
     O->Blue =(unsigned char)I1->Blue ;
   }
   I1++;
   I2++;
   O++;
 }
}

void VerticalWipeTowardsCenterWithSoftEdge(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Image "I2" appears from the top and bottom edges as t goes 0 -> 1
 // Edges between the two images appears soft as the two image blend
 // together over a small fraction of the width of the image.
 //
 // ny is the percentage of the height of the image over which the blend
 // between the incoming and outgoing images are mixed.
 //
 long i,j,ny=10;
 long w1,w2,dy;
 float f;
 dy=(Height*ny)/100;
 w1 =  -dy + (long)((float)(Height/2+dy)*t);
 w2 = Height + dy - (long)((float)(Height/2+dy)*t);
 for(i=0;i<Height;i++)
 {
   for(j=0;j<Width;j++)
   {
     if(i < w1 || i >= w2)
     {
       O->Red  =(unsigned char)I2->Red  ;
       O->Green=(unsigned char)I2->Green;
       O->Blue =(unsigned char)I2->Blue ;
     }
     else if(i < w1+dy || i >= w2-dy)
     {
       if(i < Height/2)f = 1.0-(float)(i-w1)/(float)dy;
       else            f = (float)(i-(w2-dy))/(float)dy;
       O->Red   = (unsigned char)((float)(I1->Red  )*(1.0-f)+(float)I2->Red  *f);
       O->Green = (unsigned char)((float)(I1->Green)*(1.0-f)+(float)I2->Green*f);
       O->Blue  = (unsigned char)((float)(I1->Blue )*(1.0-f)+(float)I2->Blue *f);
     }
     else
     {
       O->Red  =(unsigned char)I1->Red  ;
       O->Green=(unsigned char)I1->Green;
       O->Blue =(unsigned char)I1->Blue ;
     }
     I1++;
     I2++;
     O++;
   }
 }
}

void VerticalWipeAwayFromCenterWithSoftEdge(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Image "I2" appears in a horizontal band at the center of the image
 // it expands towards the top and bottom of framebuffer as t goes 0 -> 1
 // Edges between the two images appears soft as the two image blend
 // together over a small fraction of the width of the image.
 //
 // ny is the percentage of the height of the image over which the blend
 // between the incoming and outgoing images are mixed.
 //
 long i,j,ny=10;
 long w1,w2,dy;
 float f;
 dy=(Height*ny)/100;
 w1 = Height/2 - (long)((float)(Height/2+dy)*t);
 w2 = Height/2 + (long)((float)(Height/2+dy)*t);
 for(i=0;i<Height;i++)
 {
   for(j=0;j<Width;j++)
   {
     if(i > w1+dy && i < w2-dy)
     {
       O->Red  =(unsigned char)I2->Red  ;
       O->Green=(unsigned char)I2->Green;
       O->Blue =(unsigned char)I2->Blue ;
     }
     else if(i < w1 || i >= w2)
     {
       O->Red  =(unsigned char)I1->Red  ;
       O->Green=(unsigned char)I1->Green;
       O->Blue =(unsigned char)I1->Blue ;
     }
     else
     {
       if(i < Height/2)f = (float)(i-w1)/(float)dy;
       else            f = 1.0-(float)(i-(w2-dy))/(float)dy;
       O->Red   = (unsigned char)((float)(I1->Red  )*(1.0-f)+(float)I2->Red  *f);
       O->Green = (unsigned char)((float)(I1->Green)*(1.0-f)+(float)I2->Green*f);
       O->Blue  = (unsigned char)((float)(I1->Blue )*(1.0-f)+(float)I2->Blue *f);
     }
     I1++;
     I2++;
     O++;
   }
 }
}

void BasicSquareWipeFromCenter(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Image I2 emerges from the center of the screen
 //
 long i,j;
 long w,u;
 w=(long)((float)Width/2*t);
 u=(long)((float)Height/2*t);
 for(i=0;i<Height;i++)for(j=0;j<Width;j++)
 {
   if(j > Width/2-w && j < Width/2+w &&
      i > Height/2-u && i < Height/2+u)
   {
     O->Red  =(unsigned char)I2->Red  ;
     O->Green=(unsigned char)I2->Green;
     O->Blue =(unsigned char)I2->Blue ;
   }
   else
   {
     O->Red  =(unsigned char)I1->Red  ;
     O->Green=(unsigned char)I1->Green;
     O->Blue =(unsigned char)I1->Blue ;
   }
   I1++;
   I2++;
   O++;
 }
}

void CircularWipeToCentre(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Image I1 fades away towards the
 // center of the screen.
 // Edges between the two images appears soft as the two image blend
 // together over a small fraction of the width of the image.
 //
 // nx is the percentage of the width of the image over which the blend
 // between the incoming and outgoing images are mixed.
 //
 long i,j,nx=10;
 float r,rw,dr,f,rmax;
 rmax=sqrt((float)(Height/2)*(float)(Height/2)
     +(float)(Width/2)*(float)(Width/2));
 dr=(rmax*(float)nx)/100.0;
 rw = rmax+dr-(rmax+dr)*t;
 for(i=0;i<Height;i++)
 {
   for(j=0;j<Width;j++)
   {
     r=sqrt((float)(i-Height/2)*(float)(i-Height/2)
      +(float)(j-Width/2)*(float)(j-Width/2));
     if(r > rw)
     {
       O->Red  =(unsigned char)I2->Red  ;
       O->Green=(unsigned char)I2->Green;
       O->Blue =(unsigned char)I2->Blue ;
     }
     else if(r > rw-dr)
     {
       f = (r-(rw-dr))/dr;
       O->Red   = (unsigned char)((float)(I1->Red  )*(1.0-f)+(float)I2->Red  *f);
       O->Green = (unsigned char)((float)(I1->Green)*(1.0-f)+(float)I2->Green*f);
       O->Blue  = (unsigned char)((float)(I1->Blue )*(1.0-f)+(float)I2->Blue *f);
     }
     else
     {
       O->Red  =(unsigned char)I1->Red  ;
       O->Green=(unsigned char)I1->Green;
       O->Blue =(unsigned char)I1->Blue ;
     }
     I1++;
     I2++;
     O++;
   }
 }
}

void CircularWipeFromCentre(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Image "I2" appears from the center of the screen in a soft edged disk.
 //
 // nx is the percentage of the width of the image over which the blend
 // between the incoming and outgoing images are mixed.
 //
 long i,j,nx=10;
 float r,rw,dr,f,rmax;
 rmax=sqrt((float)(Width/2)*(float)(Width/2)
     +(float)(Height/2)*(float)(Height/2));
 dr=(rmax*(float)nx)/100.0;
 rw = (rmax+dr)*t;
 for(i=0;i<Height;i++)
 {
   for(j=0;j<Width;j++)
   {
     r=sqrt((float)(i-Height/2)*(float)(i-Height/2)
      +(float)(j-Width/2)*(float)(j-Width/2));
     if(r > rw)
     {
       O->Red  =(unsigned char)I1->Red  ;
       O->Green=(unsigned char)I1->Green;
       O->Blue =(unsigned char)I1->Blue ;
     }
     else if(r > rw-dr)
     {
       f = 1.0-(r-(rw-dr))/dr;
       I1->Red   = (unsigned char)((float)(I1->Red  )*(1.0-f)+(float)I2->Red  *f);
       I1->Green = (unsigned char)((float)(I1->Green)*(1.0-f)+(float)I2->Green*f);
       I1->Blue  = (unsigned char)((float)(I1->Blue )*(1.0-f)+(float)I2->Blue *f);
     }
     else
     {
       O->Red  =(unsigned char)I2->Red  ;
       O->Green=(unsigned char)I2->Green;
       O->Blue =(unsigned char)I2->Blue ;
     }
     I1++;
     I2++;
     O++;
   }
 }
}

void WipeWithRectangles(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Wipe the image by exposing the second image in an incresing number of
 // rectangles that spread in lines down from the top left corner.
 //
 // nnx is the number of blocks to uses across the width of the image and
 // nny is the mumber to use to cover the height of the image.
 //
 lpPIXEL s,o;
 long nnx=10,nny=10;
 long i,j,k,l,nx,ny,n,ii,jj,id;
 nx = Width/nnx; ny=Height/nny;
 if(nx*nnx < Width)nx++;
 if(ny*nny < Height)ny++;
 n=(long)((float)nnx*(float)nny*t);
 ii=0;
 for(i=0;i<Height;i+=ny)
 {
   jj=0;
   for(j=0;j<Width;j+=nx)
   {
     id=ii*nnx+jj;
     for(k=i;k<i+ny;k++)for(l=j;l<j+nx;l++)
     {
       if(id < n && l < Width && k < Height)
       {
         o=(O+k*Width+l);
         s=(I2+k*Width+l);
         o->Red  =(unsigned char)s->Red  ;
         o->Green=(unsigned char)s->Green;
         o->Blue =(unsigned char)s->Blue ;
       }
       else if(l < Width && k < Height)
       {
         o=(O+k*Width+l);
         s=(I1+k*Width+l);
         o->Red  =(unsigned char)s->Red  ;
         o->Green=(unsigned char)s->Green;
         o->Blue =(unsigned char)s->Blue ;
       }
     }
     jj++;
   }
   ii++;
 }
}

void MirrorImage(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Make the output the mirror image of input I1,  I2 and t are unused
 //
 lpPIXEL RowO,RowI;
 long i,j;
 for(i=0;i<Height;i++)
 {
   RowO=(O+i*Width);
   RowI=(I1+i*Width);
   for(j=0;j<Width;j++)
   {
     (RowO+j)->Red   = (RowI+Width-1-j)->Red  ;
     (RowO+j)->Green = (RowI+Width-1-j)->Green;
     (RowO+j)->Blue  = (RowI+Width-1-j)->Blue ;
     (RowO+j)->Alpha = (RowI+Width-1-j)->Alpha;
   }
 }
}

void UpsideDown(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Turn the image recorded in I1 upside down.
 // "I2" and "t" are unused.
 //
 long i,j;
 lpPIXEL Row1,Row2;
 for(i=0;i<Height/2;i++)
 {
   Row1=(O+i*Width);
   Row2=(I1+(Height-1-i)*Width);
   memcpy(Row1,Row2,Width*sizeof(PIXEL));
 }
}

void SliceHorizontally(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Slice the image in I1 into "ny" strips - siding the slices apart
 // to reveal the image in "I2"
 //
 int right, ny=10;
 long i,j,k,dy,dx;
 lpPIXEL s,b,o;
 //
 // Width of each strip
 //
 dy=Height/ny;
 //
 // How far to slide each strip
 //
 dx=(long)((float)Width*t);
 //
 // Slide each strip in opposite directions
 //
 right=1;
 for(k=0;k<Height;k+=dy)
 {
   //
   // Do for each strip
   //
   for(i=k;i<k+dy;i++)for(j=0;j<Width;j++)
   {
     if(i >= Height)continue;
     o=(O+i*Width+j);
     b=(I2+i*Width+j);
     if(right)
     {
       if(j > dx)
       {
         if(j-dx < Width)
         {
           s=(I1+i*Width+j-dx);
           o->Red   = s->Red  ;
           o->Green = s->Green;
           o->Blue  = s->Blue ;
         }
       }
       else
       {
         o->Red   = b->Red  ;
         o->Green = b->Green;
         o->Blue  = b->Blue ;
       }
     }
     else
     {
       if(j < Width-dx)
       {
         if(j+dx < Width && j+dx >= 0)
         {
           s=(I1+i*Width+j+dx);
           o->Red   = s->Red  ;
           o->Green = s->Green;
           o->Blue  = s->Blue ;
         }
       }
       else
       {
         o->Red   = b->Red  ;
         o->Green = b->Green;
         o->Blue  = b->Blue ;
       }
     }
   }
   right ^= 1;
 }
 return;
}

//
// The following functions are "helper" functions for a wipe that
// uses random squares to merge between the first image into the second.
// The "squares" can be anu size from a single pixel up to 1/4 of the
// screen area.
//
// The progress of the wipe is such that when "t=0" the first image is
// visible, when "t=1" the second is visible, when t=0.5 half the screen
// are shows one image and half the other.  For example if the size of
// the block is 1 x 1 pixels and the image is 320x200 then 32000 pixels
// (chosen randomly) show image 1 and the other 32000 show image 2.
// To choose 32000 pixels from the 64000 in the image at random is NOT
// a trivial, simply using a random mumber generator won't do because
// it can lead to repeated selection (of the same pixel). In addition
// the standard "C" rand() function only generated itegers in the range
// 0-64535 and this is not sufficient to select randomly in images larger
// than about 320x200.
//
// Thus to create the wipe the "helper" functions must facilitate a
// proportion of the number of random squares in the image to be selected.
// This proportion is the parameter "t"
//

// The first group of helper functions generate double precision random
// numbers using the algorithm set out at the end of chapter 1.
// The C functions "srand" and "rand" are used to seed the random number
// generator in the function "sRandom" and function "Random" returns a
// double precision rnadom number in the interval [0-1] because it is
// a double precision result the "non-uniqueness" problem with "rand()"
// is avoided.


static long IX=9123,IY=8844,IZ=20846;

static void  sRandom(long i){
 srand(i); IX=rand(); IY=rand(); IZ=rand();
}

static double Random(void){
 IX=imod(171*IX,30269);
 IY=imod(172*IY,20207);
 IZ=imod(170*IZ,30323);
 return (double)fmod((double)IX/30269.0+
                     (double)IY/30307.0+
                     (double)IZ/30323.0,1.0);
}

static void SortArray(long n, long *id, double *r)
//
// A standard sorting procedure that arranges the elements in two lists
// according to the magnitiude of the elements in one of them.
// In this specific case the lists of "n" elements (pointed to) by "id"
// and "r" are sorted into ascending values of entries in the "r" list.
// For example if n=5
//    Before:                              After:
//  (entry)  (id)    (r)           (entry)   (id)      (r)
//    0        0    10.0              0        1       1.0
//    1        1     1.0              1        4       2.0
//    2        2    50.0              2        3       3.0
//    3        3     3.0              3        0      10.0
//    4        4     2.0              4        2      50.0
//
//
{
 long i,j,gap,itemp;
 double rtemp;
 for (gap = n; gap > 0; gap /=2)
 {
   for (i = gap; i < n; i++)
   {
     for (j = i-gap; j >= 0 && *(r+j) < *(r+j+gap); j -= gap)
     {
       rtemp = *(r+j);
       *(r+j) = *(r+j+gap);
       *(r+j+gap) = rtemp;
       itemp = *(id+j);
       *(id+j) = *(id+j+gap);
       *(id+j+gap) = itemp;
     }
   }
 }
}

static long *GenerateRandomSelection(long nb)
{
 //
 // This function generates a list of "nb" numbers so that all the integers
 // up to "nb" (0,1,2 ... nb-1) are present in the list but ORDERED RANDOMLY.
 //
 // Thus when this list is used in the RandomBlock video mix we can
 // guarentee that a certain proportion of the blocks covering the screen
 // will be filled. For example of there are 5 blocks the random list
 // returned by this function might look like.
 //  (entry)   (value)
 //     0         4
 //     1         0
 //     2         2
 //     3         3
 //     4         1
 //
 // If the mix has not progressed then no blocks are taken from I2.
 // If the mix has progressed so that two blocks should be filled then
 // blocks  4 and 0  are taken from I2.
 // If the mix is such that 4 blocks from I2 are required then blocks
 // 4,0,2 and 3 are filled.
 // At the end of the mix all blocks are needed they are taken in order
 // 4,0,2,3,1
 //
 //
 // The function works by making two lists:
 // (1) Containing the ordered integers 0,1,2,3,4,...(nb-1)
 // (2) "nb" randomly generated double precision integers.
 // Sorting these lists together so that the random double precision list
 // is ordered by magnitude will result in the list of integers becoming
 // rnadomly orderded but VITALLY it will still contain the integers
 // 0 <= i < nb !!!
 //
 long i,*blockid;
 double *blockV;
 if((blockid = (long *)malloc(nb*sizeof(long))) == NULL)return NULL;
 if((blockV = (double *)malloc(nb*sizeof(double))) == NULL){
   free(blockid); return NULL;
 }
 for(i=0;i<nb;i++)
 {
   blockid[i]=i;
   blockV[i]=Random();
 }
 SortArray(nb,blockid,blockV);
 free(blockV);
 return blockid;
}

void RandomBlockMix(
     lpPIXEL I1,      // first input stream  (pointer to framebuffer)
     lpPIXEL I2,      // second input stream
     lpPIXEL O,       // pointer to frame buffer for result
     long    Width,   // width of images in framebuffers
     long    Height,  // height of images in frame buffers
     float   t)       // if t=0 then  O=I1  if t=1 then O=I2
{
 //
 // Mixes two images together by creating an ever increasing number of
 // "squares" of I2 in I1 until eventually I2 replaces I1. The squares
 // occur at random over the screen and can be any size from 1 pixel up.
 //
 // nnx and nny set the size of each random unit (width and height)
 // when they are set to 1 the mix occurs on a pixel basis.
 //
 long nnx=1, nny=1;
 long i,ii,j,jj,k,l,m,nx,ny,n,nb,id,*blockid;
 lpPIXEL s,bb;
 nx = nnx; ny=nny;
 nnx = Width/nx;
 nny = Height/ny;
 //
 // Count the number of blocks uses "nb" holds the result
 //
 ii=0;
 for(i=0;i<Height;i+=ny)
 {
   jj=0;
   for(j=0;j<Width;j+=nx)
   {
     jj++;
   }
   ii++;
 }
 nb=ii*jj;
 //
 // n is the number of blocks of I2 to appear in I1 (proprotional to t)
 //
 n=(long)((float)nnx*(float)nny*t);
 //
 // This is the key step to the effect "GenerateRandomSelection" returns
 // an array of integers that identify the order in which the "blocks"
 // should be filled! (See comments in function for further details).
 //
 if((blockid=GenerateRandomSelection(nb)) == NULL){
   //
   // The function failed so no effect can be created
   //
   return;
 }
 //
 // Copy I1 to O the parts of O that will show the emerging I2 will be
 // over-written in the next step.
 //
 memcpy(O,I1,Width*Height*sizeof(PIXEL));
 //
 // Overwrite Copy I2 to O for the "n" blocks appropriate to parameter "t"
 //
 if(n > 0)for(m=0;m<n;m++)
 {
   //
   // Retrieve the identity of which rectangle to fill for counter "m"
   //
   id=blockid[m];
   i=id/nnx;
   j=(id-(i*nnx));
   i=i*ny;
   j=j*nx;
   for(k=i;k<i+ny;k++)for(l=j;l<j+nx;l++)
   {
     if(l < Width && k < Height)
     {
       s=(O+k*Width+l);
       bb=(I2+k*Width+l);
       s->Red  =(unsigned char)bb->Red  ;
       s->Green=(unsigned char)bb->Green;
       s->Blue =(unsigned char)bb->Blue ;
     }
   }
 }
 free(blockid);
 return;
}
