//
// Pseudo 3D operations on 2D images recorded in a framebuffer.
//
//
// In all these effects the X axis runs left to right along the top of the
// image originating at the top/left corner, the Y axis runs down the
// image from the top/left corner and the Z axis runs INTO the image from
// the top/left corner (i.e. Z = vector product of X and Y [Z = X x Y])
//

//
// Include file that contains definitions for the PIXEL data structure
//

#include "chap8.h"

//
// Data type vector and matrix
//
typedef float VECTOR[3];
typedef float MATRIX[4][4];

//
// External 3D functions in file chap7_4.cpp
//
long Pseudo3dMap(VECTOR p, VECTOR Mp, VECTOR Mx, VECTOR My,
                 lpPIXEL S,
                 long Width, long Height,
                 float *red, float *green, float *blue);
void GetPointInMapPlane(VECTOR pc, VECTOR d,
                        VECTOR Mp, VECTOR Mn,
                        VECTOR p);
//
// External matrix functions in file chap5_5.cpp
//
// multiply two 4x4 matrices together Mr = Ma x Mb
extern void  MultiplyMatrix(MATRIX Ma, MATRIX Mb, MATRIX Mr);
// transform vector by multiplying it with a matrix  Vr = M x Va
extern void  TransformVector(MATRIX M, VECTOR Va, VECTOR Vr);
// create translation matrix for translation v
extern void Translate(MATRIX, VECTOR t);
// create scaling matrix by sx,sy,sz
extern void Scale(MATRIX, float sx, float sy, float sz);
// create rotation matrix of angle A round Z axis
extern void RotateRoundZ(MATRIX, float A);
// create rotation matrix of angle A round X axis
extern void RotateRoundX(MATRIX, float A);
// create rotation matrix of angle A round Y axis
extern void RotateRoundY(MATRIX, float A);
extern void RotateRoundAxis(MATRIX,  VECTOR,  VECTOR, float );
// subtract two vectors R = A - B
extern void SubtractVectors(VECTOR A, VECTOR B, VECTOR R);
// normalize the lenght of a vector
extern void Normalize(VECTOR A);

///////// The following are standard functions and definitiions that are
///////// usually part of a compilers library.
///////// Substitute then with the ones you wish to use.


extern void *malloc(long X);      // allocate X bytes, return pointer to it.
extern void free(void *P);        // release memory block pointed to by P
extern float sqrt(float D);       // return floating point square root of  D
//
//  Quick copy function to copy a block of memory to another location
//
extern void memcpy(void *Destination, void *Source, long NumberOfBytes);

#define PI 3.141592657
//
// Maximum and Minimum operations
//
#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif


//////////////////////////////////////////////////////////////////////////
//////// Now the effects /////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void BasicPictureInPicture(
     lpPIXEL I1,      // first input stream  (pointer to framebuffer)
     lpPIXEL I2,      // second input stream
     lpPIXEL O,       // pointer to frame buffer for result
     long    Width,   // width of images in framebuffers
     long    Height,  // height of images in frame buffers
     float   t)       // if t=0 then  O=I1  if t=1 then O=mixed result
{
 //
 // This effect creates a small scaled down version of I1 inside the
 // image I2. A small black border round I1 is added and the effect
 // progressively reduces the size of the rectangle over time.
 //
 // nx and ny are the screen coordinates where the picture in picture
 // ends up
 //
 long nx=10,ny=10;
 VECTOR p,Mp,Mx,My;
 float r,g,b;
 long i,j,w;
 //
 // Set up vectors in 3D space to where the three corners of the image
 // are moved by the video effect.
 // In this case this is a rectangle that decreases in size as the effect
 // progresses.
 //
 Mp[2]=Mx[2]=My[2]=p[2]=0.0;
 Mx[1]=0.0; My[0]=0.0;
 Mp[0]=t*(float)nx;
 Mp[1]=t*(float)ny;
 Mx[0]=Mp[0]+(float)Width+t*(float)(Width/2-Width);
 My[1]=Mp[1]+(float)Height+t*(float)(Height/2-Height);
 //
 // For each pixel in the output buffer determine the color value
 //
 for(i=0;i<Height;i++)for(j=0;j<Width;j++)
 {
   //
   // "p" is the point for which the color is to be determined i.e.
   // at screen coordinate (j,i) note that because the image map still
   // lies in the plane of the screen after transformation "p" remains in
   // in the plane of the image map.
   //
   p[0]=(float)j; p[1]=(float)i;
   //
   // Pseudo3dMap performs the inverse mapping from rectangle given by
   // top/left point "Mp" bottom/right point (Mp+Mx+My) to the
   // rectangle top/left (0,0) bottom/right (Width,Height) it then applies
   // this mapping to "p" and returns interpolated RGB color values
   // at the location in image "I2" where "p" is mapped.
   //
   if((w=Pseudo3dMap(p,Mp,Mx,My,I1,Width,Height,&r,&g,&b)) > 0)
   {
     //
     //  "p" falls inside the region where "I1" is mapped and thus
     //  Pseudo3dMap() returns interpolated RGB values.
     //  Place these in the output buffer.
     //
     O->Red  =(unsigned char)r;
     O->Green=(unsigned char)g;
     O->Blue =(unsigned char)b;
   }
   else if(w < 0)
   {
     //
     // "p" is in a boundary region round the mapped image
     //
     O->Red  =0;
     O->Green=0;
     O->Blue =0;
   }
   else
   {
     //
     // "p" is not in a mapped region so copy the pixel from the second
     // image I2 to the output buffer.
     //
     O->Red  =(unsigned char)I2->Red  ;
     O->Green=(unsigned char)I2->Green;
     O->Blue =(unsigned char)I2->Blue ;
   }
   I2++;
   O++;
 }
}

void ZoomOutFromImage(
     lpPIXEL I1,      // first input stream  (pointer to framebuffer)
     lpPIXEL I2,      // second input stream
     lpPIXEL O,       // pointer to frame buffer for result
     long    Width,   // width of images in framebuffers
     long    Height,  // height of images in frame buffers
     float   t)       // if t=0 then  O=I1  if t=1 then O=mixed result
{
 //
 // This effect creates the illusion of the image receeding away into
 // the distance. The effect is generated by gradually scaling down the
 // size of the image and moving it towards the center of the screen.
 //
 // nx and ny are the screen coordinates where the picture in picture
 // ends up
 //
 long nx=10,ny=10;
 VECTOR p,Mp,Mx,My;
 float r,g,b;
 long i,j,w;
 //
 // Set up vectors in 3D space to where the three corners of the image
 // are moved by the video effect.
 // In this case this is a rectangle that decreases in size to zero and
 // moves towards the center of the screen
 // progresses.
 //
 Mp[2]=Mx[2]=My[2]=p[2]=0.0;
 Mx[1]=0.0; My[0]=0.0;
 Mx[0]=max(1.0,(float)Width-t*(float)(Width));
 My[1]=max(1.0,(float)Height-t*(float)(Height));
 Mp[0]=t*(float)Width/2;
 Mp[1]=t*(float)Height/2;
 //
 // For each pixel in the output buffer determine the color value
 //
 for(i=0;i<Height;i++)for(j=0;j<Width;j++)
 {
   //
   // "p" is the point for which the color is to be determined i.e.
   // at screen coordinate (j,i) note that because the image map still
   // lies in the plane of the screen after transformation "p" remains in
   // in the plane of the image map.
   //
   p[0]=(float)j; p[1]=(float)i;
   //
   // Pseudo3dMap performs the inverse mapping from rectangle given by
   // top/left point "Mp" bottom/right point (Mp+Mx+My) to the
   // rectangle top/left (0,0) bottom/right (Width,Height) it then applies
   // this mapping to "p" and returns interpolated RGB color values
   // at the location in image "I2" where "p" is mapped.
   //
   if((w=Pseudo3dMap(p,Mp,Mx,My,I1,Width,Height,&r,&g,&b)) > 0)
   {
     //
     //  "p" falls inside the region where "I1" is mapped and thus
     //  Pseudo3dMap() returns interpolated RGB values.
     //  Place these in the output buffer.
     //
     O->Red  =(unsigned char)r;
     O->Green=(unsigned char)g;
     O->Blue =(unsigned char)b;
   }
   else
   {
     //
     // "p" is not in a mapped region so copy the pixel from the second
     // image I2 to the output buffer.
     //
     O->Red  =(unsigned char)I2->Red  ;
     O->Green=(unsigned char)I2->Green;
     O->Blue =(unsigned char)I2->Blue ;
   }
   I2++;
   O++;
 }
}

void RotatePictureRoundZaxis(
     lpPIXEL I1,      // first input stream  (pointer to framebuffer)
     lpPIXEL I2,      // second input stream
     lpPIXEL O,       // pointer to frame buffer for result
     long    Width,   // width of images in framebuffers
     long    Height,  // height of images in frame buffers
     float   t)       // if t=0 then  O=I1  if t=1 then O=mixed result
{
 //
 // nx is the angle through which the image will be rotated about an axis
 // perpendicular to the screen.
 //
 long nx=360;
 //
 // In all these effects
 //
 VECTOR p,pc,d,Mn={0.0,0.0,1.0},Mx,My,Mp,dx;
 float r,g,b;
 long i,j;
 float Tx[4][4],Ta[4][4],Tb[4][4],angle;
 angle = nx*t;
 angle *= PI/180.0;
 //
 // Set up vectors in 3D space to where the three corners of the image
 // are moved by the video effect. First place Mp Mx My at the corners
 // of the image.
 //
 Mp[0]=0.0;          Mp[0]=0.0;           Mp[0]=0.0;
 Mx[0]=(float)Width; Mx[1]=0.0;           Mx[2]=0.0;
 My[0]=0.0,          My[1]=(float)Height; My[2]=0.0;
 //
 // Build a transformation matrix that represents the effect.
 // Because rotation is about the center of the screen a translation must
 // be done to move the origin to the center of the screen and after rotation
 // it must be undone.
 //
 dx[0]=-(float)Width*0.5; dx[1]=-(float)Height*0.5; dx[2]=0.0;
 Translate(Ta,dx);
 RotateRoundZ(Tx,angle);
 MultiplyMatrix(Tx,Ta,Tb);
 dx[0]=(float)Width*0.5; dx[1]=(float)Height*0.5; dx[2]=0.0;
 Translate(Tx,dx);
 MultiplyMatrix(Tx,Tb,Ta);
 //
 // Matrix "Ta" now represents the transformation for the effect. It
 // must be applied to the points "Mp" "Mx" and "My" and importantly
 // any rotational component must also apply to the normal to the
 // plane "Mn" (this can be done in several ways but in the case "Mn"
 // is not effected by "Ta" so it can be left alone).
 //
 TransformVector(Ta,Mp,Mp);
 TransformVector(Ta,Mx,Mx);
 TransformVector(Ta,My,My);
 //
 // "pc" is the equivalent "viewpoint" from where the effect is being
 // viewed. It's Z coordinate is offest 'in front of' the screen.
 //
 pc[0] = (float)(Width/2);
 pc[1] = (float)(Height/2);
 pc[2] = (float)(-Width*2);
 //
 // For each pixel in the output buffer determine the color value
 //
 for(i=0;i<Height;i++)for(j=0;j<Width;j++)
 {
   //
   // "d" is the 'look' direction from the viewpoint at "pc" it passes
   // through pixel (j,i). The vector in direction "d" from "pc" will
   // 'see' point "p" located on the image plane.
   //
   d[0]  = (float)(j-Width/2);
   d[1]  = (float)(i-Height/2);
   d[2]  = -pc[2];
   GetPointInMapPlane(pc,d,Mp,Mn,p);
   //
   // Pseudo3dMap performs the inverse mapping from rectangle given by
   // top/left point "Mp" bottom/right point (Mp+Mx+My) to the
   // rectangle top/left (0,0) bottom/right (Width,Height) it then applies
   // this mapping to "p" and returns interpolated RGB color values
   // at the location in image "I2" where "p" is mapped.
   //
   if(Pseudo3dMap(p,Mp,Mx,My,I1,Width,Height,&r,&g,&b) > 0)
   {
     //
     //  "p" falls inside the region where "I1" is mapped and thus
     //  Pseudo3dMap() returns interpolated RGB values.
     //  Place these in the output buffer.
     //
     O->Red  =(unsigned char)r;
     O->Green=(unsigned char)g;
     O->Blue =(unsigned char)b;
   }
   else
   {
     //
     // "p" is not in a mapped region so copy the pixel from the second
     // image I2 to the output buffer.
     //
     O->Red  =(unsigned char)I2->Red  ;
     O->Green=(unsigned char)I2->Green;
     O->Blue =(unsigned char)I2->Blue ;
   }
   I2++;
   O++;
 }
}

void Tumble(
     lpPIXEL I1,      // first input stream  (pointer to framebuffer)
     lpPIXEL I2,      // second input stream
     lpPIXEL O,       // pointer to frame buffer for result
     long    Width,   // width of images in framebuffers
     long    Height,  // height of images in frame buffers
     float   t)       // if t=0 then  O=I1  if t=1 then O=mixed result
{
 //
 // "nx" is the angle through which the image will be tumbled.  To tumble the
 // the image it is simultaneously rotated about several axes.
 // "ny" is the % size the final image takes relative to the screen size
 //
 long nx=360,ny=10;
 //
 // In all these effects
 //
 VECTOR p,pc,d,Mn={0.0,0.0,1.0},Mx,My,Mp,dx;
 VECTOR wt,Mw={1.0,0.0,0.0},Mz={0.0,0.0,0.0},temp;
 float r,g,b;
 long i,j;
 float Tx[4][4],Ta[4][4],Tb[4][4],Ts[4][4],angle,ss;
 angle = nx*t;
 angle *= PI/180.0;
 //
 // Set up vectors in 3D space to where the three corners of the image
 // are moved by the video effect. First place Mp Mx My at the corners
 // of the image.
 //
 Mp[0]=0.0;          Mp[0]=0.0;           Mp[0]=0.0;
 Mx[0]=(float)Width; Mx[1]=0.0;           Mx[2]=0.0;
 My[0]=0.0,          My[1]=(float)Height; My[2]=0.0;
 //
 // Build a transformation matrix that represents the effect. Rotate
 // round Z axis and then about an axis that originally is along the
 // X axis but is itself rotated as the image rotates round Z.
 //
 dx[0]=-(float)Width*0.5; dx[1]=-(float)Height*0.5; dx[2]=0.0;
 Translate(Ta,dx);
 RotateRoundZ(Tx,angle);
 TransformVector(Tx,Mw,wt);
 RotateRoundAxis(Tx,Mz,wt,angle*180.0/PI);
 MultiplyMatrix(Tx,Ta,Tb);
 ss=(float)ny * 0.01;
 ss=1.0-t*(1.0-ss);
 Scale(Ts,ss,ss,ss);
 MultiplyMatrix(Ts,Tb,Ta);
 dx[0]=(float)Width*0.5; dx[1]=(float)Height*0.5; dx[2]=0.0;
 Translate(Tx,dx);
 MultiplyMatrix(Tx,Ta,Tb);
 //
 // Tb is the transformation matrix representing the effect
 //
 TransformVector(Tb,Mp,Mp);
 TransformVector(Tb,Mx,Mx);
 TransformVector(Tb,My,My);
 //
 // The normal "Mn" is must also be subjected to the transformation.
 // Because it is direction vector it must transformed by first converting
 // it into a position vector and after transformation converted back to a
 // direction vector.
 //
 TransformVector(Tb,Mn,temp);
 SubtractVectors(temp,Mp,Mn);
 Normalize(Mn);
 //
 // "pc" is the equivalent "viewpoint" from where the effect is being
 // viewed. It's Z coordinate is offest 'in front of' the screen.
 //
 pc[0] = (float)(Width/2);
 pc[1] = (float)(Height/2);
 pc[2] = (float)(-Width*2);
 //
 // For each pixel in the output buffer determine the color value
 //
 for(i=0;i<Height;i++)for(j=0;j<Width;j++)
 {
   //
   // "d" is the 'look' direction from the viewpoint at "pc" it passes
   // through pixel (j,i). The vector in direction "d" from "pc" will
   // 'see' point "p" located on the image plane.
   //
   d[0]  = (float)(j-Width/2);
   d[1]  = (float)(i-Height/2);
   d[2]  = -pc[2];
   GetPointInMapPlane(pc,d,Mp,Mn,p);
   //
   // Pseudo3dMap performs the inverse mapping from rectangle given by
   // top/left point "Mp" bottom/right point (Mp+Mx+My) to the
   // rectangle top/left (0,0) bottom/right (Width,Height) it then applies
   // this mapping to "p" and returns interpolated RGB color values
   // at the location in image "I2" where "p" is mapped.
   //
   if(Pseudo3dMap(p,Mp,Mx,My,I1,Width,Height,&r,&g,&b) > 0)
   {
     //
     //  "p" falls inside the region where "I1" is mapped and thus
     //  Pseudo3dMap() returns interpolated RGB values.
     //  Place these in the output buffer.
     //
     O->Red  =(unsigned char)r;
     O->Green=(unsigned char)g;
     O->Blue =(unsigned char)b;
   }
   else
   {
     //
     // "p" is not in a mapped region so copy the pixel from the second
     // image I2 to the output buffer.
     //
     O->Red  =(unsigned char)I2->Red  ;
     O->Green=(unsigned char)I2->Green;
     O->Blue =(unsigned char)I2->Blue ;
   }
   I2++;
   O++;
 }
}
