//
// The video processing function Pesudo3dMap( ) performs 3D mapping
// by returning an RGB pixel value at a given location (column,row) in
// the output imagebuffer. The action of the function is decribed in
// the text.
//

// Data type vector

typedef float VECTOR[3];

////////// The following functions will need to be provided to fulfil the
////////// described action.

extern float  Dot(VECTOR A, VECTOR B);
//  Scale the vector A with the scalar s,   R = sA
extern void   ScaleVector(float s,VECTOR A, VECTOR R);
// add two vectors R = A + B
extern void AddVectors(VECTOR A, VECTOR B, VECTOR R);
// subtract two vectors R = A - B
extern void SubtractVectors(VECTOR A, VECTOR B, VECTOR R);

///////// The following are standard functions that are usually part
///////// of a compilers library.
///////// Substitute then with the ones you wish to use.

extern float fabs(float x);   // return absolute value

//
// Include file that contains definitions for the PIXEL data structure
//

#include "chap8.h"

///////////////////////////////////////////////////////////////////
//////////////// Functions ////////////////////////////////////////

long PointRelativeTo(VECTOR vi,
                     VECTOR p0, VECTOR px, VECTOR py,
                     float *alpha, float *beta)
{
// vi is position vector of point to determine where it lies inside the
// polygon formed by:  p0 - px
//                     |    |
//                     py -
// Note: (vi must lies in the plane of the polygon)
// "alpha" and "beta" are coordinates of "vi" in the rectangles frame
// of reference. Thus if a rectangular image is mapped into the rectangle
// (p0,px,py) alpha and beta can be used to interpolate a pixel value
// visible on screen at the equivalent of point "vi"
// Note: At p0 (alpha,beta) = (0,0)  Top-Left corner
//       At py (alpha,beta) = (0,1)  Bottom-Left corner
//       At px (alpha,beta) = (1,0)  Top-Right corner
//
// Function returns  1 if successful      0 if failed
//
 VECTOR w,u,v;
 float a,b,ab,den,uu,vv,uv,wu,wv;
 SubtractVectors(vi,p0,w);
 SubtractVectors(px,p0,u);
 SubtractVectors(py,p0,v);
 uu=Dot(u,u);
 vv=Dot(v,v);
 uv=Dot(u,v);
 wu=Dot(w,u);
 wv=Dot(w,v);
 den=uu*vv-uv*uv;
 if(fabs(den) < 1.0e-6)return 0; // denominator zero
 *alpha = (wu*vv - wv*uv)/den;
 *beta  = (wv*uu - wu*uv)/den;
 return 1;
}

long Pseudo3dMap(VECTOR p,    // Point (in the plane of map) to find color
                              // values for.
                 VECTOR Mp,   // Top/left corner of the quadilateral to where
                              // the video effect has transformed the
                 VECTOR Mx,   //
                 VECTOR My,   //
                 lpPIXEL S,   // Pointer to the image distorted by the effect
                 long Width,  // Width of image recorded in S
                 long Height, // Height of image recorded in S
                 float *red,  //
                 float *green,// Returned color values
                 float *blue) //
{
 //
 // Point "p" is determined as linear combination of vectors
 // p = alpha * (Mx-Mp) + beta * (My-Mp) Function PointRelativeTo()
 // returns "alpha" and "beta".
 // "alpha" and "beta" are then used to interpolate color values for return.
 //
 // Function return a code as follows:
 // -1 if "p" is outside the mapped quadilateral
 //  0 of "p" is outside the mapped quadilateral but it lies in a "border"
 //    region round the rectangular map
 //  1 "p" is inside the map
 //
 long  ialpha,ibeta,i,j,k,kl,kr,kt,kb,pixel;
 float alpha,beta,a,b,cc0,cc1,cc2,cl0,cl1,cl2,
       cr0,cr1,cr2,ct0,ct1,ct2,cb0,cb1,cb2;
 unsigned char *pP,*pR,*pG,*pB;
 j=Width;
 //
 // Determine the location relative to the image of the point "p" which
 // is assumed to lie in the plane of the image.
 //
 if(!PointRelativeTo(p,Mp,Mx,My,&alpha,&beta))return -2;
 //
 // Now that coordinates in the frame of reference of the original image
 // have been determined use bi-linear interpolation to obtain appropriate
 // color values for "red" "green" and "blue"
 //
 if(alpha < -0.01 || alpha > 1.01)return -1;
 if(beta  < -0.02 || beta  > 1.02)return -1;
 alpha=Width*alpha;
 beta =Height*beta;
 ialpha = (long)alpha; alpha -= (float)ialpha;
 ibeta  = (long)beta;  beta  -= (float)ibeta;
 if(ialpha < 0 || ialpha >= Width)return 0;
 if(ibeta  < 0 || ibeta  >= Height)return 0;
 k =ibeta*j+ialpha;
 if(ialpha <   1)kl=k; else kl=k-1;
 if(ialpha > j-2)kr=k; else kr=k+1;
 if(ibeta  <   1)kb=k; else kb=k-j;
 if(ibeta  > Height-2)kt=k; else kt=k+j;
 cc0 = (float)S[k ].Red; cc1 = (float)S[k].Green;  cc2 = (float)S[k].Blue;
 cl0 = (float)S[kl].Red; cl1 = (float)S[kl].Green; cl2 = (float)S[kl].Blue;
 cr0 = (float)S[kr].Red; cr1 = (float)S[kr].Green; cr2 = (float)S[kr].Blue;
 ct0 = (float)S[kt].Red; ct1 = (float)S[kt].Green; ct2 = (float)S[kt].Blue;
 cb0 = (float)S[kb].Red; cb1 = (float)S[kb].Green; cb2 = (float)S[kb].Blue;
 if(alpha > 0.5){
   alpha = alpha-0.5;
   if(beta > 0.5){
     beta = beta-0.5;
     *red   = (cr0-cc0)*alpha+(ct0-cc0)*beta+cc0;
     *green = (cr1-cc1)*alpha+(ct1-cc1)*beta+cc1;
     *blue  = (cr2-cc2)*alpha+(ct2-cc2)*beta+cc2;
   }
   else{
     beta = 0.5-beta;
     *red   = (cr0-cc0)*alpha+(cb0-cc0)*beta+cc0;
     *green = (cr1-cc1)*alpha+(cb1-cc1)*beta+cc1;
     *blue  = (cr2-cc2)*alpha+(cb2-cc2)*beta+cc2;
   }
 }
 else{
   alpha = 0.5-alpha;
   if(beta > 0.5){
     beta = beta-0.5;
     *red   = (cl0-cc0)*alpha+(ct0-cc0)*beta+cc0;
     *green = (cl1-cc1)*alpha+(ct1-cc1)*beta+cc1;
     *blue  = (cl2-cc2)*alpha+(ct2-cc2)*beta+cc2;
   }
   else{
     beta = 0.5-beta;
     *red   = (cl0-cc0)*alpha+(cb0-cc0)*beta+cc0;
     *green = (cl1-cc1)*alpha+(cb1-cc1)*beta+cc1;
     *blue  = (cl2-cc2)*alpha+(cb2-cc2)*beta+cc2;
   }
 }
 return 1;
}

void GetPointInMapPlane(
  VECTOR pc, // point of origin of viewing ray (center of screen)
  VECTOR d,  // direction of viewing ray
  VECTOR Mp, // point at top/left corner of transformed image map
  VECTOR Mn, // normal to the plane of transformed image map
  VECTOR p)  // point in plane of transformed image map
{
 //
 // In some of the 3D effects the image appears to move out of the plane
 // of the screen. In those effects it is necessary to project a viewing
 // ray from the center of the screen at "pc" in a direction "d" through
 // the pixel of interest and find the point where it intersects the
 // the plane in which the transformed image lies. To do this
 // requires a calculation involving the top left corner of the transformed
 // plane "Mp" and a unit vector "Mn" normal to the transformed plane.
 //
 // The function returns "p" a position vector to a point lying in the plane
 // of the image which may then be passed to the function "PointRelativeTo()"
 // for determination of RGB values.
 //
 float mu;
 VECTOR dp;
 SubtractVectors(Mp,pc,dp);
 mu=Dot(Mn,dp)/Dot(Mn,d);
 ScaleVector(mu,d,d);
 AddVectors(pc,d,p);
}
