//
// Mixcellaneous video effects including artisitic effects
//
//
// Include file that contains definitions for the PIXEL data structure
//

#include "chap8.h"

///////// The following are standard functions that are usually part
///////// of a compilers library.
///////// Substitute then with the ones you wish to use.


extern void *malloc(long X);      // allocate X bytes, return pointer to it.
extern void free(void *P);        // release memory block pointed to by P
extern float sqrt(float D);       // return floating point square root of  D
extern float sin(float a);        // return the sin of angle "a" (radians)
extern float cos(float a);        // return the cosine of angle "a" (radians)
//
//  Quick copy function to copy a block of memory to another location
//
extern void memcpy(void *Destination, void *Source, long NumberOfBytes);
//
// Random number functions
//
void srand(unsigned int);  // initialize the random number generator
int  rand(void);           // generate random number "C" language spec
                           // says this is in range [-32767 +32767]
//
// Modulo functions
//
double fmod(double, double); // floating point modulo
//
// integer modulo - implemented as macro
//
#define imod(z,a) ((z) - ((z)/(a))*(a))
//
// Define the NULL
//
#define NULL 0
//
// Maximum and Minimum operations
//
#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
//
// Define PI
//
#define PI 3.141592657

//////////////////////////////////////////////////////////////////////
////////////////// Helper functions //////////////////////////////////
//////////////////////////////////////////////////////////////////////

//
// The first two functions perform conversion between the RGB description
// of a color and the HSV (H)ue (S)aturation (V)alue system.
// The RGB uses one unsigned byte to record the Red, Green and Blue colors
// each in the range (0 - 255).
// the HSV system uses floating point numbers in the intervals:
//  Hue    (0.0 <= h <360.0) Represents the "color" on a wheel 360 degrees
//                           goes round the wheel starting at Red and ending
//                           at Blue.
//  Saturation ( 0.0 <= s < 1.0)  This represents the "ammound" of color
//                                zero saturation = grey scale image
//  Value (0.0 <= v < 1.0)  The "brightness"  0=black 1=full brightness
//

void RGBtoHSV(unsigned char r,
              unsigned char g,
              unsigned char b,
              float *h,
              float *s,
              float *v)
{
 //
 // Convert from an RGB value to an HSV representation of a color.
 //
 float rr,gg,bb,hh,ss,vv,maxcol,mincol;
 rr=(float)r; gg=(float)g; bb=(float)b;
 rr /= 255.0; gg /= 255.0; bb /= 255.0;
 maxcol=max(rr,max(gg,bb));
 mincol=min(rr,min(gg,bb));
 vv=maxcol;
 if(maxcol == mincol){ss=0; hh=0;}
 else{
  ss=(maxcol-mincol)/maxcol;
  if(mincol == bb)
    hh = 120.0*(gg-mincol)/(rr+gg-2.0*mincol);
  else if(mincol == rr)
    hh = 120.0*(1.0+(bb-mincol)/(bb+gg-2.0*mincol));
  else
    hh = 120.0*(2.0+(rr-mincol)/(rr+bb-2.0*mincol));
 }
  *h=hh;
  *s=ss;
  *v=vv;
}

void HSVtoRGB(float h,
              float s,
              float v,
              unsigned char *r,
              unsigned char *g,
              unsigned char *b)
{
 //
 // Convert from an HSV value to an RGB representation of a color.
 //
 float rr,gg,bb,hh,ss,vv,maxcol,mincol;
 hh=(float)h; ss=(float)s; vv=(float)v;
 mincol=vv*(1.0-ss);
 if(hh <= 120.0)
 {
   bb=mincol;
   if(hh <=  60){rr=vv; gg=mincol+hh*(vv-mincol)/(120.0-hh);}
   else         {gg=vv; rr=mincol+(120.0-hh)*(vv-mincol)/hh;}
 }
 else if( hh <= 240)
 {
   rr=mincol;
   if(hh <= 180){gg=vv; bb=mincol+(hh-120.0)*(vv-mincol)/(240.0-hh);}
   else         {bb=vv; gg=mincol+(240.0-hh)*(vv-mincol)/(hh-120.0);}
 }
 else
 {
   gg=mincol;
   if(hh <= 300){bb=vv; rr=mincol+(hh-240.0)*(vv-mincol)/(360.0-hh);}
   else         {rr=vv; bb=mincol+(360.0-hh)*(vv-mincol)/(hh-240.0);}
 }
 rr = rr*255.0; *r=(unsigned char)min(255,rr);
 gg = gg*255.0; *g=(unsigned char)min(255,gg);
 bb = bb*255.0; *b=(unsigned char)min(255,bb);
}


//////////////////////////////////////////////////////////////////////////
//////// Now the effects /////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////


static void FadeToMonochrome(
     lpPIXEL I1,      // first input stream  (pointer to framebuffer)
     lpPIXEL I2,      // second input stream
     lpPIXEL O,       // pointer to frame buffer for result
     long    Width,   // width of images in framebuffers
     long    Height,  // height of images in frame buffers
     float   t)       // if t=0 then  O=I1  if t=1 then O=I2
{
 //
 // Gradually remove the color from input I1  (all color gone when t=1)
 //
 long i,j;
 float mono;
 for(i=0;i<Height;i++)for(j=0;j<Width;j++)
 {
   mono=0.3333333*((float)I1->Red +(float)I1->Green+(float)I1->Blue);
   O->Red   = (unsigned char)min(255.0,(float)(I1->Red  )*(1.0-t)+mono*t);
   O->Green = (unsigned char)min(255.0,(float)(I1->Green)*(1.0-t)+mono*t);
   O->Blue  = (unsigned char)min(255.0,(float)(I1->Blue )*(1.0-t)+mono*t);
   I1++;
   O++;
 }
}

void FadeToNegative(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // gradually make a negative image of I1
 //
 long i,j;
 for(i=0;i<Height;i++)for(j=0;j<Width;j++){
   O->Red   = (unsigned char)((float)I1->Red  *(1.0-t)+(float)(255 - I1->Red  )*t);
   O->Green = (unsigned char)((float)I1->Green*(1.0-t)+(float)(255 - I1->Green)*t);
   O->Blue  = (unsigned char)((float)I1->Blue *(1.0-t)+(float)(255 - I1->Blue )*t);
   I1++;
   O++;
 }
}

void QuantizeColor(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Fix the brightness and saturation of an image into certain bands
 // This function mimics the "colorizing" effect of digitial video mixers
 // channed I2 is unused
 //
 long i,j,nx;
 unsigned char R,G,B;
 float h,s,v;
 for(i=0;i<Height;i++)for(j=0;j<Width;j++)
 {
   RGBtoHSV(I1->Red  ,I1->Green,I1->Blue,&h,&s,&v);
   if     (v > 0.8)v=1.0;
   else if(v > 0.6)v=0.8;
   else if(v > 0.4)v=0.6;
   else if(v > 0.2)v=0.4;
   else v=0.1;
   if     (s > 0.8)s=1.0;
   else if(s > 0.6)s=0.8;
   else if(s > 0.4)s=0.6;
   else if(s > 0.2)s=0.4;
   else s=0.1;
   HSVtoRGB(h,s,v,&R,&G,&B);
   O->Red   = (unsigned char)((float)I1->Red  *(1.0-t)+(float)R*t);
   O->Green = (unsigned char)((float)I1->Green*(1.0-t)+(float)G*t);
   O->Blue  = (unsigned char)((float)I1->Blue *(1.0-t)+(float)B*t);
   I1++;
   O++;
 }
}

void RotateHue(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // By rotating the Hue vector the image can gain the illusing of
 // "color cycling" For example: If a Red raster is give as I1 the
 // output can be made to change color through yellow, green, blue,
 // purple an back to red over the time interval as 0 <= t < 1
 //
 // angle specifies the maximum change in hue. 360.0 is a full cycle of
 // all colors in the spectrum
 //
 float angle=360.0;
 long i,j;
 unsigned char R,G,B;
 float h,s,v;
 for(i=0;i<Height;i++)for(j=0;j<Width;j++)
 {
   RGBtoHSV(I1->Red  ,I1->Green,I1->Blue ,&h,&s,&v);
   h += angle*t;
   while(h > 360.0)h -= 360.0;
   HSVtoRGB(h,s,v,&R,&G,&B);
   O->Red   = (unsigned char)min(255,R);
   O->Green = (unsigned char)min(255,G);
   O->Blue  = (unsigned char)min(255,B);
   I1++;
   O++;
 }
}

void Pixelate(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Make the input image look like it is made up of large pixels.
 // The color value for each large pixel is made by averaging all the
 // actual pixels that lie within it. "nxx" and "nyy" can be adjusted to
 // specify the maximum number of big pixels across and down the screen.
 //
 long nxx=10,nyy=10;
 long i,j,k,l,nx,ny,n;
 float Rav,Gav,Bav;
 lpPIXEL s;
 nx = Width/nxx; ny=Height/nyy;
 if(nx*nxx < Width)nx++;
 if(ny*nyy < Height)ny++;
 nx=1+t*(nx-1);
 ny=1+t*(ny-1);
 for(i=0;i<Height;i+=ny)for(j=0;j<Width;j+=nx)
 {
   //
   // Calculate the average
   //
   Rav=Gav=Bav=0.0; n=0;
   for(k=i;k<i+ny;k++)for(l=j;l<j+nx;l++)
   {
     if(l < Width && k < Height)
     {
       s=(I1+k*Width+l);
       Rav += (float)s->Red  ;
       Gav += (float)s->Green;
       Bav += (float)s->Blue ;
       n++;
     }
   }
   Rav /= (float)n; Gav /= (float)n; Bav /= (float)n;
   for(k=i;k<i+ny;k++)for(l=j;l<j+nx;l++)
   {
     //
     // Write it into the output buffer
     //
     if(l < Width && k < Height){
       s=(O+k*Width+l);
       s->Red   = (unsigned char)(Rav);
       s->Green = (unsigned char)(Gav);
       s->Blue  = (unsigned char)(Bav);
     }
   }
 }
}

void RandomNoise(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Make an output image that consists of nothing more than random noise.
 // This effect is similar to the picture of a TV when no aerial is connected
 // and no transmitting station is selected.
 //
 // "param" is the maximum percentage of noise inserted. If it is 50% this
 // effect can simulate "poor TV reception" (watching in a snowstorm)

 int param=100;
 float pc,r;
 long i,j,k,l,step,d;
 lpPIXEL s;
 pc = 0.01*(float)param*t;
 step=min(1,Width/320);
 //
 // copy input to output
 //
 memcpy(O,I1,Width*sizeof(PIXEL));
 //
 // add the noise
 //
 for(i=0;i<Height;i+=step)for(j=0;j<Width;j+=step){
   r=(float)rand()/32767.0;
   d=min(255,(long)((rand()*255)/32767));
   for(k=i;k<i+step;k++)for(l=j;l<j+step;l++)
   {
     if(k >= Height || l >= Width)continue;
     s=(O+k*Width+l);
     if(r < pc){
       s->Red   = (unsigned char)d;
       s->Green = (unsigned char)d;
       s->Blue  = (unsigned char)d;
     }
   }
 }
 return;
}

void ScrollImage(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Scroll the image towards the right, parts of the image that disappear
 // off the right edge reappear at the left hand side. To scroll to the
 // left set t=(1-t)   nx is % that constitutes a full scroll
 //
 long nx=100;
 long i,j,w;
 lpPIXEL B,s,b;
 w=(long)((float)((Width*nx)/100)*t);
 B=(lpPIXEL)malloc(Width*sizeof(PIXEL));
 if(B != NULL)
 {
   for(i=0;i<Height;i++)
   {
     s=(I1+i*Width); b=B;
     memcpy(b,s,Width*sizeof(PIXEL));
     s=(O+i*Width+w);
     for(j=w;j<Width;j++)
     {
       s->Red   = b->Red  ;
       s->Green = b->Green;
       s->Blue  = b->Blue ;
       s->Alpha = b->Alpha;
       b++; s++;
     }
     s=(O+i*Width);
     if(w > 0)for(j=0;j<w;j++)
     {
       s->Red   = b->Red  ;
       s->Green = b->Green;
       s->Blue  = b->Blue ;
       s->Alpha = b->Alpha;
       b++; s++;
     }
   }
   free(B);
 }
}

void RollImage(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 //
 // Roll the image down the dislpay. (Simulate TV that suffers loss
 // of vertical hold.
 // To scroll up set t=(1-t)   ny is % that constitutes a full scroll
 //
 long i,j,w,ny=100;
 lpPIXEL B,s,b;
 w=(long)((float)((Height*ny)/100)*t);
 B=(lpPIXEL)malloc(Height*sizeof(PIXEL));
 if(B != NULL)
 {
   for(i=0;i<Width;i++)
   {
     for(j=0;j<Height;j++)
     {
       b=(B+j); s=(I1+j*Width+i);
       memcpy(b,s,sizeof(PIXEL));
     }
     for(j=w;j<Height;j++)
     {
       b=(B+(j-w)); s=(O+j*Width+i);
       s->Red   = b->Red  ;
       s->Green = b->Green;
       s->Blue  = b->Blue ;
       s->Alpha = b->Alpha;
     }
     if(w > 0)for(j=0;j<w;j++)
     {
       b=(B+(Height-w+j)); s=(O+j*Width+i);
       s->Red   = b->Red  ;
       s->Green = b->Green;
       s->Blue  = b->Blue ;
       s->Alpha = b->Alpha;
     }
   }
   free(B);
 }
}

int InterpolateBilinear(
             float alpha, float beta,   // input coordinates (Horiz,Vertic)
             lpPIXEL S,                 // input image
             long Width, long Height,   // dimensions of the image
             float *red,                // Output red green blue
             float *green,              // values (floating point numbers
             float *blue)               // in range (0.0 - 255.0)
{
 //
 // This function performs bi-linear interpolation to return values of the
 // Red, Green and Blue components from pixels in frame buffer "S"
 // mapping at co-ordinates (alpha,beta). If alpha and beta are integers
 // then exact pixel values are returned. If they are not then interpolation
 // with the nearest neighboring pixels is performed.
 //
 // Function returns 1 if (alpha,beta) is inside the image
 // (0 <= alpha < Width) and ( 0 <= beta < Height) otherwise returns 0
 //
 long  ialpha,ibeta,i,j,k,kl,kr,kt,kb,pixel;
 float a,b,cc0,cc1,cc2,cl0,cl1,cl2,
        cr0,cr1,cr2,ct0,ct1,ct2,cb0,cb1,cb2;
 unsigned char *pP,*pR,*pG,*pB;
 j=Width;
 ialpha = (long)alpha; alpha -= (float)ialpha;
 ibeta  = (long)beta;  beta  -= (float)ibeta;
 if(ialpha < 0 || ialpha >= Width)return 0;
 if(ibeta  < 0 || ibeta  >= Height)return 0;
 k =ibeta*j+ialpha;
 if(ialpha <   1)kl=k; else kl=k-1;
 if(ialpha > j-2)kr=k; else kr=k+1;
 if(ibeta  <   1)kb=k; else kb=k-j;
 if(ibeta  > Height-2)kt=k; else kt=k+j;
 cc0 = (float)S[k ].Red; cc1 = (float)S[k ].Green; cc2 = (float)S[k ].Blue;
 cl0 = (float)S[kl].Red; cl1 = (float)S[kl].Green; cl2 = (float)S[kl].Blue;
 cr0 = (float)S[kr].Red; cr1 = (float)S[kr].Green; cr2 = (float)S[kr].Blue;
 ct0 = (float)S[kt].Red; ct1 = (float)S[kt].Green; ct2 = (float)S[kt].Blue;
 cb0 = (float)S[kb].Red; cb1 = (float)S[kb].Green; cb2 = (float)S[kb].Blue;
 if(alpha > 0.5)
 {
   alpha = alpha-0.5;
   if(beta > 0.5)
   {
     beta = beta-0.5;
     *red   = (cr0-cc0)*alpha+(ct0-cc0)*beta+cc0;
     *green = (cr1-cc1)*alpha+(ct1-cc1)*beta+cc1;
     *blue  = (cr2-cc2)*alpha+(ct2-cc2)*beta+cc2;
   }
   else
   {
     beta = 0.5-beta;
     *red   = (cr0-cc0)*alpha+(cb0-cc0)*beta+cc0;
     *green = (cr1-cc1)*alpha+(cb1-cc1)*beta+cc1;
     *blue  = (cr2-cc2)*alpha+(cb2-cc2)*beta+cc2;
   }
 }
 else
 {
   alpha = 0.5-alpha;
   if(beta > 0.5)
   {
     beta = beta-0.5;
     *red   = (cl0-cc0)*alpha+(ct0-cc0)*beta+cc0;
     *green = (cl1-cc1)*alpha+(ct1-cc1)*beta+cc1;
     *blue  = (cl2-cc2)*alpha+(ct2-cc2)*beta+cc2;
   }
   else
   {
     beta = 0.5-beta;
     *red   = (cl0-cc0)*alpha+(cb0-cc0)*beta+cc0;
     *green = (cl1-cc1)*alpha+(cb1-cc1)*beta+cc1;
     *blue  = (cl2-cc2)*alpha+(cb2-cc2)*beta+cc2;
   }
 }
 return 1;
}

void StirImage(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
 {
 //
 // Transform the image as if it was being stirred round the center of the
 // screen. Similar effect to what would happen to a drawing on the surface
 // of a viscous liquid if it is gently stirred round.
 //
 // The parameter nx controls how far orund the image is stirred,
 // nx=360*turns. If linear=1 a linear model is used, this gives a sharper
 // edge to the effect than if the quadratic model (linear=0) is used.
 //
 long linear=0;
 long nx=360;
 long i,j,k,l,X2,Y2,X1,Y1;
 lpPIXEL s;
 float r,rmax,anglemax,angle,dx,dy,as,ac,dxx,dyy,R,G,B;
 rmax=min((float)(Height/2),(float)(Width/2));
 anglemax=(float)nx*t*PI/180.0;
 X1=Width-1; Y1=Height-1; X2=Width/2; Y2=Width/2;
 for(i=0;i<Height;i++)
 {
   for(j=0;j<Width;j++)
   {
     dy=(float)(i-Y2);
     dx=(float)(j-X2);
     r=sqrt(dy*dy+dx*dx);
     if(r < rmax)
     {
       //
       // Stir the region inside the radius  "rmax" calculate the angle
       // to stir as a function of radial distance from the center of
       // the image. Use either a linear or quadratic model.
       // This gives the illusion of the image "tearing" away from the
       // outside edge.
       //
       if(linear)angle = anglemax*(1.0-r/rmax);
       else angle = anglemax*((rmax-r)/rmax)*((rmax-r)/rmax);
       as=sin(angle); ac=cos(angle);
       dxx=dx*ac-dy*as;
       dyy=dx*as+dy*ac;
       //
       // The value for pixel (j,i) is to be obtained from coordinates
       // (X2+dxx, Y2+dyy) in the original image. Because the image is
       // discrete (made from pixels) we need to interpolate to obtain the
       // appropriate (RGB) values. Use bi-linear interpolation for this.
       //
       if(!InterpolateBilinear((float)X2+dxx,(float)Y2+dyy,I1,
                               Width,Height,
                               &R,&G,&B))
       {
         R=G=B=0.0;
       }
       O->Red  =(unsigned char)R;
       O->Green=(unsigned char)G;
       O->Blue =(unsigned char)B;
     }
     else
     {
       s=(I1+i*Width+j);
       O->Red  =s->Red  ;
       O->Green=s->Green;
       O->Blue =s->Blue ;
     }
     O++;
   }
 }
}

//////////////   Artistic effects Crystallize and Pointillize  ////////////
//////////////   Effects use a set or randomomly placed        ////////////
//////////////   Z buffered cones.  See "Paint by Numbers"     ////////////
//////////////   by P. Haeberli  Computer Graphics Vol 24      ////////////
//////////////   No. 4  Aug 1990 (SIGGRAPH)                    ////////////

#if 0


////////////// Helper functions ///////////////////////////////////////////

static void GetConeC(long x, long y, long r,
                     long Xmax, long Ymax,
                     unsigned char *R,
                     unsigned char *G,
                     unsigned char *B,
                     fullscreenbuffer *S){
 *R=(S+y*Xmax+x)->R;
 *G=(S+y*Xmax+x)->G;
 *B=(S+y*Xmax+x)->B;
 return;
}

static void DrawCone(long x, long y, long r,
                     long Xmax, long Ymax,
                     unsigned char R,
                     unsigned char G,
                     unsigned char B,
                     fullscreenbuffer *S,
                     float            *Z,
                     int type){
 long  i,j;
 float d,z;
 fullscreenbuffer *s;
 for(i=max(x-r,0);i<min(x+r,Xmax-1);i++)
 for(j=max(y-r,0);j<min(y+r,Ymax-1);j++){
   d=sqrt((float)(x-i)*(float)(x-i)+
          (float)(y-j)*(float)(y-j));
   if(d < (float)r){
     z=(r-d);
     if(z > *(Z+j*Xmax+i)){
       s=(S+j*Xmax+i);
       s->R=R;
       s->G=G;
       s->B=B;
       *(Z+j*Xmax+i) = z;
     }
   }
 }
 return;
}

////////////////////   Artistic effects  ////////////////////////

void CrystallizeImage(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
 unsigned char R,G,B;
 long v,x,y,i,j,iRad,Xres,Yres,nXcones,nYcones,type;
 char name_path[255];
 float rad=0.125,*Zb;
 fullscreenbuffer *S,*So;
 sscanf(PrmList,"%s %ld %ld %f",name_path,&v,&type,&rad);
 Xres=lpXimage->Xmax; Yres=lpXimage->Ymax;
 iRad=(long)(rad*(float)Xres);
// iRad=10;
// iRad=40;
 nXcones=Xres/iRad+1;
 nYcones=Yres/iRad+1;

 if(type == 0){
   iRad *= 2.0;  // for crystalize
   nXcones *= 1.5; nYcones *= 1.5;
 }
 else{
   iRad *= 0.5;  // for pointillize
   nXcones *= 1.5; nYcones *= 1.5;
 }
 So=(fullscreenbuffer *)EVI_Malloc(Xres*Yres*(long)sizeof(fullscreenbuffer));
 if(So == NULL){
   MessageBox(GetFocus(),"Memory Out",NULL,MB_OK);
   return 0;
 }
 Zb=(float *)EVI_Malloc(Xres*Yres*(long)sizeof(float));
 if(So == NULL){
   EVI_Free(So);
   MessageBox(GetFocus(),"Memory Out",NULL,MB_OK);
   return 0;
 }
 S=lpXimage->Screen;
 memset(So,0xFF,sizeof(fullscreenbuffer)*Xres*Yres);
 for(i=0;i<Xres;i++)for(j=0;j<Yres;j++)*(Zb+j*Xres+i)=0.0;
 for(i=0;i<nXcones;i++)for(j=0;j<nYcones;j++){
   x=(long)(Xres*(float)rand()/32767.0);
   y=(long)(Yres*(float)rand()/32767.0);
   GetConeC(x,y,iRad,Xres,Yres,&R,&G,&B,S);
   DrawCone(x,y,iRad,Xres,Yres,R,G,B,So,Zb,type);
 }

 memcpy(S,So,sizeof(fullscreenbuffer)*Xres*Yres);
 EVI_Free(So);
 EVI_Free(Zb);
 return;
}


void PointillizeImage(
     lpPIXEL I1,
     lpPIXEL I2,
     lpPIXEL O,
     long    Width,
     long    Height,
     float   t)
{
}

#endif
