//
//  Procedural textures. This code implements the effects illustrated
//  in chapter11 when uses by a client renderer to return a color value
//  at a specific point on a visible surface.
//  Each texture is called using the standard parameters described in the
//  text.
//

/////////////////////////////////////////////////////////////////////////
/////////////////   Prototypes for C/C++ library functions //////////////

extern float fabs(float D);   // return floating point absolute value of D
extern float sqrt(float D);   // return floating point square root of  D
extern float cos(float a);    // return the cosine of "a" radians
extern float sin(float a);    // return the sine of "a" radians
extern float floor(float a);  // return the truncated value of "a"
extern float fmod(float x, float a);   // return the floating modulo (x mod a)
extern float pow(float a,float e);     // return a**e

/////////////////////////////////////////////////////////////////////////
/////////   Defines and structures for common vector operations   ///////

/////// constants ///////////
#ifndef PI
#define PI 3.1415926
#endif
#define ROOT32  0.866025403           // root of 3 divided by 2
#define ROOT3   1.73205080
#define TRUE    1
#define FALSE   0

/////// data types //////////
#define BOOL  unsigned char
typedef float VECTOR[3];
typedef float MATRIX3[3][3];

//////// macros /////////////
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
//////// Vector arithmetic and products /////////
#define VECCOPY(a,b)    { b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; }
#define VECSUM(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2];}
#define VECSUB(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2];}
#define VECSCALE(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2];}
#define DOT(a,b)        ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define CROSS(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                        }
//////////// Integer modulo ///////////////
#define IMOD(z,a) ((z) - ((z)/(a))*(a))
///////////// Hashing function for addressing array of randon numbers
#define PATTERN(x,y,z)   ( ((z)+16384L)*64536L +  \
                           ((y)+16384L)*32768L +  \
                           ((x)+16384L))
/////////// a floor function that works with -ve numbers ////////////
#define FLOOR(x) ((x) >= 0.0 ? floor(x) : (0.0 - floor(0.0 - (x)) - 1.0))
/////////// a floating modulo function that works with -ve numbers //
#define FMOD(x,a) ((x) >= 0.0 ? fmod(x,a) :(a - fmod(-(x),a)))
// //////// noise scaled to range  [-1, 1] //////////////////////////
#define ssNoise(x,y,z) (2.0*(sNoise(x,y,z)-0.5))

/////////////////////////////////////////////////////////////////////////
////////////////// File containing a set of 2000 random vectors /////////

#include "chap9_1.h"

/////////////////////////////////////////////////////////////////////////
/////////////////   Prototypes for external functions      //////////////

extern void normalize(VECTOR v); // normalize the vector "v"

extern float fNoise(  // return a "Noise" value in the range [ 0.0 - 1.0]
             float x, // Coordinates of the point at which the Noise value
             float y, // is to be calculated.
             float z  //
            );

extern float fTurbulence( // return a " 1/f Noise" value in range [ 0.0 - 1.0]
             float x,     // Same as for fNoise function. Note that each
             float y,     // function will return the same number if called
             float z      // with the same "x y z" values.
                 );

extern void  vNoise(  // return a normalized random vector with the
             VECTOR v,// returned vector
             float x, // Coordinates of the point at which the vector
             float y, // is to be calculated.
             float z  //
            );

extern void  vTurbulence( // return a " 1/f" random vector
             VECTOR v,    // returned vector
             float x,     // Same as for fNoise function.
             float y,     // ""
             float z      // ""
                 );


/////////////////////////////////////////////////////////////////////////
//////////////////  Texturing Helper functions            ///////////////

void mix(float *r, float *g, float *b, float mx,
         float r1, float g1, float b1,
         float r2, float g2, float b2){
//
//  Mix 3 color values in the proportion "mx"
//
 *r = mx*r1 + (1.0-mx)*r2;
 *g = mx*g1 + (1.0-mx)*g2;
 *b = mx*b1 + (1.0-mx)*b2;
}

//
//  Use the cosine function to produce a "sigmoidal" like curve
//
float sigmoid(float t, float tmin, float tmax){
 t=PI*(2.0*(t-tmin)/(tmax-tmin));
 return 0.5*(1.0-cos(t));
}

//
// return the value of a "sin" function in the range [0 - 1]
// the function is periodic with period 2PI
//
float sintab(float value){
 if (value >= 0.0)
   return 0.5*(1.0+sin((value-floor(value))*PI*2));
 return 0.0;
}

//
// return a periodic "cos" function in the range [0 - 1]
//
float costab(float value){
  if (value < 0.0)return 0.0;
  return 0.5*(1.0+cos((value-floor(value))*PI*2));
}

//
//  pass the input value through a triangular function
//
float triangle(float value){
  register float offset;
  if (value >= 0.0) offset = value - floor(value);
  else offset = value - (-1.0 - floor(fabs(value)));
  if (offset >= 0.5) return (2.0 * (1.0 - offset));
  else return (2.0 * offset);
}

////////////   Table of cell topology information for textures /////////////
////////////   that need to relate a linear integer counter    /////////////
////////////   to a 3 dimensional network of adjacent cells    /////////////

static long ip[27][3]={      // For textures that need the offset of
   { 0, 0, 0},               // adjacent cells. i.e.  cell i,j,k => 0,0,0
   { 1, 0, 0},               // cell   (i-1),j,k  =>  -1,0,0
   { 1, 1, 0},               // etc. etc.
   { 0, 1, 0},
   {-1, 1, 0},
   {-1, 0, 0},
   {-1,-1, 0},
   { 0,-1, 0},
   { 1,-1, 0},
   { 0, 0,-1},
   { 1, 0,-1},
   { 1, 1,-1},
   { 0, 1,-1},
   {-1, 1,-1},
   {-1, 0,-1},
   {-1,-1,-1},
   { 0,-1,-1},
   { 1,-1,-1},
   { 0, 0, 1},
   { 1, 0, 1},
   { 1, 1, 1},
   { 0, 1, 1},
   {-1, 1, 1},
   {-1, 0, 1},
   {-1,-1, 1},
   { 0,-1, 1},
   { 1,-1, 1}
 };

/////////////////////////////////////////////////////////////////////////
/////////////////   Texturing functions                    //////////////
/////////////////////////////////////////////////////////////////////////

#include "chap9_1.cpp"
#include "chap9_2.cpp"
#include "chap9_3.cpp"
#include "chap9_4.cpp"
#include "chap9_5.cpp"
