///////////////////////////////////////////////////////////////////////////
//  Stucco1  (the basic stucco texture) half the surface area is
//           raised, i.e. plateaus.

BOOL Stucco1_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 float turb=0.0;    // rough edges
 float pos=0.48;    // where in the [0 - 1] Noise output the slope occurs
 float width=0.06;  // the width of slope
 ////////////////////////////////////
 VECTOR v,vt;
 float noise,s;
 float xr,xl,yr,yl,zr,zl;
 //
 // Move the input (sampling point) with turbulence to get rough edges
 //
 vTurbulence(v, xx, yy, zz);
 xx += v[0] * turb;
 yy += v[1] * turb;
 zz += v[2] * turb;
 //
 // The stucco textures are obtained by getting the gradient of the
 // noise function at the sampling point.
 //
 noise = fNoise(xx, yy, zz);
 //
 // fNoise is an undulating value in the range [0 - 1]  If it lies in a
 // small band [pos  - (pos+2*width)] then this is the sloping side of
 // the plateau or risges or canyons.
 //
 if(noise > pos && noise < pos+2*width)
 {
   //
   // Use finite differences to calculate the gradient of noise
   // function at (xx,yy,zz)
   //
   xr=fNoise(xx+0.01, yy, zz);
   xl=fNoise(xx-0.01, yy, zz);
   yr=fNoise(xx, yy+0.01,zz);
   yl=fNoise(xx, yy-0.01,zz);
   zr=fNoise(xx, yy,zz+0.01);
   zl=fNoise(xx, yy,zz-0.01);
   v[0]= -(xr-xl); v[1]= -(yr-yl); v[2]= -(zr-zl);
   normalize(v);
   //
   // Bend the surface normal after transforming the slope vector "v"
   // to a global frame of reference "vt".
   //
   VECSCALE(v[0],U,vt)
   VECSUM(v[1]*V,vt,vt)
   VECSUM(-v[2]*W,vt,vt)
   normalize(vt);
   //
   // The sloping edges in a Stucco texture are smoothed to give a gradual
   // transition from flat to sloping side.
   //
   if(noise < pos+width)
   {
     s=(noise-pos)/width;
     s=s*s;
   }
   else{
     s=1.0-(noise-(pos+width))/width;
     s=s*s;
   }
   VECSUM(s*vt,n,n)
   normalize(n);
 }
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
//  Stucco2  Canyons are cut into the textured surface.

BOOL Stucco2_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 int type=0;
 ////////////////////////////////////
 VECTOR v,vt;
 float t=0.48,dt=0.075,s=0.5,turb=0.4;
 float noise;
 float xr,xl,yr,yl,zr,zl;
 if(type == 1)
 {
   //
   // Move the input (sampling point) with turbulence to get
   // irregular featuer edges
   //
   vTurbulence(v, xx, yy, zz);
   xx += v[0] * turb;
   yy += v[1] * turb;
   zz += v[2] * turb;
 }
 //
 // The stucco textures are obtained by getting the gradient of the
 // noise function at the sampling point.
 //
 noise = fNoise(xx, yy, zz);
 if(noise > t && noise < t+3*dt)  // 3 regions  [down / flat / up]
 {
   //
   // Use finite differences to calculate the gradient of noise
   // function at (xx,yy,zz)
   //
   xr=fNoise(xx+0.01, yy, zz);
   xl=fNoise(xx-0.01, yy, zz);
   yr=fNoise(xx, yy+0.01,zz);
   yl=fNoise(xx, yy-0.01,zz);
   zr=fNoise(xx, yy,zz+0.01);
   zl=fNoise(xx, yy,zz-0.01);
   v[0]= -(xr-xl); v[1]= -(yr-yl); v[2]= -(zr-zl);
   normalize(v);
   VECSCALE(v[0],U,vt)
   VECSUM(v[1]*V,vt,vt)
   VECSUM(-v[2]*W,vt,vt)
   normalize(vt);
   //
   // Bend the surface normal after transforming the slope vector "v"
   // to a global frame of reference "vt".
   //
   if(noise < t+dt)                 // down  slope
   {
     s=(noise-t)/dt;
     s=-s*s;
   }
   else if(noise < t+2*dt)s=0.0;    // canyon floor
   else                             // up slope
   {
     s=1.0-(noise-(t+2*dt))/dt;
     s = s*s;
   }
   VECSUM(1.4*s*vt,n,n)
   normalize(n);
 }
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
//  Stucco3 - Ridges roughened by adding some noise

BOOL Stucco3_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 int bumpy=0;  // ridges superimposed on bumpy
 int type=0;   // ridges are rough or not (0)
 ////////////////////////////////////
 VECTOR v,vt;
 float t=0.48,dt=0.05,s=0.5,turb=0.2;
 float noise;
 float xr,xl,yr,yl,zr,zl;
 if(bumpy)
 {
   vNoise(v,xx*2.5,yy*2.5,zz*2.5);  // undulating  surface
   VECSUM(0.5*v,n,n)
   normalize(n);
 }
 //
 // Move the input (sampling point) with turbulence to get irregular edges
 //
 vTurbulence(v, xx, yy, zz);
 xx += v[0] * turb;
 yy += v[1] * turb;
 zz += v[2] * turb;
 //
 // The stucco textures are obtained by getting the gradient of the
 // noise function at the sampling point.
 //
 noise = fNoise(xx, yy, zz);
 if(noise > t && noise < t+3*dt)
 {
   //
   // Use finite differences to calculate the gradient of noise
   // function at (xx,yy,zz)
   //
   xr=fNoise(xx+0.01, yy, zz);
   xl=fNoise(xx-0.01, yy, zz);
   yr=fNoise(xx, yy+0.01,zz);
   yl=fNoise(xx, yy-0.01,zz);
   zr=fNoise(xx, yy,zz+0.01);
   zl=fNoise(xx, yy,zz-0.01);
   v[0]= -(xr-xl); v[1]= -(yr-yl); v[2]= -(zr-zl);
   normalize(v);
   //
   // Bend the surface normal after transforming the slope vector "v"
   // to a global frame of reference "vt".
   //
   VECSCALE(v[0],U,vt)
   VECSUM(v[1]*V,vt,vt)
   VECSUM(-v[2]*W,vt,vt)
   normalize(vt);
   if(noise < t+dt)                  // up slope
   {
     s=(noise-t)/dt;
     s=-s;
   }
   else if(noise < t+2*dt)s=0.0;     // top of ridge
   else{
     s=1.0-(noise-(t+2*dt))/dt;      // down slope
   }
   VECSUM(-s*vt,n,n)
   if(type == 1)
   {
     vTurbulence(vt,xx*8,yy*8,zz*8);  // Make the ridges rough
     VECSUM(0.8*vt,n,n)
   }
   normalize(n);
 }
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
//  Leather1   ( based on 4 overlapping stucco Canyons)

BOOL Leather1_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 int iturb=0;  // Irregular edges to the features
 ////////////////////////////////////
 int i;
 //
 // "ins" is used as a flag to prevent sloping side of the
 // canyons from overlapping the floor where canyons intersect.
 // Canyon floors take precedence over sides and the sloping vector is
 // only added if it occurs at a place where none of the 4 stucco
 // canyon textures have a floor.  ins=TRUE if on canyon floor.
 //
 BOOL ins;
 VECTOR v,dn,vt;
 float t=0.48,dt=0.06,s=0.5,turb=0.4;
 float noise;
 float xr,xl,yr,yl,zr,zl;
 VECCOPY(n,dn)
 ins=FALSE;
 if(iturb){
   //
   // Move the input (sampling point) with turbulence to get rough edges
   //
   vTurbulence(v, xx, yy, zz);
   xx += v[0] * turb;
   yy += v[1] * turb;
   zz += v[2] * turb;
 }
 //
 // Leather composed of 4 overlapping Stucco depressions
 //
 for(i=0;i<4;i++)
 {
 //
 // The stucco textures are obtained by getting the gradient of the
 // noise function at the sampling point.
 //
 noise = fNoise(xx, yy, zz);
 //
 // fNoise is an undulating value in the range [0 - 1]  If it lies in a
 // small band [pos  - (pos+2*width)] then this is the sloping side of
 // the plateau or risges or canyons.
 //
 if(noise > t && noise < t+3*dt)
 {
   //
   // Use finite differences to calculate the gradient of noise
   // function at (xx,yy,zz)
   //
   xr=fNoise(xx+0.01, yy, zz);
   xl=fNoise(xx-0.01, yy, zz);
   yr=fNoise(xx, yy+0.01,zz);
   yl=fNoise(xx, yy-0.01,zz);
   zr=fNoise(xx, yy,zz+0.01);
   zl=fNoise(xx, yy,zz-0.01);
   v[0]= -(xr-xl); v[1]= -(yr-yl); v[2]= -(zr-zl);
   normalize(v);
   //
   // Bend the surface normal after transforming the slope vector "v"
   // to a global frame of reference "vt".
   //
   VECSCALE(v[0],U,vt)
   VECSUM(v[1]*V,vt,vt)
   VECSUM(-v[2]*W,vt,vt)
   normalize(vt);
   if(noise < t+dt)         // down slope
   {
     s=(noise-t)/dt;
     s = -s*s;
   }
   else if(noise < t+2*dt)  // canyon floor
   {
     ins=TRUE;
     s=0.0;
   }
   else                     // up slope
   {
     s=1.0-(noise-(t+2*dt))/dt;
     s = s*s;
   }
   VECSUM(s*vt,n,n)
   normalize(n);
 }
 //
 // As each Stucco is added offset the textures origin and decrease
 // slightly the feature scale.
 //
 xx += 30; yy += 40; zz += 50;
 xx *= 1.01; yy *= 1.02; zz *= 1.03;
 } // end of "i" loop
 //
 // If on the floor of a canyon then put back the original surface
 // normal so that canyon floors take precedence over any sloping
 // sides that may occur at "p" due to one of the 4 texture applications.
 //
 if(ins)VECCOPY(dn,n)
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
//  Leather2  - This leather is based on the Pea texture so consult that
//              code for details. To transform Peas into leather the bottom
//              of the peas are flattened off.

BOOL Leather2_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 ////////////////////////////////////
 static long N=1;
 static float bradius=1.2,lradius=1.0;
 long i;
 float a;
 VECTOR u,v;
 float  dpoint[27];
 float dradius[27];
 VECTOR pp[27];
 float dx,dy,dz,cx,cy,cz,zb,ccc,dc,zbb,radius;
 long ix,iy,iz;
 long j,jid,id,pattern,patterni;
 float rad;
 VECTOR ppc;

 id=0;
 ix=FLOOR(xx);
 iy=FLOOR(yy);
 iz=FLOOR(zz);
 cx=xx-(float)ix;
 cy=yy-(float)iy;
 cz=zz-(float)iz;
 zb=-10.0;
 id = -1;

 VECCOPY(P,v)        // See Pea code for details
 VECSUM(v,cx*U,v)
 VECSUM(v,cy*V,v)
 VECSUM(v,cz*W,v)

 for(j=0;j<27;j++)  // test neighboring texture cells - global reference frame
 {
   pattern=PATTERN(ix+ip[j][0],iy+ip[j][1],iz+ip[j][2]);
   pattern=IMOD(N*pattern,1900);
   i=0;
   jid=j*N+i;
   patterni=pattern+i;
   VECSUM((float)ip[j],rand_v[patterni],ppc)
   VECSUM(P,ppc[0]*U,pp[jid])
   VECSUM(pp[jid],ppc[1]*V,pp[jid])
   VECSUM(pp[jid],ppc[2]*W,pp[jid])
   radius=(bradius+lradius*(rand_v[patterni][1]-0.5));
   dradius[jid]=radius;
   dpoint[jid] = -1.0;
   dx=(v[0]-pp[jid][0]);
   dy=(v[1]-pp[jid][1]);
   dz=(v[2]-pp[jid][2]);
   if(fabs(dx) > radius)continue;
   if(fabs(dy) > radius)continue;
   if(fabs(dz) > radius)continue;
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < radius)
   {
     dpoint[jid]=a;
     dc=(radius-a);
     if(dc > zb)
     {
       id=jid;
       zb=dc;
     }
   }
 }
 if(id >= 0)  // in sphere "id"
 {
     VECCOPY(pp[id],ppc)
     rad=dradius[id];
     VECSUB(v,ppc,u)
     if((a=sqrt((u[0])*(u[0])+
                (u[1])*(u[1])+
                (u[2])*(u[2]))) < rad)
     {                                    // Get new surface normal
       float mu,moddn,rmu;
       VECTOR c,dn;
       mu=DOT(u,n);
       c[0]=ppc[0]+mu*n[0];
       c[1]=ppc[1]+mu*n[1];
       c[2]=ppc[2]+mu*n[2];
       VECSUB(v,c,dn)
       moddn=sqrt(dn[0]*dn[0]+dn[1]*dn[1]+dn[2]*dn[2]);
       rmu=sqrt(rad*rad-mu*mu);
       if(moddn < rmu)
       {
         a=1.0-moddn/rmu;
         if(a > 0.4)                   // for flat edge to Peas
         {
           normalize(dn);
           a=pow(a,0.25);
           VECSUM((1.0-a)*dn,(a)*n,n)
           vNoise(v,xx*5,yy*5,zz*5);
           VECSUM(0.15*v,n,n)          // bend surface normal
         }
         normalize(n);
       }
     }
 }

 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
//  Cauliflower  -  Again based on the Pea texture. This time 3 pea
//                  textures are overlapped at decreasing scales.

BOOL Cauliflower_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 ////////////////////////////////////
 static long N=1;
 static float bradius=2.0,lradius=0.0;
 float a;
 VECTOR u,v;
 float  dpoint[27];
 float dradius[27];
 VECTOR pp[27];
 float dx,dy,dz,cx,cy,cz,zb,ccc,dc,zbb,radius;
 long i,ix,iy,iz;
 long k,j,jid,id,pattern,patterni;
 float rad;
 VECTOR ppc;
 //
 // Repeat the texture three times.
 //
 for(k=0;k<3;k++)
 {
 id=0;
 ix=FLOOR(xx);
 iy=FLOOR(yy);
 iz=FLOOR(zz);
 cx=xx-(float)ix;
 cy=yy-(float)iy;
 cz=zz-(float)iz;
 zb=-10.0;
 id = -1;

 VECCOPY(P,v)        // See Pea code for details
 VECSUM(v,cx*U,v)
 VECSUM(v,cy*V,v)
 VECSUM(v,cz*W,v)

 for(j=0;j<27;j++)  // test neighboring texture cells - global reference frame
 {
   pattern=PATTERN(ix+ip[j][0],iy+ip[j][1],iz+ip[j][2]);
   pattern=IMOD(N*pattern,1900);
   i=0;
   jid=j*N+i;
   patterni=pattern+i;
   VECSUM((float)ip[j],rand_v[patterni],ppc)
   VECSUM(P,ppc[0]*U,pp[jid])
   VECSUM(pp[jid],ppc[1]*V,pp[jid])
   VECSUM(pp[jid],ppc[2]*W,pp[jid])
   radius=(bradius+lradius*(rand_v[patterni][1]-0.5));
   dradius[jid]=radius;
   dpoint[jid] = -1.0;
   dx=(v[0]-pp[jid][0]);
   dy=(v[1]-pp[jid][1]);
   dz=(v[2]-pp[jid][2]);
   if(fabs(dx) > radius)continue;
   if(fabs(dy) > radius)continue;
   if(fabs(dz) > radius)continue;
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < radius)
   {
     dpoint[jid]=a;
     dc=(radius-a);
     if(dc > zb)
     {
       id=jid;              // id of sphere containing "p"
       zb=dc;
     }
   }
 }
 if(id >= 0)                // "p" lies inside sphere "id"
 {                          // see "Pea" and "random dots" code for details
     VECCOPY(pp[id],ppc)
     rad=dradius[id];
     VECSUB(v,ppc,u)
     if((a=sqrt((u[0])*(u[0])+
                (u[1])*(u[1])+
                (u[2])*(u[2]))) < rad)
     {
       float mu,moddn,rmu;            // get the displaced surface normal
       VECTOR c,dn;
       mu=DOT(u,n);
       c[0]=ppc[0]+mu*n[0];
       c[1]=ppc[1]+mu*n[1];
       c[2]=ppc[2]+mu*n[2];
       VECSUB(v,c,dn)
       moddn=sqrt(dn[0]*dn[0]+dn[1]*dn[1]+dn[2]*dn[2]);
       rmu=sqrt(rad*rad-mu*mu);
       if(moddn < rmu)
       {
         a=1.0-moddn/rmu;
         normalize(dn);
         VECSCALE(0.5,dn,dn);
         VECSUM((1.0-a)*dn,(a)*n,n)
         //
         //  As each "k" texture is applies "n" is further modified
         //
         normalize(n);
       }
     }
 }
 //
 //  On each texture combination reduce the feature scale
 //
 xx *= 4.0; yy *= 4.0; zz *=4.0;
 //

 } // overlapping textures loop "k"

 return TRUE;
}
